﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
/// <summary>
///   Standard implementation for cwCollections.
/// </summary>
unit cwCollections.Factory;
{$ifdef fpc}
  {$mode delphiunicode}
  {$MODESWITCH ADVANCEDRECORDS}
  {$MODESWITCH TYPEHELPERS}
  {$MODESWITCH NESTEDPROCVARS}
  {$ifdef CPU64}{$define CPU64BITS}{$endif}
  {$MODESWITCH ARRAYOPERATORS-}
{$endif}


interface
uses
  cwTypes
, cwCollections
;

type
  ///  <summary>
  ///    Factory record for instancing IList<T>
  ///  </summary>
  TList<T> = record
    type TArrayOfT = array of T;
    class function Create( const Granularity: nativeuint = 32; const isOrdered: boolean = false; const isPruned: boolean = false ): IList<T>; overload; static;
    class function Create( const Items: TArrayOfT ): IList<T>; overload; static;
  end;

  ///  <summary>
  ///    Factory record for instancing IRingBuffer<T>
  ///  </summary>
  TRingBuffer<T> = record
    class function Create( ItemCount: nativeuint = 128 ): IRingBuffer<T>; static;
  end;

  ///  <summary>
  ///    Factory record for instancing IStack<T>
  ///  </summary>
  TStack<T> = record
    class function Create( const Granularity: nativeuint = 0; const IsPruned: boolean = false ): IStack<T>; static;
  end;

  ///  <summary>
  ///    Factory record for instancing IDictionary<K,V>
  ///  </summary>
  TDictionary<K,V> = record
    class function Create( const Granularity: nativeuint = 32; const isOrdered: boolean = false; const isPruned: boolean = false ): IDictionary<K,V>; overload; static;
  end;

  ///  <summary>
  ///    Factory record for instancing IStringList
  ///  </summary>
  TStringList = record
    class function Create( const Granularity: nativeuint = 32; const isOrdered: boolean = false; const isPruned: boolean = false ): IStringList; overload; static;
    class function Create( const SourceStrings: TArrayOfString ): IStringList; overload; static;
  end;

implementation
uses
  cwCollections.List.Standard
, cwCollections.RingBuffer.Standard
, cwCollections.Stack.Standard
, cwCollections.Dictionary.Standard
, cwCollections.StringList.Standard
;

class function TList<T>.Create(const Granularity: nativeuint; const isOrdered: boolean; const isPruned: boolean): IList<T>;
begin
  Result := TStandardList<T>.Create( Granularity, isOrdered, isPruned );
end;

class function TRingBuffer<T>.Create(ItemCount: nativeuint): IRingBuffer<T>;
begin
  Result := TStandardRingBuffer<T>.Create( ItemCount );
end;

class function TStack<T>.Create(const Granularity: nativeuint; const IsPruned: boolean): IStack<T>;
begin
  Result := TStandardStack<T>.Create( Granularity, IsPruned );
end;

class function TDictionary<K,V>.Create( const Granularity: nativeuint = 32; const isOrdered: boolean = false; const isPruned: boolean = false ): IDictionary<K,V>;
begin
  Result := TStandardDictionary<K,V>.Create( Granularity, isOrdered, isPruned );
end;

class function TStringList.Create(const Granularity: nativeuint; const isOrdered: boolean; const isPruned: boolean): IStringList;
begin
  Result := TStandardStringList.Create( Granularity, isOrdered, isPruned );
end;

class function TStringList.Create(const SourceStrings: TArrayOfString): IStringList;
var
  S: string;
begin
  Result := TStringList.Create;
  if Length( SourceStrings ) = 0 then exit;
  for S in SourceStrings do Result.Add( S );
end;

class function TList<T>.Create(const Items: TArrayOfT): IList<T>;
var
  idx: nativeuint;
begin
  Result := TList<T>.Create;
  if Length(Items)=0 then exit;
  for idx := 0 to pred(Length(Items)) do Result.Add(Items[idx]);
end;

end.
