﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwDebugLog;

interface

type
  IDebugLog = interface
    ['{83619226-7127-40F6-BB67-A0B93E6EFB54}']
    {$ifdef DEBUG}
    procedure DebugLn( const Msg: string );
    {$endif}
  end;


{$ifndef DEBUG}
// This one is here when not debugging, so that it will be optimized out in release builds
procedure DebugLn( const Msg: string );
{$endif}

function DebugLog( const _Class: TObject; const Method: string ): IDebugLog;

implementation
{$ifdef DEBUG}
uses
  cwWin32
, cwPosix
;
{$endif}

{$ifndef DEBUG}
procedure DebugLn( const Msg: string );
begin
end;

function DebugLog( const _Class: TObject; const Method: string ): IDebugLog;
begin
  Result := nil;
end;
{$endif}

{$ifdef DEBUG}

var
  Mutex: {$ifdef MSWINDOWS} SRWLOCK; {$else} pthread_mutex_t; {$endif}

procedure AcquireLock;
begin
  {$ifdef MSWINDOWS}
  AcquireSRWLockExclusive(Mutex);
  {$else}
  pthread_mutex_lock(@Mutex);
  {$endif}
end;

procedure ReleaseLock;
begin
  {$ifdef MSWINDOWS}
  ReleaseSRWLockExclusive(Mutex);
  {$else}
  pthread_mutex_unlock(@Mutex)
  {$endif}
end;

procedure CreateLock;
begin
  {$ifdef MSWINDOWS}
  InitializeSRWLock(Mutex);
  {$else}
  pthread_mutex_init(@Mutex, nil);
  {$endif}
end;

procedure DisposeLock;
begin
  {$ifndef MSWINDOWS}
  pthread_mutex_destroy(@Mutex);
  {$endif}
end;

type
  TDebugLog = class( TInterfacedObject, IDebugLog )
  private
    fUnit: string;
    fClass: string;
    fMethod: string;
  strict private // IDebugLog
    procedure DebugLn( const Msg: string );
  public
    constructor Create( const _Class: TObject; const Method: string );
    destructor Destroy; override;
  end;

constructor TDebugLog.Create(const _Class: TObject; const Method: string);
begin
  inherited Create;
  AcquireLock;
  try
    fMethod := Method;
    if assigned(_Class) then begin
      fUnit := _Class.UnitName;
      fClass := _Class.ClassName;
    end else begin
      fUnit := '';
      fClass := '';
    end;
  finally
    ReleaseLock;
  end;
  DebugLn('ENTER');
end;

procedure TDebugLog.DebugLn(const Msg: string);
var
  Message: string;
begin
  Message := '{DEBUG} ';
  if (fUnit<>'') then begin
    Message := Message+fUnit+'.';
  end;
  if (fClass<>'') then begin
    Message := Message+fClass+'.';
  end;
  if (fMethod<>'') then begin
    Message := Message+fMethod;
  end;
  Message := Message + ' ' + Msg;
  AcquireLock;
  try
    Writeln(Message);
  finally
    ReleaseLock;
  end;
end;

destructor TDebugLog.Destroy;
begin
  DebugLn('EXIT');
  inherited;
end;

function DebugLog( const _Class: TObject; const Method: string ): IDebugLog;
begin
  Result := TDebugLog.Create( _Class, Method );
end;

initialization
  CreateLock;

finalization
  DisposeLock;

{$endif}

end.
