﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwDynLib.Factory;
{$ifdef fpc}
  {$mode delphiunicode}
  {$MODESWITCH ADVANCEDRECORDS}
  {$MODESWITCH TYPEHELPERS}
  {$MODESWITCH NESTEDPROCVARS}
  {$ifdef CPU64}{$define CPU64BITS}{$endif}
  {$MODESWITCH ARRAYOPERATORS-}
{$endif}

interface
uses
  cwDynLib
, cwTypes
, cwDynlib.HandledObject.Standard
;

type
  TDynlib = class( TInterfacedObject, IDynlib )
  private
    fLoaded: boolean;
    fLibrary: string;
    fHandle: nativeuint;
  protected //- IDynlib -//
    function getLoaded: boolean;
    function LoadLibrary( const filepath: string ): TStatus;
    function FreeLibrary: boolean;
    function GetProcAddress( const funcName: string; out ptrProc: pointer ): TStatus;
  public
    constructor Create; reintroduce;
    destructor Destroy; override;
  end;

type
  ///  <summary>
  ///    This type helper for Handle provides the Resolve() method, which
  ///    may be used within a module entry point to resolve an instance of
  ///    the object for which a handle was generated. Example call:
  ///    <br/>
  ///    procedure DoSomething( HObject: Handle );
  ///    var
  ///      obj: IMyHandledObject;
  ///    begin
  ///      if not HObject.Resolve(obj).Success then begin
  ///        //- Handle invalid status here.
  ///      end;
  ///      //- Use Obj here.
  ///    end;
  ///  </summary>
  {$ifndef fpc}
  THandleHelper = record helper for Handle
  {$else}
  THandleHelper = type helper for Handle
  {$endif}
    function Resolve(out  anObject; const _interface: TGUID): TStatus;
    function getExtra1(out extra: pointer): TStatus;
    function setExtra1(const extra: pointer): TStatus;
    function getExtra2(out extra: pointer): TStatus;
    function setExtra2(const extra: pointer): TStatus;
  end;

  ///  <summary>
  ///    Derrive THandledObject in order to implement IHandledObject and
  ///    make an object safe for exposing accross the module boundary.
  ///  </summary>
  THandledObject = cwDynlib.HandledObject.Standard.THandledObject;

implementation
uses
  cwWin32
, cwPosix
;

function THandleHelper.Resolve(out anObject; const _interface: TGUID): TStatus;
var
  P: pHandle;
begin
  P := Self;
  if assigned(P)
     and (pHandle(P)^._signature.EqualTo(cValidGUID))
     and (assigned(pHandle(P)^._object)) then begin
    Supports(P^._object,_Interface,anObject);
    Result := TStatus.Success;
  end else begin
    IUnknown(anObject) := nil;
    Result := TStatus(stInvalidHandle);
  end;
end;

function THandleHelper.getExtra1(out extra: pointer): TStatus;
var
  P: pHandle;
begin
  P := Self;
  if assigned(P)
     and (pHandle(P)^._signature.EqualTo(cValidGUID))
     and (assigned(pHandle(P)^._object)) then begin
    extra := P^._extra1;
    Result := TStatus.Success;
  end else begin
    extra := nil;
    Result := TStatus(stInvalidHandle);
  end;
end;

function THandleHelper.setExtra1(const extra: pointer): TStatus;
var
  P: pHandle;
begin
  P := Self;
  if assigned(P)
     and (pHandle(P)^._signature.EqualTo(cValidGUID))
     and assigned(pHandle(P)^._object) then begin
    P^._extra1 := extra;
    Result := TStatus.Success;
  end else begin
    Result := TStatus(stInvalidHandle);
  end;
end;

function THandleHelper.getExtra2(out extra: pointer): TStatus;
var
  P: pHandle;
begin
  P := Self;
  if assigned(P)
     and (pHandle(P)^._signature.EqualTo(cValidGUID))
     and assigned(pHandle(P)^._object) then begin
    extra := P^._extra2;
    Result := TStatus.Success;
  end else begin
    extra := nil;
    Result := TStatus(stInvalidHandle);
  end;
end;

function THandleHelper.setExtra2(const extra: pointer): TStatus;
var
  P: pHandle;
begin
  P := Self;
  if assigned(P)
     and (pHandle(P)^._signature.EqualTo(cValidGUID))
     and assigned(pHandle(P)^._object) then begin
    P^._extra2 := extra;
    Result := TStatus.Success;
  end else begin
    Result := TStatus(stInvalidHandle);
  end;
end;

constructor TDynlib.Create;
begin
  inherited Create;
  fLoaded := False;
  fHandle := 0;
end;

destructor TDynlib.Destroy;
begin
  if fLoaded and (fHandle<>0) then begin
    FreeLibrary;
  end;
  inherited Destroy;
end;

function TDynlib.FreeLibrary: boolean;
begin
  Result := False;
  if not fLoaded then exit;
  if fHandle=0 then exit;
  {$ifdef MSWINDOWS}
  if not cwWin32.FreeLibrary(fHandle) then begin
    fHandle := 0;
    exit;
  end;
  {$else}
  {$hints off}
  if not (dlClose(pointer(fHandle)) = cSuccess) then begin
  {$hints on}
    fHandle := 0;
    exit;
  end;
  {$endif}
  fHandle := 0;
  Result := True;
end;

function TDynlib.GetProcAddress(const funcName: string; out ptrProc: pointer): TStatus;
var
  utf8FuncName: pointer;
begin
  Result := TStatus.Unknown;
  utf8FuncName := TUtf8String.Alloc(funcName);
  try
    {$ifdef MSWINDOWS}
      ptrProc := cwWin32.GetProcAddress( fHandle, utf8FuncName );
    {$else}
      {$hints off}
      ptrProc := dlSym( pointer(fHandle), utf8FuncName );
      {$hints on}
    {$endif}
    if not assigned(ptrProc) then begin
      Result := stFailedToLoadEntryPoint;
      exit;
    end;
  finally
    TUtf8String.Free(utf8FuncName);
  end;
  Result := TStatus.Success;
end;

function TDynlib.getLoaded: boolean;
begin
  Result := fLoaded;
end;

function TDynlib.LoadLibrary(const filepath: string): TStatus;
var
  ucFilepath: pointer;
begin
  Result := TStatus.Unknown;
  {$ifdef MSWINDOWS}
  ucFilePath := TUTF16LEString.Alloc(filepath);
  try
    fHandle := cwWin32.LoadLibraryW( ucFilepath );
  finally
    TUTF16LEString.Free(ucFilePath);
  end;
  {$else}
  ucFilePath := TUTF8String.Alloc(filepath);
  try
    {$hints off} fHandle := NativeUInt( dlOpen(ucFilepath,cRTLD_LAZY) ); {$hints on}
  finally
    TUTF8String.Free(ucFilePath);
  end;
  {$endif}
  if fHandle=0 then begin
    Result := stModuleNotLoaded;
    exit;
  end;
  fLoaded := True;
  fLibrary := filepath;
  Result := TStatus.Success;
end;

end.


