﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
/// <summary>
///   Cross platform dynamic library loader.
/// </summary>
unit cwDynLib;
{$ifdef fpc}
  {$mode delphiunicode}
  {$MODESWITCH ADVANCEDRECORDS}
  {$MODESWITCH TYPEHELPERS}
  {$MODESWITCH NESTEDPROCVARS}
  {$ifdef CPU64}{$define CPU64BITS}{$endif}
  {$MODESWITCH ARRAYOPERATORS-}
{$endif}

interface
uses
  cwTypes
;

const
  stModuleNotLoaded        : TGUID = '{C4AEB199-88D8-4DF8-9144-604D4114F5FF}';
  stFailedToLoadEntryPoint : TGUID = '{5F7F9A43-9B2B-47E8-B861-B76888ECE6A5}';
  stInvalidHandle          : TGUID = '{CFADB9BD-15F1-406D-B84A-B7E81AEFE34E}';

type
  /// <summary>
  ///   An instance of IDynlib represents a dynamic library, such as a .dll or
  ///   .so file.
  /// </summary>
  IDynlib = interface
    ['{AA731CC2-8779-4F83-8117-F481DDD2B48D}']

    ///  <summary>
    ///    Returns TRUE if there is a library currently loaded. <br/>
    ///    Getter for Loaded property.
    ///  </summary>
    function getLoaded: boolean;

    /// <summary>
    ///   Loads a library from disk into memory.
    /// </summary>
    /// <param name="filepath">
    ///   Specifies the full path and filename of the library to be loaded.
    ///   (Relative paths permitted based on target implementation).
    /// </param>
    /// <returns>
    /// </returns>
    function LoadLibrary( const filepath: string ): TStatus;

    /// <summary>
    ///   Unloads the library (previously loaded using the LoadLibrary()
    ///   method) from memory.
    /// </summary>
    /// <returns>
    /// </returns>
    function FreeLibrary: boolean;

    /// <summary>
    ///   Locates a symbol within a library (previously loaded using the
    ///   LoadLibrary() method), and returns a pointer to it.
    /// </summary>
    /// <param name="funcName">
    ///   The name of the function or symbol to locate.
    /// </param>
    /// <returns>
    ///   If successful, returns a pointer to the requested symbol. Otherwise
    ///   returns false. Typically a failure is caused either by not having
    ///   first loaded the library using the LoadLibrary() method, or because
    ///   of an incorrect symbol name specified in the funcName parameter.
    ///   Symbol names may be case sensitive, depending upon the implementation
    ///   and target.
    /// </returns>
    function GetProcAddress( const funcName: string; out ptrProc: pointer ): TStatus;

    ///  <summary>
    ///    Returns TRUE if there is a library currently loaded. <br/>
    ///  </summary>
    property Loaded: boolean read getLoaded;
  end;

type
  ///  <summary>
  ///    Used as a handle to an object allocated in another module. <br/>
  ///    Note: NOT a pointer to the object, but rather, a pointer to a
  ///    handle-object which may be resolved to the actual object within
  ///    the module in which it was allocated.
  ///  </summary>
  Handle = pointer;

type
  ///  <summary>
  ///    Reference counted objects which may be passed across the module
  ///    boundry will implement this interface. (As a convenience, derrive
  ///    from THandledObject). <br/>
  ///    These objects will be memory managed by the module which creates
  ///    them, which means that the consuming module may have an invalid
  ///    handle when the creating module disposes an object. <br/>
  ///    In order to account for the invalid handle possibility, the object
  ///    handle must be verified by any entry-point which receives a handle.
  ///    This can be done by calling the Handle.Resolve() method, applied
  ///    by type-helper.
  ///  </summary>
  IHandledObject = interface
    ['{29734EB7-74D5-4854-944F-5D1B84817FCE}']

    ///  <summary>
    ///    Returns this objects handle.
    ///  </summary>
    function getHandle: Handle;
  end;

type
   {$ifdef fpc}
    TEnumerateMethodGlobal   = procedure( const item: pointer; var breakout: boolean );
    TEnumerateMethodOfObject = procedure( const item: pointer; var breakout: boolean ) of object;
    TEnumerateMethodIsNested = procedure( const item: pointer; var breakout: boolean ) is nested;
    {$else}
    TEnumerateMethod         = reference to procedure( const item: pointer; var breakout: boolean );
    TEnumerateMethodLocal    = procedure( const item: pointer; var breakout: boolean );
    {$endif}  

type
  ///  <summary>
  ///    A utility record for managing a zero-terminated array of pointers
  ///    for passing arrays of records/strings across the module boundry. <br/>
  ///    A pointer array is allocated by adding the first item to the
  ///    array, and retaining the pointer. i.e. <br/>
  ///
  ///        lpPtrArray := TPointerArray.Add( nil, NewPointer );
  ///
  ///    What is returned is a pointer to the first element which contains
  ///    the new pointer, which is followed by a second element which
  ///    contains the array zero-terminator. <br/>
  ///  </summary>
  TPointerArray = record
  public
    ///  <summary>
    ///    Given a pointer to the first element in the array, this method
    ///    returns the number
    ///  </summary>
    class function Count( const pFirstElement: pointer ): nativeuint; static;

    ///  <summary>
    ///    Disposes of the memory used to store the array whos first item is
    ///    pointed to by pFirstElement.
    ///  </summary>
    class procedure Free( const pFirstElement: pointer ); static;

    ///  <summary>
    ///    Allocates space for an additional pointer in the array and adds the
    ///    new pointer to the end of the array. <br/>
    ///    If NewPointer is nil, this method does nothing. <br/>
    ///    If the pFirstElement pointer is nil, a new array is allocated. <br/>
    ///    Returns a pointer to the FIRST element in the array, as the array
    ///    will be reallocated to accomodate the new pointer. (!!!)
    ///  </summary>
    class function Add( const pFirstElement: pointer; const NewPointer: pointer ): pointer; static;

    ///  <summary>
    ///    Returns an item from the array by index. <br/>
    ///    Caution: If the index is out of bounds, this method simply returns nil.
    ///  </summary>
    class function GetItem( const pFirstElement: pointer; const idx: nativeuint ): pointer; static;

    ///  <summary>
    ///    Iterates over each item in the array and calls the for-each handler
    ///    for that item. (ecludes the zero terminator).
    ///  </summary>
    {$ifdef fpc}
    class procedure ForEach( const pFirstElement: pointer; const EnumMethod: TEnumerateMethodGlobal ); overload; static;
    class procedure ForEach( const pFirstElement: pointer; const EnumMethod: TEnumerateMethodOfObject ); overload; static;
    class procedure ForEach( const pFirstElement: pointer; const EnumMethod: TEnumerateMethodIsNested ); overload; static;
    {$else}
    class procedure ForEach( const pFirstElement: pointer; const EnumMethod: TEnumerateMethod ); overload; static;
    class procedure ForEach( const pFirstElement: pointer; const EnumMethod: TEnumerateMethodLocal ); overload; static;
    {$endif}

  end;

implementation

class function TPointerArray.Count(const pFirstElement: pointer): nativeuint;
var
  pCurrentElement: pointer;
begin
  Result := 0;
  if not assigned(pFirstElement) then exit;
  pCurrentElement := pFirstElement;
  while assigned(pointer(pCurrentElement^)) do begin
    inc(Result);
    {$hints off} pCurrentElement := pointer( nativeuint(pCurrentElement) + sizeof(pointer) ); {$hints on}
  end;
end;

class function TPointerArray.Add(const pFirstElement: pointer; const NewPointer: pointer): pointer;
var
  NewElementPtr: pointer;
  uCount: nativeuint;
begin
  Result := pFirstElement;
  uCount := TPointerArray.Count( pFirstElement );
  if NewPointer=nil then exit;
  //- Special case if the array is empty, allocate space for two pointers,
  //- to include the nil.
  if uCount = 0 then begin
    {$ifndef fpc}
    GetMem(Result, sizeof(pointer)*2);
    {$else}
    Result := GetMem(sizeof(pointer)*2);
    {$endif}
    FillChar(Result^,sizeof(pointer)*2,0);
    Move(NewPointer,Result^,sizeof(pointer));
    exit;
  end;
  //- If we have an existing array, it must be reallocated in order to add the new one.
  inc(uCount); // increment ucount to include the existing zero terminator.
  {$ifndef fpc}
  GetMem(Result, sizeof(pointer)*succ(uCount));                                                       //- Allocate space for the existing array and one new pointer.
  {$else}
  Result := GetMem(sizeof(pointer)*succ(uCount));                                                       //- Allocate space for the existing array and one new pointer.
  {$endif}
  FillChar(Result^,sizeof(pointer)*succ(uCount),0);                                                     //- Clear out the new space.
  Move(pFirstElement^,Result^,sizeof(pointer)*uCount);                                                  //- Copy the old data to the new array.
  {$hints off} NewElementPtr := pointer( nativeuint( Result ) + (sizeof(pointer)*pred(uCount)) ); {$hints on} //- Get a pointer to the new element.
  Move(NewPointer,NewElementPtr^,sizeof(Pointer));                                                      //- Copy the new element into the array.
  TPointerArray.Free(pFirstElement);                                                                    //- Free the old array, we have a new one now.
end;

class function TPointerArray.GetItem(const pFirstElement: pointer; const idx: nativeuint): pointer;
var
  pCurrentElement: pointer;
  counter: nativeuint;
begin
  Result := nil;
  if not assigned(pFirstElement) then exit;
  counter := 0;
  pCurrentElement := pFirstElement;
  while assigned(pointer(pCurrentElement^)) do begin
    if (counter=idx) then begin
      Result := pointer(pCurrentElement^);
      exit;
    end;
    inc(counter);
    {$hints off} pCurrentElement := pointer( nativeuint(pCurrentElement) + sizeof(pointer) ); {$hints on}
  end;
end;


{$ifdef fpc}
class procedure TPointerArray.ForEach(const pFirstElement: pointer; const EnumMethod: TEnumerateMethodGlobal);
var
  pCurrentElement: ^pointer;
  breakout: boolean;
begin
  if pFirstElement=nil then exit;
  breakout := false;
  pCurrentElement := pFirstElement;
  while assigned(pointer(pCurrentElement^)) do begin
    EnumMethod(pCurrentElement^, breakout);
    if breakout then exit;
    {$hints off} pCurrentElement := pointer( nativeuint(pCurrentElement) + sizeof(pointer) ); {$hints on}
  end;
end;
{$endif}

{$ifdef fpc}
class procedure TPointerArray.ForEach(const pFirstElement: pointer; const EnumMethod: TEnumerateMethodOfObject);
var
  pCurrentElement: ^pointer;
  breakout: boolean;
begin
  if pFirstElement=nil then exit;
  breakout := false;
  pCurrentElement := pFirstElement;
  while assigned(pointer(pCurrentElement^)) do begin
    EnumMethod(pCurrentElement^, breakout);
    if breakout then exit;
    {$hints off} pCurrentElement := pointer( nativeuint(pCurrentElement) + sizeof(pointer) ); {$hints on}
  end;
end;
{$endif}

{$ifdef fpc}
class procedure TPointerArray.ForEach(const pFirstElement: pointer; const EnumMethod: TEnumerateMethodIsNested);
var
  pCurrentElement: ^pointer;
  breakout: boolean;
begin
  if pFirstElement=nil then exit;
  breakout := false;
  pCurrentElement := pFirstElement;
  while assigned(pointer(pCurrentElement^)) do begin
    EnumMethod(pCurrentElement^, breakout);
    if breakout then exit;
    {$hints off} pCurrentElement := pointer( nativeuint(pCurrentElement) + sizeof(pointer) ); {$hints on}
  end;
end;
{$endif}

{$ifndef fpc}
class procedure TPointerArray.ForEach( const pFirstElement: pointer; const EnumMethod: TEnumerateMethod );
var
  pCurrentElement: ^pointer;
  breakout: boolean;
begin
  if pFirstElement=nil then exit;
  breakout := false;
  pCurrentElement := pFirstElement;
  while assigned(pointer(pCurrentElement^)) do begin
    EnumMethod(pCurrentElement^, breakout);
    if breakout then exit;
    {$hints off} pCurrentElement := pointer( nativeuint(pCurrentElement) + sizeof(pointer) ); {$hints on}
  end;
end;

class procedure TPointerArray.ForEach( const pFirstElement: pointer; const EnumMethod: TEnumerateMethodLocal );
var
  pCurrentElement: ^pointer;
  breakout: boolean;
begin
  if pFirstElement=nil then exit;
  breakout := false;
  pCurrentElement := pFirstElement;
  while assigned(pointer(pCurrentElement^)) do begin
    EnumMethod(pCurrentElement^, breakout);
    if breakout then exit;
    {$hints off} pCurrentElement := pointer( nativeuint(pCurrentElement) + sizeof(pointer) ); {$hints on}
  end;
end;

{$endif}

class procedure TPointerArray.Free(const pFirstElement: pointer);
var
  uCount: nativeuint;
begin
  if pFirstElement=nil then exit;
  uCount := TPointerArray.Count( pFirstElement );
  inc(uCount); //- to include nil terminator.
  FreeMem(pFirstElement,sizeof(pointer)*uCount);
end;


initialization
  TStatus(stModuleNotLoaded        ).Register('Dynlib: Module not loaded.');
  TStatus(stFailedToLoadEntryPoint ).Register('Dynlib: failed to load entry point.');
  TStatus(stInvalidHandle          ).Register('Invalid handle.');

end.

