﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwLog.Factory;
{$ifdef fpc}{$mode delphiunicode}{$endif}

interface
uses
  cwLog
, cwIO
;

type
  ///  <summary>
  ///    Factory record for instancing ILog.
  ///  </summary>
  TLog = record

    ///  <summary>
    ///    Creates a new instance of ILog from the standard implementation. <br/>
    ///    See SysLog() before instancing a new log.
    ///  </summary>
    class function Create: ILog; static;

    ///  <summary>
    ///    Creates a log target which simply outputs to the console. <br/>
    ///    *Note: Conditionally compiled out for UI applications in Delphi.
    ///  </summary>
    class function CreateConsoleLogTarget: ILogTarget; static;

    ///  <summary>
    ///    Creates a stream to target a unicode stream,
    ///    such as a file-stream.
    ///  </summary>
    class function CreateStreamLogTarget( const TargetStream: IUnicodeStream; const Format: TUnicodeFormat; const WriteBOM: boolean = FALSE ): ILogTarget; static;

    ///  <summary>
    ///    Creates a log target which writes to a utf-8 format text file. <br/>
    ///    The CreateMode parameter sets the policy for logging to a file which
    ///    already exists. The file may be overwritten or may be appended.
    ///  </summary>
    class function CreateFileLogTarget( const Filepath: string; const CreateMode: TFileCreateMode = fcAppend ): ILogTarget; static;
  end;

implementation
uses
  cwTypes
, cwLog.Log.Standard
, cwLog.LogTarget.Console
, cwLog.LogTarget.Stream
, cwLog.LogTarget.FileTarget
;

class function TLog.Create: ILog;
begin
  Result := cwLog.Log.Standard.TLog.Create;
end;

class function TLog.CreateConsoleLogTarget: ILogTarget;
begin
  Result := TConsoleLogTarget.Create;
end;

class function TLog.CreateStreamLogTarget(const TargetStream: IUnicodeStream; const Format: TUnicodeFormat; const WriteBOM: boolean): ILogTarget;
begin
  Result := TStreamLogTarget.Create( TargetStream, Format, WriteBOM );
end;

class function TLog.CreateFileLogTarget(const Filepath: string; const CreateMode: TFileCreateMode): ILogTarget;
begin
  try
    Result := TFileLogTarget.Create(Filepath,CreateMode);
  except
    Result := nil;
    raise;
  end;
end;

end.

