﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwLog;
{$ifdef fpc}{$mode delphiunicode}{$endif}

interface
uses
  cwTypes
, cwCollections
;

type
  ///  <summary>
  ///    Enumeration to provide an indication of the severity of
  ///    a log entry.
  ///  </summary>
  TLogSeverity = (

    ///  <summary>
    ///    Log entries inserted with a severity of lsNone are
    ///    simply ignored. This log level is used when setting
    ///    the log level to surpress logging entirely.
    ///  </summary>
    lsNone    = $00,

    ///  <summary>
    ///    Reserved to provide information to the developer during
    ///    debugging. Generally, the log level of 'lsDebug' will be
    ///    set within conditional defines, such that the log entry
    ///    is only inserted into the log targets while debugging
    ///    the application.
    ///  </summary>
    lsDebug   = $01,

    ///  <summary>
    ///    Used to provide logging of low-level events for use during
    ///    fault finding. lsVerbose is most useful for fault finding
    ///    for host/environment issues. For example, verbose logging
    ///    may be used on release level builds in q/a or production
    ///    environments during testing or stand-up, to identify
    ///    faults specific to the host environment.
    ///  </summary>
    lsVerbose = $02,

    ///  <summary>
    ///    Application level logging to inform the end-user of
    ///    information about what the application is doing at
    ///    any given time.
    ///  </summary>
    lsInfo    = $03,

    ///  <summary>
    ///    Hints are usually inserted into the log to provide the end
    ///    user of the application with notifications of configuration
    ///    settings which could be altered to improve performance, stability
    ///    or other key metrics for the application.
    ///  </summary>
    lsHint    = $04,

    ///  <summary>
    ///    Warnings are usually inserted into the log to caution the
    ///    end user of undesirable configuration or operation states which
    ///    do not necessarily indicate a fault, but could lead to fault
    ///    at a later stage.
    ///  </summary>
    lsWarning = $05,

    ///  <summary>
    ///    Errors are inserted into the log when the application enters an
    ///    erroneous state, but may still continue to function or recover.
    ///  </summary>
    lsError   = $06,

    ///  <summary>
    ///    Fatal errors are inserted into the log when an error state has
    ///    been encountered from which the application is unable to recover.
    ///    While fatal log entries are the last entry inserted into the log
    ///    before the application shut-down process begins. (Not necessarily
    ///    the very last message since the application may continue logging
    ///    during shutdown)
    ///  </summary>
    lsFatal   = $07
  );

type
  /// <summary>
  ///   For creating a log file, if the file
  ///   already exists, specify if the file should
  ///   be overwritten or appended.
  /// </summary>
  TFileCreateMode = (

    /// <summary>
    ///   Overwrite file if it exists.
    /// </summary>
    fcOverwrite,

    /// <summary>
    ///   Append the file if it exists.
    /// </summary>
    fcAppend
  );

type
  ///  <summary>
  ///    Provides a target file, stream, output for log entries
  ///    to be written to.
  ///  </summary>
  ILogTarget = interface
    ['{A3C99B81-D426-4371-AAD7-A1F6CAA18C72}']

    ///  <summary>
    ///    Inserts a log entry into this log target. <br/>
    ///    Note: Each log target implementation may used the parameters differently.
    ///    The default behavior of a textual log entry is to simply insert the Translated string
    ///    into the log target, however, additional implementations may be created to intercept
    ///    specific log entries for filtering. i.e. We might log into a database, and wish to
    ///    capture the severity parameter into it's own indexed field. <br/>
    ///    See the documentation for your chosen log target understand how it behaves with
    ///    regards to these parameters. <br/>
    ///    <param name="Translated"> The full text string to be inserted into the log target. </param>
    ///    <param name="Status"> The TStatus (GUID) representation of the log entry. </param>
    ///    <param name="Severity"> The level of severity of the log entry. </param>
    ///    <param name="Parameters"> The parameter values supplied for the log entry. </param>
    ///  </summary>
    procedure Insert( const Translated: string; const Status: TStatus; const Severity: TLogSeverity; const Parameters: TArrayOfString );
  end;

const
  stLogMessage: TGUID = '{7F57E4F8-2BC5-4375-9BA5-9B0CDF9375CB}';

type
  ///  <summary>
  ///    Represents a single log which may send log messages to
  ///    any of a number of log targets. <br/>
  ///  </summary>
  ILog = interface
    ['{D8CDEAFE-40FB-4287-903F-455BB4805A27}']

    ///  <summary>
    ///    Returns the date/time format string used to
    ///    format timestamps into the log. This string
    ///    should match that required by fpc/delphi's
    ///    FormatDateTime() function. <br/>
    ///    Defaults to 'YYYY-MM-DD HH:nn:SS:zzzz' <br/>
    ///    Getter for DateTimeFormat property.
    ///  </summary>
    function getDateTimeFormat: string;

    ///  <summary>
    ///    Sets the date/time format string used to
    ///    format timestamps into the log. This string
    ///    should match that required by fpc/delphi's
    ///    FormatDateTime() function. <br/>
    ///    Defaults to 'YYYY-MM-DD HH:nn:SS:zzzz' <br/>
    ///    Setter for DateTimeFormat property.
    ///  </summary>
    procedure SetDateTimeFormat( const Format: string );

    ///  <summary>
    ///    Returns a reference to the list of log targets
    ///    installed into this log. Each log entry raised, will be
    ///    sent to each log target within this list. <br/>
    ///    Getter for Targets property.
    ///  </summary>
    function GetTargets: IReadOnlyList<ILogTarget>;

    ///  <summary>
    ///    The log level indicates the lowest log severity which will
    ///    be accepted by the log. Levels beneath this are simply discarded
    ///    upon insertion. See setLogLevel(); <br/>
    ///    Getter for LogLevel property.
    ///  </summary>
    function GetLogLevel: TLogSeverity;

    ///  <summary>
    ///    The log level indicates the lowest log severity which will
    ///    be accepted by the log. Levels beneath this are simply discarded. <br/>
    ///    Setter for LogLevel property.
    ///  </summary>
    procedure SetLogLevel( const Value: TLogSeverity );

    ///  <summary>
    ///    Inserts an entry into the log, and therefore into the associated log targets. <br/>
    ///    Parameters is optional through overload.
    ///  </summary>
    function Insert( const Status: TStatus; const Severity: TLogSeverity; const Parameters: TArrayOfString ): TStatus; overload;
    ///  <exclude/>
    function Insert( const Status: TStatus; const Severity: TLogSeverity ): TStatus; overload;
    ///  <exclude/> (provides additional insert between log severitry and message, useful in derivatives
    function Insert( const Status: TStatus; const Severity: TLogSeverity; const Additional: string; const Parameters: TArrayOfString ): TStatus; overload;

    ///  <summary>
    ///    Removes any log targets so that the target(s) may be dynamically
    ///    altered at runtime.
    ///  </summary>
    procedure ClearLogTargets;

    ///  <summary>
    ///    Adds a target to the log to recieve log entries as they are inserted. <br/>
    ///  </summary>
    procedure AddTarget( const Target: ILogTarget );

    ///  <summary>
    ///    Returns a reference to the list of log targets
    ///    installed into this log. Each log entry raised, will be
    ///    sent to each log target within this list. <br/>
    ///  </summary>
    property Targets: IReadOnlyList<ILogTarget> read getTargets;

    ///  <summary>
    ///    The log level indicates the lowest log severity which will
    ///    be accepted by the log. Levels beneath this are simply discarded
    ///    upon insertion.
    ///  </summary>
    property Level: TLogSeverity read getLogLevel write setLogLevel;

    ///  <summary>
    ///    The date/time format string used to
    ///    format timestamps into the log. This string
    ///    should match that required by fpc/delphi's
    ///    FormatDateTime() function. <br/>
    ///    Defaults to 'YYYY-MM-DD HH:nn:SS:zzzz' <br/>
    ///  </summary>
    property DateTimeFormat: string read getDateTimeFormat write setDateTimeFormat;
  end;

implementation

initialization
  TStatus(stLogMessage).Register('(%message%)');

end.

