﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwThreading.Factory;
{$ifdef fpc}
  {$mode delphiunicode}
  {$modeswitch nestedprocvars}
{$endif}

interface
uses
  cwTypes
, cwCollections
, cwThreading
, cwThreading.Promise.Custom
{$ifdef fpc}
, cwThreading.PromiseAction.Standard
{$endif}
{$ifndef fpc}
, rtti
{$endif}
;

type
  ///  <summary>
  ///    Factory record for instancing ICriticalSection
  ///  </summary>
  TCriticalSection = record
    class function Create: ICriticalSection; static;
  end;


type
  ///  <summary>
  ///    Factory record for instancing ISignaledCriticalSection
  ///  </summary>
  TSignaledCriticalSection = record
    class function Create( const SleepTimeoutMilliseconds: uint64 = 0 ): ISignaledCriticalSection; static;
  end;

type
  /// <summary>
  ///   <para>
  ///     Descend TPromise to create your own classes which are able to be
  ///     called from any thread, but which perform their methods within the
  ///     promise thread. <br />
  ///   </para>
  ///   <para>
  ///     <b>Delphi Conditions</b><br />Your derived promise class must have
  ///     type info enabled with the <i>{$M+}</i> directive. <br /><br />Any
  ///     methods which will be called as promises, must either be published,
  ///     or else, the following directive must be used to specify the
  ///     protection level to include in RTTI. <br />
  ///   </para>
  ///   <code lang="Delphi">{$RTTI EXPLICIT METHODS([vcPublic, vcProtected, vcPrivate])} </code>
  ///   <para>
  ///     <br /><b>FPC Conditions</b><br />You must include the directive to
  ///     enable nested procedure variables. <br />
  ///   </para>
  ///   <code lang="FPC">{$modeswitch nestedprocvars}</code>
  /// </summary>
  TPromise = class( TCustomPromise )
  protected

    /// <summary>
    ///   Returns TRUE if called from within the promise thread, or else
    ///   returns FALSE. <br /><br />See the TPromise.Perform method
    ///   documentation for example of usage. <br />
    /// </summary>
    /// <returns>
    ///   Boolean indicating if this method was called from the deferred
    ///   promise thread (TRUE) or some other thread (FALSE).
    /// </returns>
    function IsPromise: boolean;

    /// <summary>
    ///   <para>
    ///     <br /><b>This method prototype differs from Delphi to FPC!</b><br /><br />
    ///     The Perform() method schedules a method call, deferring its
    ///     execution to the thread which hosts the promise. <br /><br />From
    ///     the promises sample... <br /><br />
    ///   </para>
    ///   <code lang="Delphi / FPC">function TMyPromise.Add( const A: int32; const B: int32 ): int32;
    ///
    ///   procedure Promise;
    ///   begin
    ///     Result := A + B;
    ///   end;
    ///
    /// begin
    ///   if IsPromise then begin
    ///     Promise;
    ///   end else begin
    ///     {$ifdef fpc}
    ///     Perform( Promise );
    ///     {$else}
    ///     Result := Perform( 'Add', [A,B] ).AsOrdinal;
    ///     {$endif}
    ///   end;
    /// end;</code>
    ///   <para>
    ///     <b>Example explained.. <br /></b><br />The actual functionality
    ///     of the TMyPromise.Add() method is placed into a nested procedure.
    ///     The nested procedure has access to the arguments and result of
    ///     the parent procedure, allowing for function calls of any
    ///     prototype to be handled by the nested function. <br /><br />
    ///     Within the main body, TMyPromise.Add() first calls (inherited)
    ///     IsPromise() to determine if it has been called by the promise
    ///     thread. If it has been called by the promise thread, it simply
    ///     calls to the nested procedure to perform the functionality of the
    ///     Add() method. <br /><br />If TMyPromise.Add() is called by any
    ///     thread other than the promise thread (as determined by a call to
    ///     IsPromise()), it must defer execution to the promise thread. It
    ///     does this by calling the (inherited) Perform() method. The
    ///     prototype for this method differs between Delphi and
    ///     FreePascal... <br /><br />In the case of FreePascal, the compiler
    ///     features support for passing nested procedures as references,
    ///     which it accomplishes by passing both a pointer to the procedure,
    ///     and a reference to the stack frame. We can therefore call
    ///     Perform(), passing in the nested procedure. The operating promise
    ///     thread is able to call this procedure directly. <br /><br />In
    ///     the case of Delphi, passing nested procedures by reference is not
    ///     a supported feature of the compiler. Fortunately, the RTTI
    ///     capabilities of Delphi exceed those of FreePascal and it is
    ///     possible to invoke methods via RTTI. In this case, we call the
    ///     (inherted) Perform() method, passing in the name of the method to
    ///     be executed, and the parameters as a TArray&lt;TValue&gt;, and as
    ///     a result we receive a TValue&lt;&gt; which may be converted back
    ///     to the return type. <br /><br />See Remarks for more. <br />
    ///   </para>
    ///   <br />
    /// </summary>
    /// <remarks>
    ///   <br />In the example above, the IsPromise() will never return TRUE
    ///   for freepascal. <br />The freepascal implementation of perform has
    ///   the promise thread call the nested 'promise' method directly, and
    ///   because the operating promise thread has no knowledge of the
    ///   TMyPromise.Add() method, it will never call it. <br /><br />This is
    ///   in contrast to the Delphi approach, in which the nested 'promise'
    ///   procedure cannot be called directly. Instead the operating promise
    ///   thread is instructed to make a second call to TMyPromise.Add().
    ///   Because that second call comes from within the promise thread,
    ///   IsPromise() returns true, and the nested procedure is called in this
    ///   second pass. <br /><br /><b>Why not a consistent perform() method for
    ///   both Delphi and FreePascal? <br /><br /></b> You may be tempted to
    ///   think that the obvious answer to this question is that the compilers
    ///   support different features, and ultimately this is true, but is is
    ///   not the whole story... <br /><br />FreePascal DOES support the
    ///   invoking of methods via RTTI, however, at the time of this writing
    ///   the feature is new, experimental, and is not supported for all
    ///   calling conventions or on all target platforms. If the RTTI approach
    ///   were used for the FreePascal implementation, it would severely limit
    ///   the available targets, which passing a nested reference does not. <br /><br />
    ///   Delphi CAN be made to call a nested procedure directly, however, this
    ///   would have to be done through some unsafe code. When FreePascal
    ///   creates a reference to a nested procedure, the reference includes
    ///   both a pointer to the method code, and a copy of the current
    ///   stack-frame at the time the call is made. Delphi does not support
    ///   such a feature, and so the only way to implement this would be by
    ///   passing the pointer alone, which risks corrupting the stack
    ///   (potentially the stack of both the promise thread and the calling
    ///   thread!). <br /><br /><b>Why not use anonymous methods? <br /></b><br />
    ///   At the time of this writing, FreePascal does not support anonymous
    ///   methods. I have worked around this by using overloads of global,
    ///   object, and nested methods in other code. Even if FreePascal did
    ///   support anonymous methods, the feature could not be used for promises
    ///   because the method prototype is not known to the promises thread. <br /><br />
    /// </remarks>
    {$ifdef fpc}
    procedure Perform( const Method: TPromisedMethod );
    {$else}
    function Perform(const MethodName: string; const Args: TArray<TValue> ): TValue; overload;
    function Perform(const MethodName: string ): TValue; overload;
    {$endif}
  end;

type
  ///  <summary>
  ///    A factory for creating an instance of IList<T> which uses a critical section
  ///    to defend insertions and removals from cross-thread activity.
  ///  </summary>
  TThreadSafeList<T> = record
    type TArrayOfT = array of T;
    class function Create( const Granularity: nativeuint = 32; const isOrdered: boolean = false; const isPruned: boolean = false ): IList<T>; overload; static;
    class function Create( const Items: TArrayOfT ): IList<T>; overload; static;
  end;

///  <summary>
///    Returns the global singleton instance of IThreads.
///  </summary>
function Threads: IThreads;

implementation
uses
  cwThreading.Threads.Standard
, cwThreading.SignaledCriticalSection.Standard
, cwThreading.CriticalSection.Standard
, cwThreading.ThreadSafeList.Standard
;

var
  SingletonThreads: IThreads = nil;

class function TThreadSafeList<T>.Create(const Granularity: nativeuint; const isOrdered: boolean; const isPruned: boolean): IList<T>;
begin
  Result := cwThreading.ThreadSafeList.Standard.TThreadSafeList<T>.Create( Granularity, isOrdered, isPruned );
end;

class function TThreadSafeList<T>.Create(const Items: TArrayOfT): IList<T>;
var
  idx: nativeuint;
begin
  Result := cwThreading.ThreadSafelist.Standard.TThreadSafeList<T>.Create;
  if Length( Items ) = 0 then exit;
  for idx := 0 to pred( Length( Items ) ) do Result.Add( Items[ idx ] );
end;

class function TSignaledCriticalSection.Create( const SleepTimeoutMilliseconds: uint64 = 0 ): ISignaledCriticalSection;
begin
  Result := cwThreading.SignaledCriticalSection.Standard.TSignaledCriticalSection.Create( SleepTimeoutMilliseconds );
end;

class function TCriticalSection.Create: ICriticalSection;
begin
  Result := cwThreading.CriticalSection.Standard.TCriticalSection.Create;
end;

function Threads: IThreads;
begin
  if not assigned(SingletonThreads) then begin
    SingletonThreads := cwThreading.Threads.Standard.TThreads.Create;
  end;
  Result := SingletonThreads;
end;

function TPromise.IsPromise: boolean;
begin
  Result := DoIsPromise;
end;

{$ifdef fpc}

procedure TPromise.Perform( const Method: TPromisedMethod );
begin
  DoPerform( Method );
end;

{$else}

function TPromise.Perform( const MethodName: string; const Args: TArray<TValue> ): TValue;
begin
  Result := DoPerform( MethodName, Args );
end;

function TPromise.Perform(const MethodName: string ): TValue;
var
  EmptyArray: TArray<TValue>; // Delphi XE6 or older requires the array declared.
begin
  Result := DoPerform( MethodName, EmptyArray );
end;


{$endif}

initialization
  SingletonThreads := nil;

finalization
  SingletonThreads := nil;

end.

