﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwThreading;
{$ifdef fpc}
  {$mode delphiunicode}
  {$MODESWITCH NESTEDPROCVARS}
  {$warn 05079 off}
{$endif}

interface
uses
  cwTypes
, cwCollections
{$ifndef FPC}
, typinfo
, rtti
{$endif}
;

const
  stThreadSleepFailed                : TGUID = '{89368E3D-D247-4A10-AF0A-9CDC68E71596}';
  stThreadCreateFailed               : TGUID = '{A57FE90F-E152-47F9-80A1-0C14AB81706B}';
  stPromiseNotRunningInThread        : TGUID = '{EE30DB47-C3CA-457E-BB33-B29C12E44AC2}';
  stPromiseActionFailed              : TGUID = '{C726E6EE-9E3D-4981-9ED7-6B9C1F4CC561}';
  stObjectMustSupportInterface       : TGUID = '{D9383E5A-4B4B-4D09-9855-9941D1D2414A}';
  stObjectMustSupportEitherInterface : TGUID = '{BA27DC5B-EFFC-418F-B3E2-9A37CB48DC5F}';
  stObjectMustDescend                : TGUID = '{D49A38FC-DC71-4A90-A9FF-10F681F89EF2}';
  stNoTaskThreads                    : TGUID = '{7FC07E40-15B5-4CA4-8DC1-3872EA417973}';
  stIncorrectNumberOfParameters      : TGUID = '{0C8A7B14-8845-4C2A-A688-941D01990201}';
  stInitMutexFailed                  : TGUID = '{AA54D838-4FD9-460B-BBD3-83D55E2956C0}';
  stInitConditionVariableFailed      : TGUID = '{5726FC66-A81C-4398-A44D-ACB6FF94BA1B}';
  stThreadAttrInitFailed             : TGUID = '{D096FDC2-2A2F-48E1-8403-2125A8619EC2}';
  stThreadStackSetFailed             : TGUID = '{66AF3455-42F6-4370-91E8-1DC19358DD54}';
  stThreadAttrFnitFailed             : TGUID = '{ECDF9AB8-E6D4-4617-B483-C306209F7E23}';
  stThreadJoinFailed                 : TGUID = '{2677B0F3-B413-4760-9303-780D0D7DAFC9}';


type
  ///  <summary>
  ///    TThreadMethod is a method call-back which, once installed by
  ///    IThreads.ThreadMethod() will be called repeatedly by the
  ///    owning thread so long as it returns TRUE, or until the
  ///    owning thread is terminated using its IThreadControl reference. <br/>
  ///  </summary>
  {$ifdef fpc}
  TThreadMethod = function: boolean of object;
  {$else}
  TThreadMethod = reference to function: boolean;
  {$endif}

type
  ///  <summary>
  ///    Represents a mutex lock which may be used to protect a critical
  ///    section of code, which must be executed by only one thread at any
  ///    time.
  ///  </summary>
  ICriticalSection = interface
    ['{21F4E11C-C165-4473-82C0-1674EBD90678}']

    ///  <summary>
    ///    Acquire the mutex lock. A thread should call this to ensure that
    ///    it is executing exclusively.
    ///  </summary>
    procedure Acquire;

    ///  <summary>
    ///    Release the mutex lock. A thread calls this method to release it's
    ///    exclusive execution.
    ///  </summary>
    procedure Release;
  end;

type
  ///  <summary>
  ///    Represents a critical section controlled by a condition variable. <br/>
  ///    This works in the same way as an ICriticalSection, except that a
  ///    thread can put it's self to sleep (releasing the mutex), until it
  ///    is woken by an external signal from another thread. Once woken the
  ///    thread re-aquires the mutex lock and continues execution.
  ///  </summary>
  ISignaledCriticalSection = interface
    ['{89D86C88-78BB-4FD5-AE68-BFF81C035BF0}']

    ///  <summary>
    ///    Acquire the mutex lock. A thread should call this to ensure that
    ///    it is executing exclusively.
    ///  </summary>
    procedure Acquire;

    ///  <summary>
    ///    Release the mutex lock. A thread calls this method to release it's
    ///    exclusive execution.
    ///  </summary>
    procedure Release;

    ///  <summary>
    ///    Causes the calling thread to release the mutex lock and begin
    ///    sleeping. While sleeping, the calling thread is excluded from the
    ///    thread scheduler, allowing other threads to consume it's runtime.
    ///    <remarks>
    ///      Sleep may return at any time, regardless of the work having been
    ///      completed. You should check that the work has actually been
    ///      completed, and if not, put the signaled critical seciton back
    ///      to sleep.
    ///    </remarks>
    ///  </summary>
    procedure Sleep;

    ///  <summary>
    ///    Called by some external thread, Wake causes the sleeping thread to
    ///    re-aquire the mutex lock and to continue executing.
    ///  </summary>
    procedure Wake;
  end;

type
  ///  <summary>
  ///    IThreadControl is passed into IThreadDelegate.Execute()
  ///    to provide the thread delegate with some control over the
  ///    thread which executes it. <br/>
  ///  </summary>
  IThreadControl = interface
    ['{6A0371DC-711E-40C8-BA9A-EA13B367F41F}']

    ///  <summary>
    ///    Removes the thread from the operating system
    ///    scheduler. Execution will not resume until
    ///    Wake() is called.
    ///  </summary>
    procedure Sleep;

    ///  <summary>
    ///    Wakes a sleeping thread. <br/>
    ///    Has no effect on an already sleeping thread.
    ///  </summary>
    procedure Wake;

    ///  <summary>
    ///    Signals the thread to terminate. <br/>
    ///    The thread will not terminate immediately,
    ///    but a flag will be set to tell it to terminate
    ///    within its next execution cycle. The thread
    ///    delegate execute method will not be called
    ///    again after calling terminate.
    ///  </summary>
    procedure Terminate;
  end;

type
  ///  <summary>
  ///    Provide an implementation of IThreadDelegate to the
  ///    thread system to have it's execute method scheduled. <br/>
  ///    A new thread is created for each instance of IThreadDelegate
  ///    which is passed to the threading system. <br/>
  ///    Your thread delegate may also implement the ILifeCycleDelegate
  ///    and/or IPauseResumeDelegate, see cwDelegates for more information.
  ///  </summary>
  IThreadDelegate = interface
    ['{EA9CEB60-66E5-469B-BFF3-3574AF1FBBA7}']

    ///  <summary>
    ///    Your execute method will be called repeatedly
    ///    by the thread which is created for it. <br/>
    ///    You may use the Sleep/Wake methods of the
    ///    control to pause and resume calls to execute
    ///    respectively.<br/>
    ///    A call to the terminate method of the control
    ///    will cause the thread to terminate its self,
    ///    in which case, this execute method is no longer
    ///    called.
    ///  </summary>
    procedure Execute( const Control: IThreadControl );

  end;

type
  ///  <summary>
  ///    Provide an implementation of ITaskDelegate to be
  ///    performed by a task thread when calling
  ///    IThreads.Tasks().
  ///  </summary>
  ITaskDelegate = interface
    ['{593ABA40-23EB-4299-87BE-6A877B2925EA}']

    ///  <summary>
    ///    A task thread will call this execute
    ///    method when scheduled to perform
    ///    the task.
    ///  </summary>
    procedure Execute;
  end;

  ///  <summary>
  ///    Provide an implementation of IScheduledTaskDelegate to
  ///    be performed on a regular interval by an ITaskSchedule
  ///    driven thread.
  ///  </summary>
  IScheduledTaskDelegate = interface
    ['{4988EDA2-A920-4ED0-84D1-807358F651F6}']

    ///  <summary>
    ///    Your execute method will be called with ElapsedMilliseconds
    ///    set to the time elapsed since this method was previously
    ///    called. (DeltaT in milliseconds)
    ///  </summary>
    procedure Execute( const ElapsedMilliseconds: nativeuint );
  end;

type
  ///  <summary>
  ///    Provide an implementation of IForTaskDelegate to be
  ///    performed by a task thread when calling IThreads.ForLoop();
  ///  </summary>
  IForTaskDelegate = interface
    ['{BEB5CC41-6515-4498-97C0-A4B7D6B4518C}']

    ///  <sumamry>
    ///    Your IForTaskDelegate will be assigned a portion of the total
    ///    work to be executed by the task threads. This portion of work
    ///    is described by the provided start and stop indicies. <br/>
    ///  </summary>
    procedure Execute( const start: nativeuint; const stop: nativeuint );
  end;

type
  ///  <summary>
  ///    An IThreadingSystem is a scheduler for marshalling both long and
  ///    short running threads. <br/>
  ///    cwThreading provides a singleton instance of IThreadingSystem which
  ///    is returned from cwThreading.Standard.ThreadingSystem. <br/>
  ///  </summary>
  IThreads = interface
    ['{FE38A927-61DA-4069-B26F-5BA3DE100266}']

    ///  <summary>
    ///    Returns the number of threads allocated for performing tasks. <br/>
    ///    Getter for 'TaskThreadCount' property.
    ///  </summary>
    function getTaskThreadCount: nativeuint;

    ///  <summary>
    ///    Sets the number of threads to be allocated for performing tasks. <br/>
    ///    CAUTION: Setting this to zero while there are tasks still waiting
    ///    to be performed will cause those tasks to halt. <br/>
    ///    Setter for 'TaskThreadCount' property.
    ///  </summary>
    procedure setTaskThreadCount( const value: nativeuint );

    /// <summary>
    ///   ExecuteThread() creates a new thread which is responsible for running
    ///   the provided ThreadDelegate. <br /><br />Your implementation of
    ///   IThreadDelegate.Execute() will be called repeatedly until the thread
    ///   sleeps or terminates. <br /><br />Your implementation of
    ///   IThreadDelegate may also implement cwDelegates.ILifeCycleDelegate
    ///   and/or cwDelegates.IPauseResumeDelegate in order to respond to thread
    ///   life-cycle events. <br /><br />See Remarks.
    /// </summary>
    /// <param name="ThreadDelegate">
    ///   An implementation of IThreadDelegate to be operated by the newly
    ///   created thread.
    /// </param>
    /// <returns>
    ///   A reference to the delegate passesd in as ThreadDelegate.
    /// </returns>
    function ExecuteThread( const ThreadDelegate: TInterfacedObject ): IThreadDelegate;

    ///  <summary>
    ///    ThreadedMethod() will launch a new thread to repeatedly call
    ///    the prrovided 'ThreadMethod' callback until such time that
    ///    the callback function returns FALSE, or until the thread is
    ///    terminated using an IThreadControl() on its IThreadDelegate. <br/>
    ///    See also: ControlThread().
    ///  </summary>
    function ThreadedMethod( const ThreadMethod: TThreadMethod ): IThreadDelegate;

    /// <summary>
    ///   Given a thread delegate, ControlThread looks up the operating thread
    ///   for that delegate and returns an IThreadControl reference. This
    ///   reference may be used to Sleep, Wake or Terminate the thread which is
    ///   operating the delegate. <br /><br />Note: You should not retain this
    ///   reference as this can lead to "zombie" threads - i.e. the thread
    ///   class remains allocated after the thread has terminated. <br />
    /// </summary>
    function ControlThread( const ThreadDelegate: IThreadDelegate ): IThreadControl;

    /// <summary>
    ///   Creates a new thread to run a promise class. <br /><br />A promise
    ///   class is one which is able to defer execution of some, or all of its
    ///   methods to an operating thread. Making those methods safe to call
    ///   from external threads. <br /><br />Conditions for running a promise
    ///   class. <br /><br />1) Your promise class must descend from
    ///   cwThreading.Factory.TPromise. <br /><br />2) Your promise class must
    ///   implement an interface. This interface may be any interface of your
    ///   choosing, and is recommended to contain the methods to be deferred to
    ///   the operating thread. <br /><br />3) Deferred methods must be written
    ///   such that they operate to perform the method action when called by
    ///   the operating thread, but to defer the action when called by any
    ///   other thread. This is facilitated by the inherited protected methods
    ///   of TPromise. <br /><br />See cwThreading.Factory.TPromise and the
    ///   Promises sample application for more information. <br />
    /// </summary>
    /// <param name="Promise">
    ///   An instance which descends TPromise.
    /// </param>
    /// <returns>
    ///   An interface reference for the promise class which may be cast to the
    ///   interface of your choosing (under condition that the class implements
    ///   it).
    /// </returns>
    function ExecutePromise( const Promise: TInterfacedObject ): IInterface;

    ///  <summary>
    ///    Looks up executing promises by interface and returns a list
    ///    containing the promises found.
    ///  </summary>
    function ResolvePromise( const Iface: TGUID ): IReadOnlyList<IInterface>;

    /// <summary>
    ///   Given an interface to a promise, this causes the promise to be
    ///   removed from the thread which is operating it. <br /><br />Because
    ///   promises are able to create new 'sub-promises' on the same thread, a
    ///   single thread may be operating multiple promises. When all promises
    ///   on a thread have been terminated via a call to TerminatePromise(),
    ///   the thread will terminate its self.
    /// </summary>
    procedure TerminatePromise( const Promise: IInterface );

    /// <summary>
    ///   <para>
    ///     Tasks() schedules the tasks supplied in the Tasks
    ///     parameter to be performed by a pool of task threads. There must
    ///     be at least one task thread allocated to perform tasks, see the
    ///     'TaskThreads' property.
    ///   </para>
    ///   <para>
    ///     <br />Tasks() will return immediately after scheduling the
    ///     tasks. You may optionally receive a notification when all of the
    ///     scheduled task have been performed, by providing a call back for
    ///     the NotifyDone parameter.
    ///   </para>
    ///   <para>
    ///     The NotifyDone parameter has a simple prototype of: <br />
    ///     procedure (); <br /><br />You may provide a global procedure,
    ///     object method, anonymous method (Delphi only) or a nested method
    ///     (FPC only). <br /><br />
    ///   </para>
    /// </summary>
    /// <param name="Tasks">
    ///   An IReadOnlyList&lt;&gt; or IList&lt;&gt; of ITaskDelegate instances
    ///   to be performed by the task threads.
    /// </param>
    /// <param name="NotifyDone">
    ///   A callback to be called when all of the tasks have been completed by
    ///   task threads.
    /// </param>
    procedure Tasks( const Tasks: IReadOnlyList<ITaskDelegate>; const NotifyDone: TNotifyDelegate );

    /// <summary>
    ///   <para>
    ///     ForLoop() is essentially a parallel 'for-loop' for un-ordered
    ///     processing. At least one task thread must be allocated. See <see cref="IThreads.TaskThreadCount">
    ///     IThreads.TaskThreadCount.</see><br /><br />A call to ForLoop()
    ///     causes the 'TotalWork' to be divided among the available task
    ///     threads. <br />Each task thread then does one of two things: <br /><br />
    ///     1) Call the provided Task method, supplying the index for each
    ///     work item. <br /><br />2) Call the provided Task method,
    ///     supplying the start and stop indices of the work items that it is
    ///     responsible for. <br /><br />The determination of which action is
    ///     taken is made by the task method provided. You may supply one of
    ///     two task method prototypes... <br /><br />
    ///   </para>
    ///   <code lang="Delphi / FPC">procedure ( const index: nativeuint; const Parameters: array of TVariant = [] );
    /// // OR
    /// procedure ( const start: nativeuint; const stop: nativeuint; const Parameters: array of TVariant = [] );</code>
    ///   <para>
    ///     <br />Overloads allow for the provided task method to be a global
    ///     procedure or object method, as well as an anonymous method
    ///     (Delphi only), or nested method (FPC only). The task method may
    ///     also be provided with an optional array of variants as
    ///     parameters. <br /><br />You may optionally provide a notification
    ///     method to be called when the for loop is complete. This method
    ///     may also be a global procedure, object method, Anonymous method
    ///     (Delphi only), or Nested method (FPC only).
    ///   </para>
    ///   <para>
    ///     If a notification method is not provided, you may optionally
    ///     specify that the for loop call should wait for the work to be
    ///     completed.
    ///   </para>
    /// </summary>
    /// <param name="TotalWork">
    ///   Specify the total number of iterations to be processed by the loop.
    /// </param>
    /// <param name="Task">
    ///   A method typed as one of the prototypes defined above, to perform the
    ///   work on a per-item, or per work division basis.
    /// </param>
    /// <param name="Parameters">
    ///   (Optional) Array of variant parameters to be passed into the task
    ///   method.
    /// </param>
    /// <param name="NotifyDone">
    ///   (Optional) Method to be called when the work has been completed.
    /// </param>
    /// <param name="WaitFor">
    ///   (Optional) Boolean, if set TRUE a call to ForLoop() will block until
    ///   all work has been completed.
    /// </param>
    procedure ForLoop( const TotalWork: nativeuint; const Task: TLoopDelegate; const Parameters: array of TVariant; const NotifyDone: TNotifyDelegate ); overload;
    /// <exclude/>
    procedure ForLoop( const TotalWork: nativeuint; const Task: TLoopDelegate; const Parameters: array of TVariant; const Wait: boolean = FALSE ); overload;

    /// <summary>
    ///   Adds either an implementation of ITaskDelegate, or an implementation of IScheduledTaskDelegate
    ///   to the task scheduler, to be executed at a defined interval.
    /// </summary>
    procedure Schedule( const ScheduledTaskDelegate: TInterfacedObject; const IntervalMilliseconds: nativeuint );

    ///  <summary>
    ///    Returns or Sets the number of threads to be allocated for performing tasks. <br/>
    ///    CAUTION: Setting this to zero while there are tasks still waiting
    ///    to be performed will cause those tasks to halt. <br/>
    ///  </summary>
    property TaskThreadCount: nativeuint read getTaskThreadCount write setTaskThreadCount;

  end;

///  <summary>
///    A writeln protected by a critical section to
///    ensure that a call to writeln is completed
///    before another is made.
///  </summary>
procedure AtomicWriteln( const msg: string );

///  <summary>
///    Removes the calling thread from the OS scheduler for
///    a specified number of milliseconds. <br/>
///    Note: Millisecond timing is imprecise due to the
///    instancing and disposal of a signaled critical section.
///    This make sleep sufficient for interactive application but
///    not for precise timing scenarios. <br/>
///    For more precise timing, see IThreads.Schedule and cwTiming.
///  </summary>
procedure Sleep( const ms: nativeuint );

///  <summary>
///    Returns the ID of the calling thread.
///  </summary>
function GetCurrentThreadID: TThreadID;

///  <summary>
///    Returns the ID of the process main thread.
///  </summary>
function GetMainthreadID: TThreadID;

implementation
uses
  cwThreading.CriticalSection.Standard
, cwThreading.SignaledCriticalSection.Standard
;

var
  Atom: nativeuint = 0;
  MainThreadID: TThreadID = 0;

procedure AtomicWriteln(const msg: string);
var
  aThreadID: TThreadID;
begin
  aThreadID := GetCurrentThreadID;
  while TRUE do begin
    while Atom <> 0 do;
    Atom := aThreadID;
    if Atom = aThreadID then break;
  end;
  try
    Writeln(msg);
  finally
    Atom := 0;
  end;
end;

procedure Sleep( const ms: nativeuint );
var
  SleepCS: ISignaledCriticalSection;
begin
  SleepCS := TSignaledCriticalSection.Create( ms );
  try
    SleepCS.Acquire;
    try
      SleepCS.Sleep;
    finally
      SleepCS.Release;
    end;
  finally
    SleepCS := nil;
  end;
end;

function GetCurrentThreadID: TThreadID;
begin
  Result := cwTypes.getCurrentThreadID;
end;

function GetMainthreadID: TThreadID;
begin
  Result := MainThreadID;
end;

initialization
  MainThreadID := cwTypes.getCurrentThreadID;

  TStatus( stThreadSleepFailed                ).Register('Signaled Critical Section failed to sleep on condition variable.');
  TStatus( stThreadCreateFailed               ).Register('Attempt to create new thread failed.');
  TStatus( stPromiseNotRunningInThread        ).Register('An attempt to call a method on promise class "(%promiseclass%)" failed because the class is not hosted by a thread.');
  {$ifdef fpc}
  TStatus( stPromiseActionFailed              ).Register('Failed to execute a promise method on "(%promiseclass%)"');
  {$else}
  TStatus( stPromiseActionFailed              ).Register('Failed to execute "(%promiseclass%).(%method%)" using RTTI. Is {$M+} set for the promise class?);');
  {$endif}
  TStatus( stObjectMustSupportInterface       ).Register('The object provided for "(%object%)" must support interface "(%interface%)".');
  TStatus( stObjectMustSupportEitherInterface ).Register('The object provided for "(%object%)" must support either interface "(%interface1%)" or interface "(%interface2%)".');
  TStatus( stObjectMustDescend                ).Register('The object provided for "(%object%)" must descend "(%baseclass)".');
  TStatus( stNoTaskThreads                    ).Register('The requested operation requires that there be task threads allocated.');
  TStatus( stIncorrectNumberOfParameters      ).Register('Incorrect number of parameters provided for call to method "(%method%)"' );
  TStatus( stInitMutexFailed                  ).Register('Failed to initialize mutex.');
  TStatus( stInitConditionVariableFailed      ).Register('Failed to initialize condition variable.');
  TStatus( stThreadAttrInitFailed             ).Register('pthread_attr_init() failed with code (%retcode%).');
  TStatus( stThreadStackSetFailed             ).Register('pthread_attr_setstacksize() failed with code (%retcode%).');
  TStatus( stThreadAttrFnitFailed             ).Register('pthread_attr_destroy() failed with code (%retcode%).');
  TStatus( stThreadJoinFailed                ).Register('pthread_join() failed with code (%retcode%).');

end.

