﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwCollections.Dictionary.Standard;
{$ifdef fpc}
  {$mode delphiunicode}
  {$MODESWITCH ADVANCEDRECORDS}
  {$MODESWITCH TYPEHELPERS}
  {$MODESWITCH NESTEDPROCVARS}
  {$ifdef CPU64}{$define CPU64BITS}{$endif}
  {$MODESWITCH ARRAYOPERATORS-}
{$endif}

interface
uses
  cwTypes
, cwCollections
;

type
  TStandardDictionary<K,V> = class( TInterfacedObject, IReadOnlyDictionary<K,V>, IDictionary<K,V> )
  private
    fKeys: array of K;
    fItems: array of V;
    fCapacity: nativeuint;
    fCount: nativeuint;
    fGranularity: nativeuint;
    fPruned: boolean;
    fOrdered: boolean;
  strict private //- IEnuerable -//
    function GetEnumerator: IEnumerator<K>;
  private
    function OrderedRemoveItem( const idx: nativeuint ): boolean;
    function UnorderedRemoveItem( const idx: nativeuint ): boolean;
    procedure PruneCapacity;
    procedure Initialize(const Granularity: nativeuint = 32; const isOrdered: boolean = false; const isPruned: boolean = false);
  private //- IReadOnlyDictionary<K,V> -//
    function getCount: nativeuint;
    function getKeyByIndex( const idx: nativeuint ): K;
    function getValueByIndex( const idx: nativeuint ): V;
    function getKeyExists( const key: K ): boolean;
    function getValueByKey( const key: K ): V;
    procedure setValueByIndex( const idx: nativeuint; const value: V );
    function getAsReadOnly: IReadOnlyDictionary<K,V>;
    {$ifndef fpc}
    function Compare(const ItemA, ItemB: K): boolean;
    {$endif}
  strict private //- IDictionary<K,V> -//
    procedure setValueByKey( const key: K; const value: V );
    procedure removeByIndex( const idx: nativeuint );
    procedure Clear;
  public
    constructor Create( const Granularity: nativeuint = 32; const isOrdered: boolean = false; const isPruned: boolean = false ); reintroduce; overload;
    destructor Destroy; override;
  end;

implementation
uses
  cwCollections.Enumerator.Standard
;

procedure TStandardDictionary<K,V>.Initialize(const Granularity: nativeuint = 32; const isOrdered: boolean = false; const isPruned: boolean = false);
begin
  // Set granularity control
  if Granularity>0 then begin
    fGranularity := Granularity;
  end;
  // Set granularity pruning.
  fPruned := isPruned;
  // Set order maintenance flag
  fOrdered := isOrdered;
  // Initialize the array.
  fCount := 0;
  fCapacity := 0;
  SetLength( fKeys, fCapacity );
  SetLength( fItems, fCapacity );
end;

constructor TStandardDictionary<K,V>.Create( const Granularity: nativeuint = 32; const isOrdered: boolean = false; const isPruned: boolean = false );
begin
  inherited Create;
  Initialize(Granularity,isOrdered,isPruned);
end;

destructor TStandardDictionary<K,V>.Destroy;
begin
  SetLength( fKeys, 0 );
  SetLength( fItems, 0 );
  inherited Destroy;
end;

function TStandardDictionary<K,V>.getKeyByIndex( const idx: nativeuint ): K;
begin
  {$warnings off} Result := Default(K); {$warnings on}
  if idx<getCount then begin
    Result := fKeys[idx];
  end;
end;

{$ifndef fpc}
(* EXPERIMENTAL *)
(*
   As Delphi will not compare two items of the same type when generic,
   we'll compare them byte-for-byte to save having to have a comparison
   anonymous method. This -should- work for any type, including
   referenced types as the references should match if the items
   are equal.
2*)

function TStandardDictionary<K,V>.Compare( const ItemA: K; const ItemB: K ): boolean;
var
  idx: nativeuint;
  ptrA: ^uint8;
  ptrB: ^uint8;
begin
  Result := False;
  if sizeof(K)=0 then exit;
  ptrA := @ItemA;
  ptrB := @ItemB;
  idx := sizeof(K);
  while idx>0 do begin
    if ptrA^<>ptrB^ then exit;
    inc(ptrA);
    inc(ptrB);
    dec(idx);
  end;
  Result := True;
end;
{$endif}

function TStandardDictionary<K,V>.getKeyExists( const key: K ): boolean;
var
  idx: nativeuint;
begin
  Result := False;
  if getCount=0 then begin
    exit;
  end;
  for idx := 0 to pred(getCount) do begin
    {$ifdef fpc}
    if fKeys[idx] = key then begin
    {$else}
    if Compare( fKeys[idx] , key ) then begin
    {$endif}
      Result := True;
      Exit;
    end;
  end;
end;

function TStandardDictionary<K,V>.getValueByIndex( const idx: nativeuint ): V;
begin
  Result := Default(V);
  if idx<getCount then begin
    Result := fItems[idx];
  end;
end;

procedure TStandardDictionary<K,V>.setValueByIndex( const idx: nativeuint; const value: V );
begin
  fItems[idx] := value;
end;

function TStandardDictionary<K,V>.getValueByKey( const key: K ): V;
var
  idx: nativeuint;
begin
  Result := Default(V);
  if getCount=0 then begin
    exit;
  end;
  for idx := 0 to pred(getCount) do begin
    {$ifdef fpc}
    if fKeys[idx] = key then begin
    {$else}
    if Compare( fKeys[idx] , key ) then begin
    {$endif}
      Result := fItems[idx];
      Exit;
    end;
  end;
end;

function TStandardDictionary<K,V>.GetEnumerator: IEnumerator<K>;
begin
  Result := TEnumerator<K>.Create( @fItems, fCount );
end;

function TStandardDictionary<K,V>.OrderedRemoveItem( const idx: nativeuint ): boolean;
var
  idy: nativeuint;
begin
  Result := False; // unless..
  if fCount=0 then begin
    exit;
  end;
  if idx<pred(fCount) then begin
    for idy := idx to pred(pred(fCount)) do begin
      fItems[idy] := fItems[succ(idy)];
      fKeys[idy] := fKeys[succ(idy)];
    end;
    fItems[pred(fCount)] := Default(V);
    fKeys[pred(fCount)] := Default(K);
    dec(fCount);
    Result := True;
  end else if idx=pred(fCount) then begin
    //- Item is last on list, no need to move-down items above it.
    fItems[idx] := Default(V);
    fKeys[idx] := Default(K);
    dec(fCount);
    Result := True;
  end;
end;

function TStandardDictionary<K,V>.UnorderedRemoveItem( const idx: nativeuint ): boolean;
begin
  Result := False; // unless..
  if fCount=0 then begin
    exit;
  end;
  if idx<pred(fCount) then begin
    //- Move last item into place of that being removed.
    fItems[idx] := fItems[pred(fCount)];
    fKeys[idx] := fKeys[pred(fCount)];
    //- Clear last item
    fItems[pred(fCount)] := Default(V);
    {$warnings off} fKeys[pred(fCount)] := Default(K); {$warnings on}
    dec(fCount);
    Result := True;
  end else if idx=pred(fCount) then begin
    //- if idx=fCount then simply remove the top item and decrement
    fItems[idx] := Default(V);
    {$warnings off} fKeys[idx] := Default(K); {$warnings on}
    dec(fCount);
    Result := True;
  end;
end;

procedure TStandardDictionary<K,V>.PruneCapacity;
var
  Blocks: nativeuint;
  Remainder: nativeuint;
  TargetSize: nativeuint;
begin
  Blocks := fCount div fGranularity;
  Remainder := fCount - Blocks;
  if Remainder>0 then begin
    inc(Blocks);
  end;
  TargetSize := Blocks*fGranularity;
  //- Total number of required blocks has been determined.
  if fCapacity>TargetSize then begin
    fCapacity := TargetSize;
    SetLength( fItems, fCapacity );
    SetLength( fKeys, fCapacity );
  end;
end;

procedure TStandardDictionary<K,V>.removeByIndex( const idx: nativeuint );
begin
  // If the list is ordered, perform slow removal, else fast removal
  if fOrdered then begin
    OrderedRemoveItem( idx );
  end else begin
    UnorderedRemoveItem( idx );
  end;
  // If the list is pruning memory (to save memory space), do the prune.
  if fPruned then begin
    PruneCapacity;
  end;
end;

procedure TStandardDictionary<K,V>.Clear;
begin
  fCount := 0;
  if fPruned then begin
    fCapacity := 0;
    SetLength( fKeys, fCapacity );
    SetLength( fItems, fCapacity );
  end;
end;

procedure TStandardDictionary<K,V>.setValueByKey( const key: K; const value: V );
var
  idx: nativeuint;
begin
  if getCount>0 then begin //- Craig! Don't change this!
    for idx := pred(getCount) downto 0 do begin
      {$ifdef fpc}
      if fKeys[idx] = key then begin
      {$else}
      if Compare( fKeys[idx] , key ) then begin
      {$endif}
        fItems[idx] := value;
        exit;
      end;
    end;
  end;
  //- If we made it here, add the item.
  if (fCount=fCapacity) then begin
    fCapacity := fCapacity + fGranularity;
    SetLength(fKeys,fCapacity);
    SetLength(fItems,fCapacity);
  end;
  fKeys[fCount] := key;
  fItems[fCount] := value;
  inc(fCount);
end;

function TStandardDictionary<K,V>.getAsReadOnly: IReadOnlyDictionary<K,V>;
begin
  Result := Self as IReadOnlyDictionary<K,V>;
end;

function TStandardDictionary<K, V>.getCount: nativeuint;
begin
  Result := fCount;
end;

end.

