﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwLog.Log.Standard;
{$ifdef fpc}{$mode delphiunicode}{$endif}

interface
uses
  cwCollections
, cwThreading
, cwTypes
, cwLog
;

type
  TLog = class( TInterfacedObject, ILog )
  private
    fLogLevel: TLogSeverity;
    fTargets: IList<ILogTarget>;
    fTargetsCS: ICriticalSection;
    fDateTimeFormat: string;
  private
    function getSeverityPadded(const Severity: TLogSeverity): string;
  strict private //- ILog -//
    function  getDateTimeFormat: string;
    procedure SetDateTimeFormat( const Format: string );
    function  GetTargets: IReadOnlyList<ILogTarget>;
    function  GetLogLevel: TLogSeverity;
    procedure SetLogLevel( const Value: TLogSeverity );
    function  Insert( const Status: TStatus; const Severity: TLogSeverity; const Additional: string; const Parameters: TArrayOfString ): TStatus; overload;
    function  Insert( const Status: TStatus; const Severity: TLogSeverity; const Parameters: TArrayOfString ): TStatus; overload;
    function  Insert( const Status: TStatus; const Severity: TLogSeverity ): TStatus; overload;
    procedure ClearLogTargets;
    procedure AddTarget( const Target: ILogTarget );
  public
    constructor Create;
    destructor Destroy; override;
  end;

implementation
uses
  cwCollections.Factory
, cwThreading.Factory
;

const
  cDefaultDateTimeFormat = 'YYYY-MM-DD HH:nn:SS:zzzz';

function TLog.GetTargets: IReadOnlyList<ILogTarget>;
begin
  Result := fTargets;
end;

function TLog.GetLogLevel: TLogSeverity;
begin
  Result := fLogLevel;
end;

procedure TLog.SetLogLevel(const Value: TLogSeverity);
begin
  fLogLevel := value;
end;


function TLog.Insert(const Status: TStatus; const Severity: TLogSeverity; const Additional: string; const Parameters: TArrayOfString): TStatus;
const
  cFormatMessage = '[%s]: %s%s' + CR + LF;

var
  idx: nativeuint;
  Translated: string;
  AdditionalStr: string;
begin
  Result := Status;
  if Status then exit; // Don't insert success status.
  if fLogLevel = lsNone then exit;
  if Severity = lsNone then exit;
  if ord( Severity ) < ord( fLogLevel ) then exit;
  AdditionalStr := '';
  if (Additional.Trim<>'') then begin
    AdditionalStr := Additional.Trim+' ';
  end;
  Translated := cFormatMessage.Format(
    [
      getSeverityPadded( Severity ),
      AdditionalStr,
      Status.AsString( Parameters )
    ]
  );
  fTargetsCS.Acquire;
  try
    if fTargets.Count=0 then exit;
    for idx := 0 to pred( fTargets.Count ) do begin
      fTargets[ idx ].Insert( Translated, Status, Severity, Parameters );
    end;
  finally
    fTargetsCS.Release;
  end;
end;

function TLog.getSeverityPadded( const Severity: TLogSeverity ): string;
const
  cSeverityDebug   = ' DEBUG   ';
  cSeverityVerbose = ' VERBOSE ';
  cSeverityInfo    = ' INFO    ';
  cSeverityHint    = ' HINT    ';
  cSeverityWarning = ' WARNING ';
  cSeverityError   = ' ERROR   ';
  cSeverityFatal   = ' FATAL   ';
  cUnknown         = ' ??????? ';
begin
  case Severity of
    lsDebug:    Result := cSeverityDebug;
    lsVerbose:  Result := cSeverityVerbose;
    lsInfo:     Result := cSeverityInfo;
    lsHint:     Result := cSeverityHint;
    lsWarning:  Result := cSeverityWarning;
    lsError:    Result := cSeverityError;
    lsFatal:    Result := cSeverityFatal;
       else     Result := cUnknown;
  end;
end;

function TLog.getDateTimeFormat: string;
begin
  Result := fDateTimeFormat;
end;

procedure TLog.SetDateTimeFormat(const Format: string);
begin
  fTargetsCS.Acquire; // prevent changing while log entry being translated.
  try
    fDateTimeFormat := Format;
  finally
    fTargetsCS.Release;
  end;
end;

function TLog.Insert(const Status: TStatus; const Severity: TLogSeverity; const Parameters: TArrayOfString): TStatus;
begin
  Result := Insert( Status, Severity, '', Parameters );
end;

function TLog.Insert(const Status: TStatus; const Severity: TLogSeverity): TStatus;
const
  nullarray: TArrayOfString = nil;
begin
  Result := Insert( Status, Severity, nullarray );
end;

procedure TLog.ClearLogTargets;
begin
  fTargetsCS.Acquire;
  try
    fTargets.Clear;
  finally
    fTargetsCS.Release;
  end;
end;

procedure TLog.AddTarget(const Target: ILogTarget);
begin
  fTargetsCS.Acquire;
  try
    fTargets.Add(Target);
  finally
    fTargetsCS.Release;
  end;
end;

constructor TLog.Create;
begin
  inherited Create;
  fDateTimeFormat := cDefaultDateTimeFormat;
  fLogLevel       := TLogSeverity.lsInfo;
  fTargets        := TList<ILogTarget>.Create;
  fTargetsCS      := TCriticalSection.Create;
end;

destructor TLog.Destroy;
begin
  fTargetsCS := nil;
  fTargets := nil;
  inherited Destroy;
end;

end.

