
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
/// <summary>
///   The standard implementation of cwTest.
/// </summary>
unit cwTest.TestEditor.Standard;
{$ifdef fpc}{$mode delphiunicode}{$endif}

interface
uses
  cwTest
;


type
  TTestEditor = class
  private
    fTestSuite: ITestSuite;
    fCases: TArrayOfRegisteredTestCase;
    fSelectedTestcase: IRegisteredTestCase;
  private
    procedure ListTestCases( const Option: string = '' );
    procedure MainMenu;
    procedure MainMenuText;
    procedure DisableALLCases;
    procedure DisplayHelp( const Command: string );
    procedure EnableAllCases;
    procedure TestCaseMenu(const TestCaseName: string);
    procedure ContinuePrompt;
    function CountCasesByStatus(const Status: boolean): nativeuint;
    procedure TestCaseMenuText;
    procedure SetEnableDisableCase(const CaseName: string; const Enabled: boolean);
    procedure ListTests(const Option: string = '');
    function CountTestsByStatus(const Enabled: boolean): nativeuint;
    function FindTestcaseByName(const Name: string): IRegisteredTestCase;
    procedure EnableAllTests;
    procedure DisableAllTests;
    procedure SetEnableDisableTest(const TestName: string; const Status: boolean);
    procedure WriteSpacer;
  public
    constructor Create( const TestSuite: ITestSuite );
    destructor Destroy; override;
  public
    procedure Run;
  end;

implementation
uses
  cwTypes
, cwTest.TestReport.Console
;

type
  TListOption =( doAll, doEnabled, doDisabled );

constructor TTestEditor.Create(const TestSuite: ITestSuite);
begin
  inherited Create;
  fTestSuite := TestSuite;
  fCases := fTestSuite.getRegisteredTestCases;
end;

destructor TTestEditor.Destroy;
begin
  fTestSuite := nil;
  inherited Destroy;
end;

function TTestEditor.CountCasesByStatus( const Status: boolean ): nativeuint;
var
  idx: Integer;
begin
  Result := 0;
  if Length(fCases)=0 then exit;
  for idx := 0 to pred(Length(fCases)) do begin
    if fCases[idx].Enabled = Status then inc(Result);
  end;
end;

procedure TTestEditor.ListTestCases( const Option: string );
  procedure WritePageTop;
  begin
    Writeln(' Idx  Enabled    Name ');
    Writeln('------------------------------------------------');
  end;
const
  cPaginationLimit = 15;
var
  idx: nativeuint;
  Pagination: nativeuint;
  Enabled: string;
  ListOption: TListOption;
begin
  if Option.UppercaseTrim='ENABLED' then begin
    ListOption := doEnabled;
  end else if Option.UppercaseTrim='DISABLED' then begin
    ListOption := doDisabled;
  end else begin
    if Option.Trim<>'' then begin
      Writeln('Unknown list option "'+Option+'"');
      ContinuePrompt;
      exit;
    end;
    ListOption := doAll;
  end;
  try
    Pagination := 0;
    WritePageTop;

    if Length(fCases)=0 then begin
      Writeln('There are no test cases registed.');
      exit;
    end;

    if (ListOption=doEnabled) and (CountCasesByStatus(TRUE)=0) then begin
      Writeln('There are no enabled test cases registered.');
      exit;
    end;

    if (ListOption=doDisabled) and (CountCasesByStatus(FALSE)=0) then begin
      Writeln('There are no disabled test cases registered.');
      exit;
    end;

    for idx := 0 to pred(Length(fCases)) do begin
      if (ListOption=doEnabled) and (not fCases[idx].Enabled) then continue;
      if (ListOption=doDisabled) and (fCases[idx].Enabled) then continue;
      Enabled := '';
      if fCases[idx].Enabled then Enabled := '[X]' else Enabled := '[ ]';
      Write( ' '+idx.AsString.LeftPad(4,'0') );
      Writeln('   '+Enabled+'      '+fCases[idx].Name);
      inc(Pagination);
      if Pagination=cPaginationLimit then begin
        Writeln;
        ContinuePrompt;
        pagination := 0;
        WritePageTop;
      end;
    end;
  finally
    Writeln;
    ContinuePrompt;
  end;
end;

function TTestEditor.CountTestsByStatus( const Enabled: boolean ): nativeuint;
var
  idx: nativeuint;
begin
  Result := 0;
  if fSelectedTestCase.Count=0 then exit;
  for idx := 0 to pred(fSelectedTestCase.count) do begin
    if fSelectedTestCase.getTest(idx).Enabled=Enabled then inc(Result);
  end;
end;

procedure TTestEditor.ListTests( const Option: string );
  procedure WritePageTop;
  begin
    Writeln(' Idx  Enabled    Name ');
    Writeln('------------------------------------------------');
  end;
const
  cPaginationLimit = 15;
var
  idx: nativeuint;
  Pagination: nativeuint;
  Enabled: string;
  ListOption: TListOption;
begin
  if Option.UppercaseTrim='ENABLED' then begin
    ListOption := doEnabled;
  end else if Option.UppercaseTrim='DISABLED' then begin
    ListOption := doDisabled;
  end else begin
    if Option.Trim<>'' then begin
      Writeln('Unknown list option "'+Option+'"');
      ContinuePrompt;
      exit;
    end;
    ListOption := doAll;
  end;
  try
    Pagination := 0;
    WritePageTop;

    if (fSelectedTestCase.Count)=0 then begin
      Writeln('There are no tests registed for test-case: '+fSelectedTestCase.Name);
      exit;
    end;

    if (ListOption=doEnabled) and (CountTestsByStatus(TRUE)=0) then begin
      Writeln('There are no enabled tests.');
      exit;
    end;

    if (ListOption=doDisabled) and (CountTestsByStatus(FALSE)=0) then begin
      Writeln('There are no disabled tests.');
      exit;
    end;

    for idx := 0 to pred(fSelectedTestCase.Count) do begin
      if (ListOption=doEnabled) and (not fSelectedTestCase.getTest(idx).Enabled) then continue;
      if (ListOption=doDisabled) and (fSelectedTestCase.getTest(idx).Enabled) then continue;
      Enabled := '';
      if fSelectedTestCase.getTest(idx).Enabled then Enabled := '[X]' else Enabled := '[ ]';
      Write( ' '+idx.AsString.LeftPad(4,'0') );
      Writeln('   '+Enabled+'      '+fSelectedTestCase.getTest(idx).Name);
      inc(Pagination);
      if Pagination=cPaginationLimit then begin
        Writeln;
        ContinuePrompt;
        pagination := 0;
        WritePageTop;
      end;
    end;
  finally
    Writeln;
    ContinuePrompt;
  end;
end;

procedure TTestEditor.MainMenuText;
begin
  Writeln;
  Writeln('cwTest: Console Test Runner.');
  Writeln('Please enter one of the following commands: ');
  Writeln;
  Writeln('HELP <command>             - Lists help information about the command.');
  Writeln('LIST (<status>)            - Lists the test cases available.');
  Writeln('SELECT <test-case>         - Selects the specfied test case.');
  Writeln('DISABLE <test-case> / ALL  - Disables ALL test cases tests.');
  Writeln('ENABLE  <test-case> / ALL  - Enables ALL test cases.');
  Writeln('RUN                        - Run the test suite.');
  Writeln('EXIT                       - Exit the application.');
  Writeln;
end;

procedure TTestEditor.ContinuePrompt;
begin
  Writeln('Press [RETURN] to continue.');
  Readln;
end;

procedure TTestEditor.DisplayHelp( const Command: string );
begin
  try
    if Command='HELP' then begin
      Writeln('');
      Writeln('This command.');
      Writeln('Enter the "HELP" command followe by the name of the command you''d like help with.');
      Writeln('For example, type "HELP LIST" to see help information for the "LIST" command.');
      Writeln('');
    end else if Command='LIST' then begin
      Writeln('');
      Writeln('Lists all of the registered test cases.');
      Writeln('The LIST command will display the test number, enabled status and name.');
      Writeln('');
    end else if Command='SELECT' then begin
      Writeln('');
      Writeln('Selects a test case in order to list, enable or disable individual tests.');
      Writeln('');
      Writeln('USAGE: SELECT MyTest');
      Writeln('');
    end else begin
      Writeln('');
      Writeln(' '+Command+' is an unknown command.');
      Writeln('');
    end;
  finally
    ContinuePrompt;
  end;
end;

procedure TTestEditor.DisableALLCases;
var
  idx: nativeuint;
begin
  if Length(fCases)=0 then begin
    Writeln('There are no test cases registered.');
    ContinuePrompt;
    exit;
  end;
  try
    for idx := 0 to pred(Length(fCases)) do begin
      fCases[idx].Enabled := False;
    end;
  finally
    Writeln('All cases disabled.');
    ContinuePrompt;
  end;
end;

procedure TTestEditor.EnableAllCases;
var
  idx: nativeuint;
begin
  if Length(fCases)=0 then begin
    Writeln('There are no test cases registered.');
    ContinuePrompt;
    exit;
  end;
  try
    for idx := 0 to pred(Length(fCases)) do begin
      fCases[idx].Enabled := True;
    end;
  finally
    Writeln('All cases enabled.');
    ContinuePrompt;
  end;
end;

procedure TTestEditor.TestCaseMenuText;
begin
  Writeln;
  Writeln('cwTest: Console Test Runner.');
  Writeln('SELECTED TEST CASE: ',fSelectedTestcase.Name);
  Writeln('Please enter one of the following commands: ');
  Writeln;
  Writeln('HELP <command>     - Lists help information about the command.');
  Writeln('LIST (<status>)    - Lists the tests in ',fSelectedTestcase.Name);
  Writeln('DISABLE <test>     - Disables the specified test.');
  Writeln('ENABLE <test>      - Enables the specified test.');
  Writeln('EXIT               - Back to main menu.');
  Writeln;
end;

function TTestEditor.FindTestcaseByName( const Name: string ): IRegisteredTestCase;
var
  idx: nativeuint;
  utName: string;
begin
  Result := nil;
  utName := Name.UppercaseTrim;
  if Length(fCases)=0 then exit;
  for idx := 0 to pred(Length(fCases)) do begin
    if fCases[idx].Name.UppercaseTrim=utName then begin
      Result := fCases[idx];
      exit;
    end;
  end;
end;

procedure TTestEditor.EnableAllTests;
var
  idx: nativeuint;
begin
  if not assigned(fSelectedTestCase) then exit;
  if fSelectedTestCase.Count=0 then exit;
  for idx := 0 to pred(fSelectedTestCase.Count) do begin
    fSelectedTestCase.getTest(idx).Enabled := TRUE;
  end;
end;

procedure TTestEditor.DisableAllTests;
var
  idx: nativeuint;
begin
  if not assigned(fSelectedTestCase) then exit;
  if fSelectedTestCase.Count=0 then exit;
  for idx := 0 to pred(fSelectedTestCase.Count) do begin
    fSelectedTestCase.getTest(idx).Enabled := FALSE;
  end;
end;

procedure TTestEditor.SetEnableDisableTest( const TestName: string; const Status: boolean );
var
  idx: nativeuint;
  utTestName: string;
begin
  if not assigned(fSelectedTestCase) then exit;
  if fSelectedTestCase.Count=0 then exit;
  utTestName := TestName.UppercaseTrim;
  for idx := 0 to pred(fSelectedTestCase.Count) do begin
    if fSelectedTestCase.getTest(idx).Name.UppercaseTrim=utTestName then begin
      fSelectedTestCase.getTest(idx).Enabled := Status;
      exit;
    end;
  end;
end;

procedure TTestEditor.TestCaseMenu(const TestCaseName: string);
var
  Option: string;
  Parts: TArrayOfString;
begin
  fSelectedTestCase := FindTestcaseByName( TestCaseName );
  if not assigned(fSelectedTestCase) then begin
    Writeln('Could not find a test case named "'+TestCaseName+'"');
    ContinuePrompt;
    exit;
  end;
  repeat
    TestCaseMenuText;
    Readln(Option);
    WriteSpacer;
    Parts := Option.Explode(' ');
    if Length(Parts)=0 then continue;
    Option := Parts[0].UppercaseTrim;
    if Option='HELP' then begin

    end else if Option='LIST' then begin
      if Length(Parts)=2 then begin
        ListTests( Parts[1] );
      end else begin
        ListTests;
      end;
    end else if Option='DISABLE' then begin
      if Length(Parts)<2 then begin
        Writeln('Missing <test-name> parameter.');
        ContinuePrompt;
        continue;
      end;
      if Parts[1].UppercaseTrim='ALL' then begin
        DisableALLTests;
      end else begin
        SetEnableDisableTest( Parts[1], FALSE );
      end;
    end else if Option='ENABLE'  then begin
      if Length(Parts)<2 then begin
        Writeln('Missing <test-name> parameter.');
        ContinuePrompt;
        continue;
      end;
      if Parts[1].UppercaseTrim='ALL' then begin
        EnableALLTests;
      end else begin
        SetEnableDisableTest( Parts[1], TRUE  );
      end;
    end else if Option='EXIT' then begin
      exit;
    end;
  until FALSE;
end;

procedure TTestEditor.SetEnableDisableCase( const CaseName: string; const Enabled: boolean );
var
  idx: nativeuint;
  Found: boolean;
  utSearch: string;
begin
  try
    if Length(fCases)=0 then begin
      Writeln('There are no test cases registered.');
      ContinuePrompt;
      exit;
    end;
    Found := False;
    utSearch := CaseName.UppercaseTrim;
    for idx := 0 to pred(Length(fCases)) do begin
      if utSearch=fCases[idx].Name.UppercaseTrim then begin
        Found := True;
        fCases[idx].Enabled := Enabled;
        if Enabled then begin
          Writeln('Test case: "'+CaseName+'" is now enabled.');
        end else begin
          Writeln('Test case: "'+CaseName+'" is now disabled.');
        end;
        break;
      end;
    end;
    if not Found then begin
       Writeln('Test case: "'+CaseName+'" not found.');
    end;
  finally
    ContinuePrompt;
  end;
end;

procedure TTestEditor.WriteSpacer;
var
  idx: nativeuint;
begin
  for idx := 0 to 199 do Writeln;
end;

procedure TTestEditor.MainMenu;
var
  Option: string;
  Parts: TArrayOfString;
begin
  repeat
    MainMenuText;
    Readln(Option);
    WriteSpacer;
    Parts := Option.Explode(' ');
    if Length(Parts)=0 then continue;
    Option := Parts[0].UppercaseTrim;
    if Option='HELP'    then begin
      if Length(Parts)<>2 then begin
        Writeln('Missing <command> parameter.');
        Writeln('Please speficy a command you would like help for.');
        Writeln('For Example, "HELP LIST"');
        ContinuePrompt;
        continue;
      end;
      DisplayHelp( Parts[1] );
    end else
    if Option='LIST'    then begin
     if Length(Parts)=2 then begin
      ListTestCases( Parts[1] );
     end else begin
       ListTestCases;
     end;
    end else if Option='SELECT'  then begin
      if Length(Parts)<2 then begin
        Writeln('Missing <testcase> parameter.');
        ContinuePrompt;
        continue;
      end;
      TestCaseMenu( Parts[1] )
    end else
    if Option='DISABLE' then begin
      if Length(Parts)<2 then begin
        Writeln('Missing <testcase> parameter.');
        ContinuePrompt;
        continue;
      end;
      if Parts[1].UppercaseTrim='ALL' then begin
        DisableALLCases;
      end else begin
        SetEnableDisableCase( Parts[1], FALSE );
      end;
    end else if Option='ENABLE'  then begin
      if Length(Parts)<2 then begin
        Writeln('Missing <testcase> parameter.');
        ContinuePrompt;
        continue;
      end;
      if Parts[1].UppercaseTrim='ALL' then begin
        EnableALLCases;
      end else begin
        SetEnableDisableCase( Parts[1], TRUE  );
      end;
    end else if Option='RUN' then begin
      Writeln('Running test suite...');
      Writeln;
      fTestSuite.Run( 'test_cwRuntime', [TStandardConsoleReport.Create] );
      Writeln;
      Writeln('Test suite complete.');
      ContinuePrompt;
    end;
  until Option='EXIT';
end;

procedure TTestEditor.Run;
begin
  MainMenu;
end;

end.
