﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwThreading.Criticalsection.Standard;
{$ifdef fpc}{$mode delphiunicode}{$endif}

interface
uses
  cwTypes
, cwThreading
{$ifndef fpc}
  {$ifdef MSWINDOWS}
  , cwWin32
  {$else}
  , cwPosix
  {$endif}
{$endif}
;

type
  TCriticalSection = class( TInterfacedObject, ICriticalSection )
  private
    fLockedByThreadID: TThreadID;
    {$ifdef fpc}
    fMutex: TRTLCriticalSection;
    {$else}
    fMutex: {$ifdef MSWINDOWS} SRWLOCK; {$else} pthread_mutex_t; {$endif}
    {$endif}
  strict private //- ICriticalSection -//
    procedure Acquire;
    procedure Release;
  public
    constructor Create;
    destructor Destroy; override;
  end;

implementation

procedure TCriticalSection.Acquire;
var
  CurrentThread: TThreadID;
begin
  CurrentThread := GetCurrentThreadID;
  if fLockedByThreadID=CurrentThread then exit;
  {$ifdef fpc}
    EnterCriticalSection( fMutex );
  {$else}
    {$ifdef MSWINDOWS}
    AcquireSRWLockExclusive(fMutex);
    {$else}
    pthread_mutex_lock(@fMutex);
    {$endif}
  {$endif}
  fLockedByThreadID := CurrentThread;
end;

procedure TCriticalSection.Release;
begin
  fLockedByThreadID := 0;
  {$ifdef fpc}
    LeaveCriticalSection( fMutex );
  {$else}
    {$ifdef MSWINDOWS}
    ReleaseSRWLockExclusive(fMutex);
    {$else}
    pthread_mutex_unlock(@fMutex)
    {$endif}
  {$endif}
end;

constructor TCriticalSection.Create;
begin
  inherited Create;
  fLockedByThreadID := 0;
  {$ifdef fpc}
  InitCriticalSection( fMutex );
  {$else}
    {$ifdef MSWINDOWS}
    InitializeSRWLock(fMutex);
    {$else}
    if pthread_mutex_init(@fMutex, nil)<>0 then TStatus(stCreateMutexFailed).Raize([GetLastError().AsString]);
    {$endif}
  {$endif}
end;

destructor TCriticalSection.Destroy;
begin
  {$ifdef fpc}
  DoneCriticalSection( fMutex );
  {$else}
    {$ifndef MSWINDOWS}
    pthread_mutex_destroy(@fMutex);
    {$endif}
  {$endif}
  inherited Destroy;
end;

end.

