﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwThreading.Promise.Custom;
{$ifdef fpc}
  {$mode delphiunicode}
  {$modeswitch nestedprocvars}
{$endif}

interface
uses
  cwTypes
, cwCollections
{$ifndef fpc}
, typinfo
, rtti
{$endif}
, cwThreading
, cwThreading.PromiseAction.Standard
;

  (*
     Internal interface specific to promises for managing
     their life cycle.
  *)
  type
    IPromiseLifeCycle = interface
      ['{A2CD591A-A9BB-4611-9EBA-49F98F87DC3E}']

      ///  <summary>
      ///    Called by the operating thread to lock
      ///    access to the pending promises list while
      ///    those promises are copied up.
      ///  </summary>
      procedure LockPending;

      ///  <summary>
      ///    Called by the operating thread to
      ///    unlock access to the pending promises list
      ///    after having copied the promises up.
      ///  </summary>
      procedure UnlockPending;

      ///  <summary>
      ///    Returns a reference to any promses added
      ///    via TPromise.AddPromise, to be collected
      ///    and scheduled by the operating thread.
      ///  </summary>
      function PendingPromises: IList< IPromiseLifeCycle >;

      ///  <summary>
      ///    The calling thread will pass its control
      ///    interface to the promise, so that the promise
      ///    can take a reference to it. <br/>
      ///    The reference may then be used to wake or
      ///    terminate the calling thread.
      ///  </summary>
      procedure StartUp( const Control: IThreadControl );

      ///  <summary>
      ///    Called by the calling thread repeatedly
      ///    when awake.
      ///  </summary>
      procedure Execute;

      ///  <summary>
      ///    The promise must free its reference to the
      ///    thread control on a call to shutdown, which
      ///    the calling thread will call when the thread
      ///    terminates. Clearing the reference enables
      ///    the threading system to then dispose of
      ///    the class wrapper which hosts the
      ///    thread.
      ///  </summary>
      procedure Shutdown;
    end;

  {$M+}
  TCustomPromise = class( TInterfacedObject, IPromiseLifeCycle )
  private
    fClassName: string;
    fThreadID: TThreadID;
    fThreadControl: IThreadControl;
    fActions: IList<IPromiseAction>;
    fActionsCS: ICriticalSection;
    fPendingPromises: IList<IPromiseLifeCycle>;
    fPendingPromisesCS: ISignaledCriticalSection;
    {$ifndef fpc}
    fRTTIContext: TRTTIContext;
    fRTTIType: TRTTIType;
    {$endif}
  private
    procedure ExecuteAction( const Action: IPromiseAction );
  strict private //- IPromiseLifeCycle -//
    procedure LockPending;
    procedure UnlockPending;
    function PendingPromises: IList<IPromiseLifeCycle>;
    procedure Startup( const ThreadControl: IThreadControl );
    procedure Shutdown;
    procedure Execute;
  public
    constructor Create; virtual;
    destructor Destroy; override;
  protected //- Available to descend.
    function DoIsPromise: boolean;
    {$ifdef fpc}
    procedure DoPerform( const Method: TPromisedMethod );
    {$else}
    function DoPerform( const MethodName: string; const Args: TArray<TValue> ): TValue;
    {$endif}
  protected //- To consume
    function AddPromise( const Promise: TInterfacedObject ): IInterface;
  end;

implementation
uses
  cwWin32
, cwPosix
, cwCollections.Factory
, cwThreading.Factory
, cwDebugLog
;


procedure TCustomPromise.ExecuteAction(const Action: IPromiseAction);
var
  Method: {$ifdef fpc} TPromisedMethod {$else} TRTTIMethod {$endif};
begin
  {$ifdef fpc}
  try
    Method := Action.getMethod();
    if assigned( Method ) then Method();
    Action.setStatus( psCompleted );
  except
    Action.setStatus( psFailed );
  end;
  {$else}
  // Delphi uses RTTI to find and execute the method.
  Method := fRTTIType.GetMethod( Action.getMethodName );
  if not assigned(Method) then begin
    Action.setStatus( psFailed );
    exit;
  end;
  if Length( Method.GetParameters ) <> Action.getArgCount then begin
    TStatus( stIncorrectNumberOfParameters ).Raize( [ Method.Name ] );
  end;
  try
    Action.setMethodReturn( Method.Invoke( Self, Action.getMethodArgs ) );
  except
    Action.setStatus( psFailed );
  end;
  Action.setStatus( psCompleted );
  {$endif}
end;


procedure TCustomPromise.LockPending;
begin
  fPendingPromisesCS.Acquire;
end;

function TCustomPromise.PendingPromises: IList<IPromiseLifeCycle>;
begin
  Result := fPendingPromises;
end;

procedure TCustomPromise.Execute;
var
  idx: nativeuint;
begin
  fActionsCS.Acquire;
  try
    if fActions.Count=0 then exit;
    try
      for idx := 0 to pred( fActions.Count ) do begin
        ExecuteAction( fActions[ idx ] );
      end;
    finally
      fActions.Clear;
    end;
  finally
    fActionsCS.Release;
  end;
end;

procedure TCustomPromise.Shutdown;
begin
  fThreadID := 0;
  fThreadControl := nil;
  {$ifndef fpc}
  fRTTIType.Free;
  fRTTIContext.Free;
  {$endif}
  fActionsCS := nil;
  fActions := nil;
end;

procedure TCustomPromise.Startup( const ThreadControl: IThreadControl );
begin
  fThreadID := GetCurrentThreadID;
  fThreadControl := ThreadControl;
  {$ifndef fpc}
  fRTTIContext := TRTTIContext.Create;
  fRTTIType    := fRTTIContext.GetType({$ifdef fpc}fInterfaceTypeInfo{$else}Self.ClassInfo{$endif});
  {$endif}
end;

procedure TCustomPromise.UnlockPending;
begin
  fPendingPromisesCS.Release;
end;

function TCustomPromise.AddPromise( const Promise: TInterfacedObject ): IInterface;
var
  PromiseI: IInterface;
begin
  Result := nil;
  PromiseI := Promise as IInterface;
  if not Supports(Promise, IPromiseLifeCycle) then TStatus( stObjectMustDescend ).Raize( ['Promise','cwThreading.Factory.TPromise'] );
  fPendingPromisesCS.Acquire;
  try
    fPendingPromises.Add(PromiseI as IPromiseLifeCycle);
  finally
    fPendingPromisesCS.Release;
  end;
  Result := PromiseI;
end;

constructor TCustomPromise.Create;
begin
  inherited Create;
  with DebugLog(Self,'Create') do try

    {$ifdef fpc}
    fClassName := string(UnitName) + '.' + string(ClassName);
    {$else}
    fClassName := UnitName + '.' + ClassName;
    {$endif}
    fActions := TList<IPromiseAction>.Create(32,False,False);
    fActionsCS := TCriticalSection.Create;
    fPendingPromises := TList<IPromiseLifeCycle>.Create(8);
    fPendingPromisesCS := TSignaledCriticalSection.Create;
    Startup( nil );

  finally
  end;
end;

destructor TCustomPromise.Destroy;
begin
  with DebugLog(Self,'Create') do try

    Shutdown;
    fPendingPromisesCS.Acquire;
    try
      fPendingPromises := nil;
    finally
      fPendingPromisesCS.Release;
      fPendingPromisesCS := nil;
    end;
    inherited Destroy;

  finally
  end;
end;

function TCustomPromise.DoIsPromise: boolean;
var
  ThreadID: TThreadID;
begin
  if fThreadID = 0 then TStatus( stPromiseNotRunningInThread ).Raize( fClassName );
  ThreadID := GetCurrentThreadID;
  Result := (fThreadID = ThreadID);
end;

{$ifdef fpc}
procedure TCustomPromise.DoPerform( const Method: TPromisedMethod );
var
  Action: IPromiseAction;
begin
  if fThreadID=0 then TStatus( stPromiseNotRunningInThread ).Raize( fClassName );
  Action := TPromiseAction.Create;
  try
    Action.setMethod( Method );
    // Place promise request.
    fActionsCS.Acquire;
    try
      fActions.Add( Action );
    finally
      fActionsCS.Release;
    end;
    //- Wait for the promise action to be executed.
    fThreadControl.Wake;
    while Action.getStatus = psPending do Sleep(0);
    if Action.getStatus = psFailed then TStatus( stPromiseActionFailed ).Raize( [ fClassName ] );
  finally
    Action := nil;
  end;
end;

{$else}

function TCustomPromise.DoPerform( const MethodName: string; const Args: TArray<TValue> ): TValue;
var
  GotArgs: TArray<TValue>;
  Action: IPromiseAction;
  idx: nativeuint;
begin
  with DebugLog(Self,'DoPerform') do try
    DebugLn(MethodName);

    while not assigned(fThreadControl) do sleep(0); //- The thread has not yet started up, so wait for it.
    if fThreadID=0 then TStatus( stPromiseNotRunningInThread ).Raize( fClassName );
    Action := TPromiseAction.Create;
    try
      Action.SetMethodArgs( Args );
      Action.setMethodName( MethodName );
      // Place promise request.
      fActionsCS.Acquire;
      try
        fActions.Add(Action);
      finally
        fActionsCS.Release;
      end;
      //- Wait for the promise action to be executed.
      fThreadControl.Wake;
      while Action.getStatus = psPending do Sleep(0); //- Waiting for promise to be executed.
      //- Return
      case Action.getStatus of
        psFailed: TStatus( stPromiseActionFailed ).Raize( [ fClassName, MethodName ] );
        psCompleted: begin
          Result := Action.getMethodReturn;
          if Length( Args ) = 0 then exit;
          GotArgs := Action.getMethodArgs;
          for idx := 0 to pred( Length( Args ) ) do Args[ idx ] := GotArgs[ idx ];
        end;
      end;
    finally
      Action := nil;
    end;
  finally
  end;
end;
{$endif}

end.
