﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwThreading.PromiseThread.Standard;
{$ifdef fpc}
  {$mode delphiunicode}
  {$warn 05079 off}
{$endif}

interface
uses
  cwTypes
, cwThreading
, cwCollections
, cwThreading.Promise.Custom
;

type
  (* Used to enable look-up of supported promises. *)
  IPromiseThread = interface
    ['{5A9CF304-F46E-4F64-BA1E-B19D2B6EE542}']

    ///  <summary>
    ///    Used to inject a promise into the promise thread when
    ///    the promise thread is in-fact on the main thread.
    ///  </summary>
    procedure MainThreadInject( const Promise: IPromiseLifecycle );

    ///  <summary>
    ///    Returns a list of promises (pending or active) which
    ///    implement the specified interface.
    ///  </summary>
    function Resolve( const Iface: TGUID ): IReadOnlyList<IInterface>;
  end;

  TPromiseThread = class( TInterfacedObject, IThreadDelegate, ILifeCycleDelegate, IPromiseThread )
  private
    fExecutingThreadID: TThreadID;
    fControlRef: IThreadControl; //- is nil except during execute.
    fPendingPromises: IList<IPromiseLifeCycle>;
    fPendingPromisesCS: ISignaledCriticalSection;
    fPromises: IList<IPromiseLifeCycle>;
    fPromisesCS: ISignaledCriticalSection;
  private
    procedure ExecutePromises;
    procedure ExecutePromise(const Promise: IPromiseLifeCycle);
    procedure CollectPendingPromises;
    procedure CollectPendingFromPromise(const Promise: IPromiseLifeCycle);
    procedure CollectPendingPromise(const Promise: IPromiseLifeCycle);
    procedure AddPendingPromises;
    procedure AddPendingPromise(const Promise: IPromiseLifeCycle);
    procedure ShutdownPromise(const Promise: IPromiseLifeCycle);
    function TerminateIfDone( const Control: IThreadControl ): boolean;
  protected
    procedure DoTerminatePromise( const Promise: IInterface );
  public
    class procedure TerminatePromise( const Promise: IInterface ); static;
  strict private //- ILifeCycleDelegate -//
    procedure Startup;
    procedure Shutdown;
  strict private //- IThreadDelegate -//
    procedure Execute( const Control: IThreadControl );
  strict private //- IPromiseThread -//
    procedure MainThreadInject( const Promise: IPromiseLifecycle );
    function Resolve( const Iface: TGUID ): IReadOnlyList<IInterface>;
  public
    constructor Create( const FirstPromise: IPromiseLifeCycle );
    destructor Destroy; override;
  end;



implementation
uses
  cwWin32
, cwCollections.Factory
, cwThreading.CriticalSection.Standard
, cwThreading.SignaledCriticalSection.Standard
;


(*
   This list may be used to locate the thread operating a particular
   promise.
*)
var
  PromiseThreads: IList<TPromiseThread> = nil;
  PromiseThreadsCS: ICriticalSection = nil;

procedure EnsurePromiseThreads;
begin
  if not assigned(PromiseThreadsCS) then begin
    PromiseThreadsCS := TCriticalSection.Create;
    PromiseThreadsCS.Acquire;
    try
      PromiseThreads := TList<TPromiseThread>.Create(32);
    finally
      PromiseThreadsCS.Release;
    end;
  end;
end;

procedure DisposePromiseThreads;
begin
  if not assigned(PromiseThreadsCS) then exit;
  PromiseThreadsCS.Acquire;
  try
    PromiseThreads := nil;
  finally
    PromiseThreadsCS.Release;
    PromiseThreadsCS := nil;
  end;
end;

class procedure TPromiseThread.TerminatePromise(const Promise: IInterface);
var
  idx: nativeuint;
begin
  EnsurePromiseThreads;
  PromiseThreadsCS.Acquire;
  try
    if PromiseThreads.Count=0 then exit;
    for idx := pred(PromiseThreads.Count) downto 0 do begin
      PromiseThreads[idx].DoTerminatePromise(Promise);
    end;
  finally
    PromiseThreadsCS.Release;
  end;
end;

procedure TPromiseThread.CollectPendingPromise( const Promise: IPromiseLifeCycle );
begin
  fPendingPromisesCS.Acquire;
  try
    fPendingPromises.Add( Promise );
  finally
    fPendingPromisesCS.Release;
  end;
end;

procedure TPromiseThread.CollectPendingFromPromise( const Promise: IPromiseLifeCycle );
var
  aPromise: IPromiseLifeCycle;
begin
  Promise.LockPending;
  try
    for aPromise in Promise.PendingPromises do CollectPendingPromise( aPromise );
  finally
    Promise.PendingPromises.Clear;
    Promise.UnlockPending;
  end;
end;

procedure TPromiseThread.CollectPendingPromises;
var
  Promise: IPromiseLifecycle;
begin
  fPromisesCS.Acquire;
  try
    for Promise in fPromises do CollectPendingFromPromise( Promise );
  finally
    fPromisesCS.Release;
  end;
end;

constructor TPromiseThread.Create( const FirstPromise: IPromiseLifeCycle );
begin
  inherited Create;
  fExecutingThreadID := 0;
  fPromises          := TList<IPromiseLifeCycle>.Create;
  fPromisesCS        := TSignaledCriticalSection.Create;
  fPendingPromises   := TList<IPromiseLifeCycle>.Create;
  fPendingPromisesCS := TSignaledCriticalSection.Create;
  if assigned(FirstPromise) then begin
    fPendingPromisesCS.Acquire;
    try
      fPendingPromises.Add( FirstPromise );
    finally
      fPendingPromisesCS.Release;
    end;
  end;
  //- Add to un-referenced global list for lookup.
  EnsurePromiseThreads;
  PromiseThreadsCS.Acquire;
  try
    PromiseThreads.Add(self);
  finally
    PromiseThreadsCS.Release;
  end;
end;

destructor TPromiseThread.Destroy;
begin
  //- Remove from un-referenced global list.
  EnsurePromiseThreads;
  PromiseThreadsCS.Acquire;
  try
    PromiseThreads.Remove(pointer(self));
  finally
    PromiseThreadsCS.Release;
  end;
  //- Clear the promises.
  fPromisesCS.Acquire;
  try
    fPromises.Clear;
  finally
    fPromisesCS.Release;
  end;
  fPromisesCS := nil;
  fPromises := nil;
  inherited Destroy;
end;

procedure TPromiseThread.AddPendingPromise( const Promise: IPromiseLifeCycle );
begin
  fPromisesCS.Acquire;
  try
    Promise.StartUp( fControlRef );
    if Supports(Promise,ILifeCycleDelegate) then begin
      (Promise as ILifeCycleDelegate).Startup;
    end;
    fPromises.Add( Promise );
  finally
    fPromisesCS.Release;
  end;
end;

procedure TPromiseThread.AddPendingPromises;
var
  Promise: IPromiseLifecycle;
begin
  fPendingPromisesCS.Acquire;
  try
    if fPendingPromises.Count=0 then exit;
    try
      for Promise in fPendingPromises do AddPendingPromise( Promise );
    finally
      fPendingPromises.Clear;
    end;
  finally
    fPendingPromisesCS.Release;
  end;
end;

procedure TPromiseThread.ExecutePromise( const Promise: IPromiseLifeCycle );
begin
  Promise.Execute;
end;

procedure TPromiseThread.ExecutePromises;
var
  Promise: IPromiseLifeCycle;
begin
  for Promise in fPromises do ExecutePromise( Promise );
end;

procedure TPromiseThread.MainThreadInject( const Promise: IPromiseLifecycle );
begin
  fPendingPromisesCS.Acquire;
  try
    fPendingPromises.Add( Promise );
  finally
    fPendingPromisesCS.Release;
  end;
end;

function TPromiseThread.Resolve( const Iface: TGUID ): IReadOnlyList<IInterface>;
var
  idx: nativeuint;
  List: IList<IInterface>;
begin
  List := TList<IInterface>.Create;
  try
    fPromisesCS.Acquire;
    fPendingPromisesCS.Acquire;
    try
      if fPromises.Count>0 then begin
        for idx := 0 to pred(fPromises.Count) do begin
          if Supports( fPromises[idx], IFace ) then List.Add(fPromises[idx]);
        end;
      end;
      if fPendingPromises.Count>0 then begin
        for idx := 0 to pred(fPendingPromises.Count) do begin
          if Supports( fPendingPromises[idx], IFace ) then List.Add(fPendingPromises[idx]);
        end;
      end;
    finally
      fPendingPromisesCS.Release;
      fPromisesCS.Release;
    end;
  finally
    Result := List.getAsReadOnly;
  end;
end;

procedure TPromiseThread.DoTerminatePromise( const Promise: IInterface );
var
  idx: nativeuint;
begin
  fPromisesCS.Acquire;
  try
    if fPromises.Count>0 then begin
      for idx := pred(fPromises.Count) downto 0 do begin
        if fPromises[idx] = Promise then begin
          fPromises.RemoveItem(idx);
          exit;
        end;
      end;
    end;
  finally
    fPromisesCS.Release;
  end;
  fPendingPromisesCS.Acquire;
  try
    if fPendingPromises.Count>0 then begin
      for idx := pred(fPendingPromises.Count) downto 0 do begin
        if fPendingPromises[idx] = Promise then begin
          fPendingPromises.RemoveItem(idx);
          exit;
        end;
      end;
    end;
  finally
    fPendingPromisesCS.Release;
  end;
end;

procedure TPromiseThread.ShutdownPromise( const Promise: IPromiseLifeCycle );
begin
  if Supports(Promise,ILifeCycleDelegate) then (Promise as ILifeCycleDelegate).ShutDown;
  Promise.Shutdown;
end;

procedure TPromiseThread.Shutdown;
var
  Promise: IPromiseLifeCycle;
begin
  fPromisesCS.Acquire;
  try
    if fPromises.Count=0 then exit;
    for Promise in fPromises do ShutdownPromise( Promise );
  finally
    fPromisesCS.Release;
  end;
end;

procedure TPromiseThread.Startup;
begin
  fExecutingThreadID := cwThreading.GetCurrentThreadID;
end;

function TPromiseThread.TerminateIfDone( const Control: IThreadControl ): boolean;
begin
  Result := False;
  fPromisesCS.Acquire;
  try
    fPendingPromisesCS.Acquire;
    try
      if (fPromises.Count=0) and (fPendingPromises.Count=0) then begin
        Control.Terminate;
        Result := True;
      end;
    finally
      fPendingPromisesCS.Release;
    end;
  finally
    fPromisesCS.Release;
  end;
end;

procedure TPromiseThread.Execute( const Control: IThreadControl );
begin
  fControlRef := Control;
  try
    CollectPendingPromises;
    AddPendingPromises;
    if TerminateIfDone( Control ) then exit;
    ExecutePromises;
    Control.Sleep;
  finally
    fControlRef := nil;
  end;
end;

initialization
  EnsurePromiseThreads;

finalization
  DisposePromiseThreads;

end.
