﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwThreading.SignaledCriticalSection.Standard;
{$ifdef fpc}{$mode delphiunicode}{$endif}

interface
uses
  cwThreading
{$ifndef fpc}
  {$ifdef MSWINDOWS}
, cwWin32
  {$else}
, cwPosix
  {$endif}
{$endif}
;

type
  TSignaledCriticalSection = class( TInterfacedObject, ISignaledCriticalSection )
  private
    fLockedByThreadID: TThreadID;
    {$ifdef fpc}
      fTimeoutMS: uint32;
      fMutex: TRTLCriticalSection;
      fCondition: PRTLEvent;
    {$else}
      {$ifdef MSWINDOWS}
      fTimeoutMS: uint64;
      fMutex: SRWLOCK;
      fCondition: CONDITION_VARIABLE;
      {$else}
      fSleepTimeoutNanoseconds: uint64;
      fMutex: pthread_mutex_t;
      fCondition: pthread_cond_t;
      {$endif}
    {$endif}
  strict private //- ISignaledCriticalSection -//
    procedure Acquire;
    procedure Release;
    procedure Sleep;
    procedure Wake;
  public
    constructor Create( const SleepTimeoutMilliseconds: uint64 = 0 ); reintroduce;
    destructor Destroy; override;
  end;

implementation
uses
  cwTypes
, cwTiming
;

procedure TSignaledCriticalSection.Acquire;
var
  CurrentThread: TThreadID;
begin
  CurrentThread := GetCurrentThreadID;
  if fLockedByThreadID=CurrentThread then exit;
  {$ifdef fpc}
  EnterCriticalSection( fMutex );
  {$else}
    {$ifdef MSWINDOWS}
    AcquireSRWLockExclusive(fMutex);
    {$else}
    pthread_mutex_lock(@fMutex);
    {$endif}
  {$endif}
  fLockedByThreadID := CurrentThread;
end;

procedure TSignaledCriticalSection.Release;
begin
  fLockedByThreadID := 0;
  {$ifdef fpc}
  LeaveCriticalSection( fMutex );
  {$else}
    {$ifdef MSWINDOWS}
    ReleaseSRWLockExclusive(fMutex);
    {$else}
    pthread_mutex_unlock(@fMutex);
    {$endif}
  {$endif}
end;

procedure TSignaledCriticalSection.Sleep;
{$ifndef fpc}
  {$ifndef MSWINDOWS}
  var
    retval  : int32;
    ts      : Timespec;
    TotalNS : int64;
  {$endif}
{$endif}
begin
  {$ifdef fpc}
    Release;
    try
      if fTimeoutMS = 0 then begin
        RTLEventWaitFor( fCondition );
      end else begin
        RTLEventWaitFor( fCondition, fTimeoutMS );
      end;
      RTLEventResetEvent( fCondition );
    finally
      Acquire;
    end;
  {$else}
    {$ifdef MSWINDOWS}
    if not SleepConditionVariableSRW(fCondition, fMutex, fTimeoutMS, 0) then begin
      if GetLastError <> ERROR_TIMEOUT then TStatus(stThreadSleepFailed).Raize;
    end;
    {$else}
    if fSleepTimeoutNanoseconds=0 then begin
      if pthread_cond_wait(
        @fCondition,
        @fMutex
      ) <> 0 then begin
        TStatus(stThreadSleepFailed).Raize;
      end;

    end else begin
      //- Get the current time and add the interval which is in nanoseconds.
      clock_gettime( CLOCK_REALTIME, ts );
      TotalNS    := ( ts.tv_sec *  TTimeConstants.NanosecondsPerSecond );
      TotalNS    := ( TotalNS   +  ts.tv_nsec                          );
      TotalNS    := ( TotalNS   +  fSleepTimeoutNanoseconds            );
      ts.tv_sec  := ( TotalNS  div TTimeConstants.NanosecondsPerSecond );
      ts.tv_nsec := ( TotalNS  mod TTimeConstants.NanosecondsPerSecond );
      // Now run the wait until we timeout.
      repeat
        retval := pthread_cond_timedwait(
          @fCondition,
          @fMutex,
          @ts
        );
        {$ifndef fpc}
        if retval=ETIMEDOUT then exit;
        {$else}
        if (retval=ESysETIMEDOUT) or (retval=ETIMEDOUT) then exit;
        {$endif}
      until (RetVal<>0);
      if RetVal=0 then exit;
      TStatus(stThreadSleepFailed).Raize;
    end;
    {$endif}
  {$endif}
end;

procedure TSignaledCriticalSection.Wake;
begin
  {$ifdef fpc}
    RTLEventSetEvent( fCondition );
  {$else}
    {$ifdef MSWINDOWS}
    WakeConditionVariable(fCondition);
    {$else}
    pthread_cond_signal(@fCondition);
    {$endif}
  {$endif}
end;

constructor TSignaledCriticalSection.Create(const SleepTimeoutMilliseconds: uint64);
begin
  inherited Create;
  fLockedByThreadID := 0;
  {$ifdef fpc}
  fTimeoutMS := SleepTimeoutMilliseconds;
  InitCriticalSection( fMutex );
  fCondition := RTLEventCreate;
  {$else}
    {$ifdef MSWINDOWS}
    InitializeSRWLock(fMutex);
    InitializeConditionVariable(fCondition);
    if SleepTimeoutMilliseconds=0 then fTimeoutMS := INFINITE else fTimeoutMS := SleepTimeoutMilliseconds;
    {$else}
    fSleepTimeoutNanoSeconds := SleepTimeoutMilliseconds * TTimeConstants.NanosecondsPerMillisecond;
    if pthread_mutex_init(@fMutex, nil)<>0 then TStatus(stInitMutexFailed).Raize([GetLastError().AsString]);
    if pthread_cond_init(@fCondition,nil)<>0 then TStatus(stInitConditionVariableFailed).Raize([GetLastError().AsString]);
    {$endif}
  {$endif}
end;

destructor TSignaledCriticalSection.Destroy;
begin
  {$ifdef fpc}
  DoneCriticalSection( fMutex );
  RTLEventDestroy( fCondition );
  {$else}
    {$ifndef MSWINDOWS}
    pthread_mutex_destroy(@fMutex);
    pthread_cond_destroy(@fCondition);
    {$endif}
  {$endif}
  inherited Destroy;
end;

end.

