﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwThreading.TaskSchedule.Standard;
{$ifdef fpc} {$mode delphiunicode} {$endif}

interface
uses
  cwCollections
, cwThreading
, cwTiming
;


type
  (* Used to store tasks as scheduled items within the task schedule *)
  IScheduledTask = interface
    ['{DDF1A1C8-379C-4ACD-8856-6D0475B1FA76}']
    function Delegate: IInterface;
    procedure Execute;
  end;

type
  (* Used by TThreads to control the scheduler *)
  ITaskSchedule = interface( IThreadDelegate )
    ['{3B0E342B-8254-4A3D-9FB7-52AFCD933432}']

    (* Adds a task delegate to the schedule at the specified interval *)
    procedure AddTask( const TaskDelegate: IInterface; const IntervalMS: nativeuint );

  end;

  (* Implements ITaskScheduler as an IThreadDelegate *)
  TTaskSchedule = class( TInterfacedObject, IThreadDelegate, ITaskSchedule )
  private
    fScheduledTasks: IList<IScheduledTask>;
    fScheduleCS: ICriticalSection;
  private
    procedure ExecuteTask(const Task: IScheduledTask);
  strict private //- ITaskSchedule -//
    procedure AddTask( const TaskDelegate: IInterface; const IntervalMS: nativeuint );
  strict private //- ITaskDelegate -//
    procedure Execute( const Control: IThreadControl );
  public
    constructor Create;
    destructor Destroy; override;
  end;

implementation
uses
  cwTypes
, cwCollections.Factory
, cwThreading.CriticalSection.Standard
, cwTiming.Factory
;

type
  TScheduledTask = class( TInterfacedObject, IScheduledTask )
  private
    fTicksPerMS: nativeuint;
    fPriorticks: nativeuint;
    fIntervalMS: nativeuint;
    fDelegate: IInterface;
    fTimer: ITimer;
  strict private //- IScheduledTask -//
    function Delegate: IInterface;
    procedure Execute;
  public
    constructor Create( const Delegate: IInterface; const Interval: nativeuint );
  end;

constructor TScheduledTask.Create(const Delegate: IInterface; const Interval: nativeuint);
begin
  inherited Create;
  fTimer := TTimer.Create;
  if (not Supports(Delegate,ITaskDelegate)) and
     (not Supports(Delegate,IScheduledTaskDelegate)) then begin
    TStatus(stObjectMustSupportEitherInterface).Raize( ['TaskDelegate','ITaskDelegate','IScheduledTaskDelegate'] );
  end;
  fDelegate := Delegate;
  fIntervalMS := Interval;
  fPriorticks := 0;
  fTicksPerMS := fTimer.TicksPerSecond div TTimeConstants.MillisecondsPerSecond;
  fTimer.Clear;
end;

function TScheduledTask.Delegate: IInterface;
begin
  Result := fDelegate;
end;

procedure TScheduledTask.Execute;
var
  Ticks: nativeuint;
  DeltaTms: nativeuint;
begin
  Ticks := fTimer.DeltaTicks;
  fPriorticks := fPriorticks + Ticks;
  DeltaTms := (fPriorticks div fTicksPerMS);
  if DeltaTms>=fIntervalMS then begin
    try
      if not assigned(fDelegate) then exit;
      if Supports(fDelegate,IScheduledTaskDelegate) then begin
        (fDelegate as IScheduledTaskDelegate).Execute(DeltaTms);
      end;
      if Supports(fDelegate,ITaskDelegate) then begin
        (fDelegate as ITaskDelegate).Execute;
      end;
    finally
      fPriorticks := 0;
    end;
  end;
end;

procedure TTaskSchedule.AddTask(const TaskDelegate: IInterface; const IntervalMS: nativeuint);
var
  NewScheduledTask: IScheduledTask;
begin
  fScheduleCS.Acquire;
  try
    NewScheduledTask := TScheduledTask.Create( TaskDelegate, IntervalMS );
    fScheduledTasks.Add( NewScheduledTask );
  finally
    fScheduleCS.Release;
  end;
end;

constructor TTaskSchedule.Create;
begin
  inherited Create;
  fScheduledTasks := TList<IScheduledTask>.Create(8);
  fScheduleCS := TCriticalSection.Create;
end;

destructor TTaskSchedule.Destroy;
begin
  fScheduleCS.Acquire;
  try
    fScheduledTasks := nil;
  finally
    fScheduleCS.Release;
    fScheduleCS := nil;
  end;
  inherited;
end;

procedure TTaskSchedule.ExecuteTask( const Task: IScheduledTask );
begin
  Task.Execute;
end;

procedure TTaskSchedule.Execute( const Control: IThreadControl );
var
  Task: IScheduledTask;
begin
  fScheduleCS.Acquire;
  try
    for Task in fScheduledTasks do ExecuteTask( Task );
    Control.Sleep; //- The thread is set to 1ms by cwThreading.Threads.CreateSchedule
  finally
    fScheduleCS.Release;
  end;
end;

end.
