﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwthreading.taskthread.standard;
{$ifdef fpc}
  {$mode delphiunicode}
  {$MODESWITCH NESTEDPROCVARS}
{$endif}


interface
uses
  cwTypes
, cwThreading
, cwCollections
;

type
  ITask = interface
    ['{CEA8670C-3965-4343-A6C5-B797AEC29CE7}']
  end;

  ITaskSet = interface
    ['{1F9B110C-47D7-4E86-8EEB-19ECC82D7C28}']
  end;

type
  ITaskThread = interface ['{15BFE00C-F218-4A61-86D3-31932AA0644E}'] end; //<- So that we can locate task threads from cwthreading.threads.standard
  TTaskThread = class( TInterfacedObject, IThreadDelegate, ITaskThread )
  strict private //- IThreadDelegate -//
    procedure Execute( const Control: IThreadControl );
  private
  public
    class procedure AddTaskSet( const Tasks: IReadOnlyList<ITaskDelegate>; const NotifyDone: TNotifyDelegate ); static;
  end;

implementation
uses
  cwCollections.Factory
, cwThreading.CriticalSection.Standard
, cwThreading.Thread.Standard
;

var
  TasksEnabled: boolean = FALSE;
  TaskList: IList<ITask> = nil;
  TaskListCS: ICriticalSection = nil;


type
  TTaskSet = class( TInterfacedObject, ITaskSet )
  private
    fNotify: TNotifyDelegate;
  public
    constructor Create( const Notify: TNotifyDelegate );
    destructor Destroy; override;
  end;

constructor TTaskSet.Create( const Notify: TNotifyDelegate );
begin
  inherited Create;
  fNotify := Notify;
end;

destructor TTaskSet.Destroy;
begin
  try
    if not TasksEnabled then exit;
    fNotify.Execute;
  finally
    inherited Destroy;
  end;
end;


type
  TTask = class( TInterfacedObject, ITask )
  private
    fTask: ITaskDelegate;
    fTaskSet: ITaskSet;
  public
    constructor Create( const TaskSet: ITaskSet; const Task: ITaskDelegate );
    destructor Destroy; override;
  end;

constructor TTask.Create( const TaskSet: ITaskSet; const Task: ITaskDelegate );
begin
  inherited Create;
  fTaskSet := TaskSet;
  fTask := Task;
end;

destructor TTask.Destroy;
begin
  try
    if not TasksEnabled then exit;
    if assigned(fTask) then begin
      try
        fTask.Execute;
      finally
       fTask := nil;
       fTaskSet := nil;
      end;
    end;
  finally
    inherited Destroy;
  end;
end;


class procedure TTaskThread.AddTaskSet( const Tasks: IReadOnlyList<ITaskDelegate>; const NotifyDone: TNotifyDelegate );
var
  NewTask: ITask;
  NewTaskSet: ITaskSet;
  idx: nativeuint;
begin
  NewTaskSet := TTaskSet.Create( NotifyDone );
  if Tasks.Count>0 then begin
    TaskListCS.Acquire;
    try
      for idx := 0 to pred(Tasks.Count) do begin
        NewTask := TTask.Create( NewTaskSet, Tasks[idx] );
        TaskList.Add(NewTask);
      end;
    finally
      TaskListCS.Release;
    end;
  end;
end;


procedure TTaskThread.Execute(const Control: IThreadControl);
var
  CurrentTask: ITask;
begin
  //- Attempt to pull a task from the task list.
  CurrentTask := nil;
  TaskListCS.Acquire;
  try
    if TaskList.Count>0 then begin
      CurrentTask := TaskList[0];
      TaskList.Remove(CurrentTask);
    end;
  finally
    TaskListCS.Release;
  end;

  //- If we got no task, sleep.
  if not assigned(CurrentTask) then Control.Sleep;
  CurrentTask := nil; //- Freeing the task triggers it to be executed, and the taskset notified and freed!
end;

initialization
  TasksEnabled := TRUE;
  TaskList := TList<ITask>.Create(32);
  TaskListCS := TCriticalSection.Create;

finalization
  TasksEnabled := False;
  TaskListCS.Acquire;
  try
    TaskList := nil;
  finally
    TaskListCS := nil;
  end;

end.
