﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwThreading.Thread.Standard;
{$ifdef fpc}{$mode delphiunicode}{$endif}

interface
uses
  cwCollections
, cwThreading
{$ifndef fpc}
  {$ifdef MSWINDOWS}
, cwWin32
  {$else}
, cwPosix
  {$endif}
{$endif}
;

type
  ///  <summary>
  ///    A controller to operate the thread, for internal
  ///    use. Publically IThreadControl is used.
  ///  </summary>
  IInternalThreadControl = interface( IThreadControl )
    ['{688184CE-B089-4956-BA5F-7D4212F19354}']

    ///  <summary>
    ///    Starts the thread running.
    ///  </summary>
    procedure Start;

    procedure WaitForRunningState( const value: boolean );

    ///  <summary>
    ///    Returns true if the thread is hosting the specified
    ///    delegate.
    ///  </summary>
    function IsDelegate( const ThreadDelegate: IThreadDelegate ): boolean;

    ///  <summary>
    ///    Returns true if the thread delegate implements the specified
    ///    interface.
    ///  </summary>
    function DelegateImplements( const InterfaceGUID: TGUID ): boolean;

    ///  <summary>
    ///    Returns a reference to the delegate.
    ///  </summary>
    function getDelegate: IInterface;
  end;

type
  ///- Internal representation of a thread.
  ///- This thread is used by all other parts of the threading system.
  TThread = class( TInterfacedObject, IInternalThreadControl, IThreadControl )
  private
    {$ifdef fpc}
    fHandle          : TThreadID;
    {$else}
    fHandle          : {$ifdef MSWINDOWS} winHandle; {$else} pthread_t; {$endif}
    {$endif}
    fSleepCS         : ISignaledCriticalSection;
    fDelegate        : IThreadDelegate;
    fRunningFlag     : nativeuint;
    fTerminateFlag   : nativeuint;
    fSleepFlag       : nativeuint;
    fHasSleepTimeout : boolean;

  strict private //- IInternalThreadControl -//
    procedure Start;
    procedure WaitForRunningState( const value: boolean );
    function IsDelegate( const ThreadDelegate: IThreadDelegate ): boolean;
    function DelegateImplements( const InterfaceGUID: TGUID ): boolean;
    function getDelegate: IInterface;

  strict private //- IThreadControl -//
    procedure Sleep;
    procedure Wake;
    procedure Terminate;

  protected
    procedure Execute;

  public
    constructor Create( const ThreadDelegate: IThreadDelegate; const SleepTimeoutMilliseconds: uint64 = 0 ); reintroduce;
    destructor Destroy; override;

  end;

implementation
uses
  cwTypes
, cwCollections.Factory
, cwThreading.CriticalSection.Standard
, cwThreading.SignaledCriticalSection.Standard
;


{$ifndef MSWINDOWS}
const
  cDefaultStackSize = 2097152; // 2MB;
//  cDefaultStackSize = 833608; // 8MB;
{$endif}

const
  cFlagFALSE = 0;
  cFlagTRUE  = 1;

type
  TThreadRecord = record
    Thread: TThread;
  end;
  pThreadRecord = ^TThreadRecord;

function InternalHandler( Thread: pointer ): {$ifdef fpc} nativeuint; {$else} {$ifdef MSWINDOWS} uint32; stdcall; {$else} pointer cdecl; {$endif} {$endif}
var
  ptrThreadRecord: pThreadRecord;
begin
  Result := 0;
  ptrThreadRecord := Thread;
  ptrThreadRecord^.Thread.Execute;
  dispose( ptrThreadRecord );
end;

procedure TThread.Execute;
begin
  fRunningFlag := cFlagTRUE;
  try
    //- If there is no delegate, there is nothing to do.
    if not assigned( fDelegate ) then exit;
    //- Make sure we're able to sleep.
    if not assigned( fSleepCS )  then exit;
    //- Signal the delegate that the thread has started up
    if Supports( fDelegate, ILifeCycleDelegate ) then (fDelegate as ILifeCycleDelegate).StartUp;
    //- We're going to repeatedly execute the delegate.
    while true do begin
      fSleepCS.Acquire;
      try
        // Check that we've not been terminated.
        if fTerminateFlag = cFlagTRUE then exit;
        //- If the thread has been instructed to sleep.
        if fSleepFlag = cFlagTRUE then begin
          //- If the delegate supports the IPauseResumeDelegate, let it know
          //- the thread is going to sleep.
          if Supports( fDelegate, IPauseResumeDelegate ) then (fDelegate as IPauseResumeDelegate).Pause;
          //- Sleep the thread.
          fSleepCS.Sleep;
          if fHasSleepTimeout then fSleepFlag := cFlagFalse;
          //- if the sleep flag is cleared, let the delegate know we're awoken.
          if Supports( fDelegate, IPauseResumeDelegate ) then ( fDelegate as IPauseResumeDelegate ).Resume;
          continue; //- Loop back around.
        end;
        //- If we got here, we're neither terminated nor sleeping, run the delegate.
        // Execute the delegate
        fDelegate.Execute( Self );
      finally
        fSleepCS.Release;
      end;
    end;
  finally
    try
      if Supports( fDelegate, ILifeCycleDelegate ) then (fDelegate as ILifeCycleDelegate).ShutDown;
    finally
      fRunningFlag := cFLAGFalse;
      fDelegate := nil;
    end;
  end;
end;

function TThread.getDelegate: IInterface;
begin
  Result := fDelegate;
end;

function TThread.IsDelegate( const ThreadDelegate: IThreadDelegate ): boolean;
begin
  Result := fDelegate = ThreadDelegate;
end;

procedure TThread.Sleep;
begin
  fSleepFlag := cFlagTRUE;
end;

procedure TThread.Wake;
begin
  fSleepFlag := cFlagFALSE;
  fSleepCS.Wake;
end;

procedure TThread.Terminate;
begin
  if fRunningFlag   = cFlagFALSE then exit;
  if fTerminateFlag = cFlagTRUE then exit;
  fTerminateFlag := cFlagTRUE;
  fSleepCS.Wake;
end;

procedure TThread.WaitForRunningState( const value: boolean );
var
  WaitFor: nativeuint;
begin
  //- What are we waiting for?
  if Value then begin
    WaitFor := cFlagTRUE;
  end else begin
    WaitFor := cFlagFALSE;
  end;
  //- Already in this state?
  if fRunningFlag = WaitFor then exit;
  //- Begin waiting.
  while TRUE do begin
    fSleepCS.Wake;
    cwThreading.Sleep( 1 );
    if fRunningFlag = WaitFor then exit;
  end;
end;

procedure TThread.Start;
var
  {$ifndef fpc}
    {$ifdef MSWINDOWS}
      ThreadID: TThreadID;
    {$else}
      Attr: pthread_attr_t;
    {$endif}
  {$endif}
  ptrThreadRecord: pThreadRecord;
begin
  if fRunningFlag = cFlagTRUE then exit;
  //- Create a thread record and store it's reference.
  new( ptrThreadRecord );
  ptrThreadRecord^.Thread := Self;
  {$ifdef fpc}
    fHandle := BeginThread( @InternalHandler, ptrThreadRecord );
  {$else}
    {$ifdef MSWINDOWS}
    fHandle := CreateThread( nil, 0, @InternalHandler, ptrThreadRecord, 0, ThreadID );
    if fHandle = 0 then TStatus( stThreadCreateFailed ).Raize;
    {$else}
    if pthread_attr_init        ( attr                                             ) <> 0 then TStatus( stThreadAttrInitFailed ).Raize( [GetLastError().AsString] );
    if pthread_attr_setstacksize( attr, cDefaultStackSize                          ) <> 0 then TStatus( stThreadStackSetFailed ).Raize( [GetLastError().AsString] );
    if pthread_create           ( fHandle, attr, @InternalHandler, ptrThreadRecord ) <> 0 then TStatus( stThreadCreateFailed   ).Raize( [GetLastError().AsString] );
    if pthread_attr_destroy     ( attr                                             ) <> 0 then TStatus( stThreadAttrFnitFailed ).Raize( [GetLastError().AsString] );
    {$endif}
  {$endif}
end;


constructor TThread.Create( const ThreadDelegate: IThreadDelegate; const SleepTimeoutMilliseconds: uint64 = 0 );
begin
  inherited Create;
  fSleepCS             := TSignaledCriticalSection.Create( SleepTimeoutMilliseconds );
  fSleepFlag           := cFlagFALSE;
  fRunningFlag         := cFlagFALSE;
  fTerminateFlag       := cFlagFALSE;
  fDelegate            := ThreadDelegate;
  fHasSleepTimeout     := (SleepTimeoutMilliseconds <> 0);
end;

function TThread.DelegateImplements(const InterfaceGUID: TGUID): boolean;
begin
  Result := Supports( fDelegate, InterfaceGUID );
end;

destructor TThread.Destroy;
begin
  WaitForRunningState( FALSE );
  {$ifdef fpc}
    CloseThread( fHandle );
  {$else}
    {$ifdef MSWINDOWS}
    if fHandle<>0 then CloseHandle(fHandle);
    {$else}
    if pthread_join( fHandle, nil ) <> 0 then TStatus( stThreadJoinFailed ).Raize( [ GetLastError().AsString ] );
    {$endif}
  {$endif}
  fSleepCS := nil;
  inherited Destroy;
end;


end.

