
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwThreading.ThreadPool.Standard;
{$ifdef fpc}{$mode delphiunicode}{$endif}

interface
uses
  cwCollections
, cwThreading
, cwThreading.Thread.Standard
;

type
  TThreadPool = class
  private
    fThreads: IList<IInternalThreadControl>;
    fThreadsCS: ICriticalSection;
  private
    procedure TerminateAll;
  public
    function LocateByDelegateImplements( const InterfaceGUID: TGUID ): IReadOnlyList<IThreadControl>;
    function LocateByDelegate( const ThreadDelegate: IThreadDelegate ): IThreadControl;
    function LocateDelegateByImplements(const InterfaceGUID: TGUID): IReadOnlyList<IInterface>;
    procedure Add( const Delegate: IThreadDelegate; const SleepTimeout: nativeuint = 0 );
  public
    constructor Create;
    destructor Destroy; override;
  end;

implementation
uses
  cwTypes
, cwCollections.Factory
, cwThreading.CriticalSection.Standard
;

function TThreadPool.LocateByDelegateImplements( const InterfaceGUID: TGUID ): IReadOnlyList<IThreadControl>;
var
  idx: nativeuint;
  List: IList<IThreadControl>;
begin
  List := TList<IThreadControl>.Create;
  Result := List.getAsReadOnly;
  fThreadsCS.Acquire;
  try
    if fThreads.Count = 0 then exit;
    for idx := 0 to pred( fThreads.Count ) do begin
      if ( fThreads[ idx ] as IInternalThreadControl ).DelegateImplements( InterfaceGUID ) then begin
        List.Add( ( fThreads[ idx ] as IThreadControl ) );
      end;
    end;
  finally
    fThreadsCS.Release;
  end;
end;

function TThreadPool.LocateDelegateByImplements( const InterfaceGUID: TGUID ): IReadOnlyList<IInterface>;
var
  idx: nativeuint;
  List: IList<IInterface>;
begin
  List := TList<IInterface>.Create;
  Result := List.getAsReadOnly;
  if not assigned(fThreadsCS) then exit;
  fThreadsCS.Acquire;
  try
    for idx := 0 to pred( fThreads.Count ) do begin
      if ( fThreads[ idx ] as IInternalThreadControl ).DelegateImplements( InterfaceGUID ) then begin
        List.Add( ( fThreads[ idx ] as IInternalThreadControl ).getDelegate );
      end;
    end;
  finally
    fThreadsCS.Release;
  end;
end;

procedure TThreadPool.Add( const Delegate: IThreadDelegate; const SleepTimeout: nativeuint );
var
  ThreadControl: IInternalThreadControl;
begin
  ThreadControl := TThread.Create( Delegate, SleepTimeout );
  fThreadsCS.Acquire;
  try
    fThreads.Add( ThreadControl );
    ThreadControl.Start;
  finally
    fThreadsCS.Release;
  end;
end;

function TThreadPool.LocateByDelegate( const ThreadDelegate: IThreadDelegate ): IThreadControl;
var
  idx: nativeuint;
begin
  Result := nil;
  fThreadsCS.Acquire;
  try
    if fThreads.Count=0 then exit;
    for idx := 0 to pred( fThreads.Count ) do begin
      if not Supports( fThreads[ idx ], IInternalThreadControl ) then continue;
      if ( fThreads[ idx ] as IInternalThreadControl ).IsDelegate( ThreadDelegate ) then begin
        Result := fThreads[ idx ] as IThreadControl;
        exit;
      end;
    end;
  finally
    fThreadsCS.Release;
  end;
end;

procedure TThreadPool.TerminateAll;
var
  idx: nativeuint;
begin
  if fThreads.Count = 0 then exit;
  for idx := pred( fThreads.Count ) downto 0 do begin
    fThreads[ idx ].Terminate;
    fThreads[ idx ].WaitForRunningState( FALSE );
    fThreads.RemoveItem( idx );
  end;
end;

constructor TThreadPool.Create;
begin
  inherited Create;
  fThreads   := TList<IInternalThreadControl>.Create;
  fThreadsCS := TCriticalSection.Create;
end;

destructor TThreadPool.Destroy;
begin
  TerminateAll;
  fThreadsCS := nil;
  fThreads   := nil;
  inherited Destroy;
end;

end.

