﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
unit cwThreading.Threads.Standard;
{$ifdef fpc}
  {$mode delphiunicode}
  {$MODESWITCH NESTEDPROCVARS}
{$endif}
interface
uses
  cwTypes
, cwCollections
, cwThreading
, cwThreading.ThreadPool.Standard
;
type
  TThreads = class( TInterfacedObject, IThreads )
  private
    fThreadPool: TThreadPool;
  private
    procedure InternalForLoopWait( const Delegate: TLoopDelegate; const TotalWork: nativeuint; const Parameters: array of TVariant );
    function DivideWorkload( const Task: TLoopDelegate; const TotalWork: nativeuint; const Parameters: array of TVariant ): IList<ITaskDelegate>;
    procedure TerminateTaskThread( const Control: IThreadControl );
    procedure WakeTaskThread( const Control: IThreadControl );
    function getTaskThreads: IReadOnlyList<IThreadControl>;
  strict private //- IThreads -//
    function getTaskThreadCount: nativeuint;
    procedure setTaskThreadCount( const value: nativeuint );
    function ExecuteThread ( const ThreadDelegate: TInterfacedObject ): IThreadDelegate;
    function ThreadedMethod( const ThreadMethod: TThreadMethod ): IThreadDelegate;
    function ControlThread( const ThreadDelegate: IThreadDelegate ): IThreadControl;
    procedure Schedule( const ScheduledTaskDelegate: TInterfacedObject; const IntervalMilliseconds: nativeuint );
    function ExecutePromise( const Promise: TInterfacedObject ): IInterface;
    function ResolvePromise( const Iface: TGUID ): IReadOnlyList<IInterface>;
    procedure TerminatePromise( const Promise: IInterface );
    procedure Tasks( const Tasks: IReadOnlyList<ITaskDelegate>; const NotifyDone: TNotifyDelegate );
    procedure ForLoop( const TotalWork: nativeuint; const Task: TLoopDelegate; const Parameters: array of TVariant; const NotifyDone: TNotifyDelegate ); overload;
    procedure ForLoop( const TotalWork: nativeuint; const Task: TLoopDelegate; const Parameters: array of TVariant; const Wait: boolean = FALSE ); overload;

  public
    constructor Create;
    destructor Destroy; override;
  end;
implementation
uses
  cwCollections.Factory
, cwThreading.Factory
, cwThreading.Promise.Custom
, cwThreading.PromiseThread.Standard
, cwThreading.TaskThread.Standard
, cwThreading.ForTaskDelegate.Standard
, cwThreading.TaskSchedule.Standard
, cwThreading.ThreadedMethod.Standard
;

function TThreads.ExecuteThread( const ThreadDelegate: TInterfacedObject ): IThreadDelegate;
var
  ThreadDelegateI: IInterface;
begin
  ThreadDelegateI := ThreadDelegate as IInterface;
  if not Supports(ThreadDelegate, IThreadDelegate) then TStatus( stObjectMustSupportInterface ).Raize( ['ThreadDelegate','IThreadDelegate'] );
  fThreadPool.Add( (ThreadDelegateI as IThreadDelegate) );
  Result := ThreadDelegateI as IThreadDelegate;
end;

procedure TThreads.ForLoop(const TotalWork: nativeuint; const Task: TLoopDelegate; const Parameters: array of TVariant; const NotifyDone: TNotifyDelegate);
begin
  Tasks( DivideWorkLoad( Task, TotalWork, Parameters ), NotifyDone );
end;

procedure TThreads.ForLoop(const TotalWork: nativeuint; const Task: TLoopDelegate; const Parameters: array of TVariant; const Wait: boolean);
begin
  if Wait then begin
    InternalForLoopWait( Task, TotalWork, Parameters );
  end else begin
    Tasks( DivideWorkLoad( Task, TotalWork, Parameters ), TNotifyDelegate.Null );
  end;
end;

function TThreads.ThreadedMethod(const ThreadMethod: TThreadMethod): IThreadDelegate;
begin
  Result := TThreadedMethod.Create( ThreadMethod );
  fThreadPool.Add( Result );
end;

procedure TThreads.InternalForLoopWait( const Delegate: TLoopDelegate; const TotalWork: nativeuint; const Parameters: array of TVariant );
var
  WaitCS: ISignaledCriticalSection;
  WaitCondition: nativeuint;
  {$ifdef fpc}
  procedure ReleaseWait;
  begin
    WaitCS.Acquire;
    try
      WaitCondition := 1;
    finally
      WaitCS.Release;
    end;
    WaitCS.Wake;
  end;
  {$endif}
begin
  WaitCS := TSignaledCriticalSection.Create;
  try
    WaitCondition   := 0;
    {$ifdef fpc}
    Tasks( DivideWorkLoad( Delegate, TotalWork, Parameters ), TNotifyDelegate( TNotifyNestedCallback( ReleaseWait ) ) );
    {$else}
    Tasks( DivideWorkLoad( Delegate, TotalWork, Parameters ),
        procedure
        begin
          WaitCS.Acquire;
          try
            WaitCondition := 1;
          finally
            WaitCS.Release;
          end;
          WaitCS.Wake;
        end
    );
    {$endif}
    WaitCS.Acquire;
    try
      repeat
        if WaitCondition = 0 then begin
          WaitCS.Sleep;
        end;
      until WaitCondition<>0;
    finally
      WaitCS.Release;
    end;
  finally
    WaitCS := nil;
  end;
end;

function TThreads.ResolvePromise( const Iface: TGUID ): IReadOnlyList<IInterface>;
var
  idx: nativeuint;
  idy: nativeuint;
  PromiseDelegates: IReadOnlyList<IInterface>;
  PromiseDelegate: IPromiseThread;
  Promises: IList<IInterface>;
  Tmp: IReadOnlyList<IInterface>;
begin
  Promises := TList<IInterface>.Create;
  try
    PromiseDelegates := fThreadPool.LocateDelegateByImplements(IPromiseThread);
    if PromiseDelegates.Count>0 then begin
      for idx := 0 to pred(PromiseDelegates.Count) do begin
        PromiseDelegate := (PromiseDelegates[idx] as IPromiseThread);
        Tmp := PromiseDelegate.Resolve(IFace);
        if not assigned(Tmp) then continue;
        if Tmp.Count=0 then continue;
        for idy := 0 to pred(Tmp.Count) do Promises.Add(Tmp[idy]);
      end;
    end;
  finally
    Result := Promises.getAsReadOnly;
  end;
end;

procedure TThreads.Schedule(const ScheduledTaskDelegate: TInterfacedObject; const IntervalMilliseconds: nativeuint);
var
  I: IReadOnlyList<IThreadControl>;
  Task: IInterface;
  Scheduler: ITaskSchedule;
begin
  Task := ScheduledTaskDelegate;
  if (not Supports(Task,ITaskDelegate)) and
     (not Supports(Task,IScheduledTaskDelegate)) then begin
    TStatus(stObjectMustSupportEitherInterface).Raize( ['ScheduledTaskDelegate','ITaskDelegate','IScheduledTaskDelegate'] );
  end;
  I := fThreadPool.LocateByDelegateImplements(ITaskSchedule);
  if (not assigned(I)) or (I.Count=0) then begin
    Scheduler := TTaskSchedule.Create;
    fThreadPool.Add( (Scheduler as IThreadDelegate), 1 );
  end else begin
    Scheduler := (I[0] as ITaskSchedule);
  end;
  Scheduler.AddTask(Task,IntervalMilliseconds);
end;

constructor TThreads.Create;
begin
  inherited Create;
  fThreadPool := TThreadPool.Create;
end;

destructor TThreads.Destroy;
begin
  fThreadPool.Free;
  inherited Destroy;
end;

function TThreads.getTaskThreadCount: nativeuint;
begin
  Result := getTaskThreads.Count;
end;

procedure TThreads.setTaskThreadCount(const value: nativeuint);
var
  idx: nativeuint;
  TaskThreads: IReadOnlyList<IThreadControl>;
  Control: IThreadControl;
begin
  TaskThreads := getTaskThreads;
  if TaskThreads.Count=Value then exit;
  if TaskThreads.Count>0 then begin
    for Control in TaskThreads do TerminateTaskThread( Control );
  end;
  for idx := 0 to pred(value) do begin
    ExecuteThread(TTaskThread.Create);
  end;
end;

procedure TThreads.Tasks( const Tasks: IReadOnlyList<ITaskDelegate>; const NotifyDone: TNotifyDelegate );
var
  Control: IThreadControl;
begin
  if getTaskThreads.Count=0 then TStatus(stNoTaskThreads).Raize;
  TTaskThread.AddTaskSet( Tasks, TNotifyDelegate( NotifyDone ) );
  for Control in getTaskThreads do WakeTaskThread( Control );
end;

function TThreads.getTaskThreads: IReadOnlyList<IThreadControl>;
begin
  Result := fThreadPool.LocateByDelegateImplements(ITaskThread);
end;

procedure TThreads.TerminatePromise(const Promise: IInterface);
begin
  TPromiseThread.TerminatePromise(Promise);
end;

function TThreads.ControlThread( const ThreadDelegate: IThreadDelegate ): IThreadControl;
begin
  Result := fThreadPool.LocateByDelegate( ThreadDelegate );
end;

procedure TThreads.TerminateTaskThread( const Control: IThreadControl );
begin
  Control.Terminate;
end;

procedure TThreads.WakeTaskThread( const Control: IThreadControl );
begin
  Control.Wake;
end;

function TThreads.DivideWorkload( const Task: TLoopDelegate; const TotalWork: nativeuint; const Parameters: array of TVariant ): IList<ITaskDelegate>;
var
  idx: nativeuint;
  TaskThreadCount: nativeuint;
  division: nativeuint;
  remainder: nativeuint;
  start: nativeuint;
  stop: nativeuint;
  top: nativeuint;
  NewTask: ITaskDelegate;
begin
  Result := nil;
  if TotalWork=0 then exit;
  TaskThreadCount := getTaskThreads.Count;
  if TaskThreadCount=0 then TStatus(stNoTaskThreads).Raize;
  Result := TList<ITaskDelegate>.Create;
  if TotalWork<TaskThreadCount then begin
    for idx := 0 to pred(TotalWork) do begin
      NewTask := TForTaskDelegate.Create( idx, idx, Task, Parameters );
      Result.Add( NewTask );
    end;
  end else begin
    Division := (TotalWork div TaskThreadCount);
    Remainder := (TotalWork-(Division*TaskThreadCount));
    for idx := 0 to pred(TaskThreadCount) do begin
      Start   := (idx*Division);
      Stop    := pred(Start+Division);
      NewTask := TForTaskDelegate.Create( Start, Stop, Task, Parameters );
      Result.Add( NewTask );
    end;
    if Remainder>0 then begin
      Top := (Division*TaskThreadCount);
      for idx := 0 to pred(Remainder) do begin
        NewTask := TForTaskDelegate.Create( (Top + idx), (Top + idx), Task, Parameters );
        Result.Add( NewTask );
      end;
    end;
  end;
end;

function TThreads.ExecutePromise( const Promise: TInterfacedObject ): IInterface;
var
  PromiseThread: IThreadDelegate;
  PromiseI: IInterface;
begin
  Result := nil;
  PromiseI := Promise as IInterface;
  if not Supports(Promise, IPromiseLifeCycle) then TStatus( stObjectMustDescend ).Raize( ['Promise','cwThreading.Factory.TPromise'] );
  PromiseThread := TPromiseThread.Create( PromiseI as IPromiseLifeCycle );
  fThreadPool.Add( PromiseThread );
  Result := PromiseI;
end;

end.
