{$ifdef license}
(*
  Copyright 2020 ChapmanWorld LLC ( https://chapmanworld.com )
  Redistribution and use in source and binary forms, with or without modification,
  are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation and/or
     other materials provided with the distribution.

  3. Neither the name of the copyright holder nor the names of its contributors may be
     used to endorse or promote products derived from this software without specific prior
     written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)
{$endif}
unit cwVirtualMachine;
{$ifdef fpc}{$mode delphiunicode}{$endif}

interface
uses
  sysutils
;

{$region ' Exceptions '}

type
  ///  <summary>
  ///    Raised when attempting to read/write a bytecode buffer which has no
  ///    backing virtual memory. Also raised if attempting to run a virtual CPU
  ///    without virtual memory assigned.
  ///  </summary>
  EVirtualMemoryUnassigned = class( Exception )
    constructor Create; reintroduce;
  end;

{$endregion}

{$region ' Virtual CPU States'}

type
  ///  <summary>
  ///    Returned by IVirtualCPU.Run() method to indicate
  ///    the state of the virtual CPU upon exit of
  ///    the running program.
  ///  </summary>
  TVirtualCPUState = (

    ///  <summary>
    ///    The program successfully ran and came to an
    ///    appropriate HALT condition, such as encountering
    ///    a HALT instruction if the target CPU supports it.
    ///  </summary>
    Success,

    ///  <summary>
    ///    This is an error state caused by the virtual CPU encountering
    ///    an instruction which it was unable to decode.
    ///  </summary>
    InvalidOpCode,

    ///  <summary>
    ///    This is an error state caused by the virtual CPU attempting to
    ///    decode instructions outside the bounds of the virtual memory
    ///    space available to it.
    ///  </summary>
    MemoryOverrun,

    ///  <summary>
    ///    An unknown error condition occurred. <br/>
    ///    The details of the error condition may be reported by the
    ///    virtual CPU depending on it's implementation. i.e. There
    ///    may be a flags register containing an indication of the
    ///    cause. Check the documentation for the CPU implementation.
    ///  </summary>
    Unknown
  );



{$endregion}

{$region ' Supporting Types '}

type

  /// <summary>
  ///   Data-type for returning arrays of bytes which is compatible
  ///   with dynamic array of bytes parameters.
  /// <summary>
  TArrayOfByte = array of byte;

{$endregion}

{$region ' IVirtualCPU '}

type
  /// <exclude/> // forward declaration for TOnInstructionCycle.
  IVirtualCPU = interface;
  /// <exclude/> // forward declaration for TOnInstructionCycle.
  IVirtualMemory = interface;

  ///  <summary>
  ///    An event callback which will be fired for every instruction cycle of the
  ///    virtual CPU. <br/>
  ///  </summary>
  {$ifdef fpc}
  TOnInstructionCycle = procedure( const VirtualCPU: IVirtualCPU; const VirtualMemory: IVirtualMemory ) of object;
  {$else}
  TOnInstructionCycle = reference to procedure( const VirtualCPU: IVirtualCPU; const VirtualMemory: IVirtualMemory );
  {$endif}

  /// <summary>
  ///   This interface enables the behavior of IVirtualMachine implementations
  ///   to be altered by providing different virtual CPU implementations. <br />
  /// </summary>
  /// <remarks>
  ///   You should not need to work with this interface directly, except
  ///   whenimplementing your own virtual CPU. Access via theIVirtualMachine
  ///   interface for most use cases. <br />
  /// </remarks>
  IVirtualCPU = interface
    ['{38F2CC5B-60AC-48CE-8F4F-DDDE20E45C08}']


    ///  <summary>
    ///    Sets a reference to the call-back procedure which should be
    ///    fired on every instruction cycle if set. <br/>
    ///    This is a setter for the OnInstructionCycle property. <br/>
    ///    Note: Running a callback for every instruction cycle is
    ///    costly in terms of performance, and should therefore be
    ///    done only as a last resort or while debugging. <br/>
    ///    If the virtual CPU implementation supports a BREAK
    ///    instruction, consider detecting the necessary
    ///    conditions in byte-code and issuing a break instead.
    ///  </summary>
    procedure setOnInstructionCycle( const Handler: TOnInstructionCycle );

    ///  <summary>
    ///    Gets a reference to the call-back procedure which should be
    ///    fired on every instruction cycle if set. <br/>
    ///    This is the getter for the OnInstructionCycle property. <br/>
    ///    Note: See setOnInstructionCycle() for performance considerations.
    ///  </summary>
    function getOnInstructionCycle: TOnInstructionCycle;

    ///  <summary>
    ///    <para>
    ///      Resets the CPU, restoring initial state. <br />
    ///      Entrypoint should be set to a position within the
    ///      virutal memory buffer at which the cpu should begin
    ///      executing byte-code instructions.
    ///    </para>
    ///  </summary>
    procedure Reset( const EntryPoint: nativeuint );

    ///  <summary>
    ///    Starts the virtual CPU running. <br/>
    ///    The virtual CPU will continue to run until
    ///    a breaking condition, such as program halt or
    ///    breakpoint is encountered, or until an error
    ///    state occurs.
    ///   </summary>
    function Run: TVirtualCPUState;

    ///  <summary>
    ///    Get or Set a reference to a call-back procedure which should be
    ///    fired for every instruction executed. <br/>
    ///    See setOnInstructionCycle() for performance considerations.
    ///  </summary>
    property OnInstructionCycle: TOnInstructionCycle read getOnInstructionCycle write setOnInstructionCycle;
  end;

{$endregion}

{$region ' IVirtualMemory '}

  ///  <summary>
  ///    An implementation of IVirtualMemory represents the memory space
  ///    for use with a virtual CPU.
  ///    Each implementation of IVirtualMemory will differ depending on
  ///    the requirements of the virtual CPU implementation. <br/>
  ///  </summary>
  IVirtualMemory = interface
    ['{6066523F-9DAB-419B-8787-0F316F6D8DEE}']

    ///  <summary>
    ///    Returns a vanilla pointer to the memory buffer. <br/>
    ///    Getter for DataPtr property.
    ///  </summary>
    function getDataPtr( const ByteIndex: nativeuint ): pbyte;

    ///  <summary>
    ///    Returns the size, in bytes, of the virtual memory buffer. <br/>
    ///    Getter for DataSize property.
    ///  </summary>
    function getDataSize: nativeuint;

    ///  <summary>
    ///    Alter the size of allocated memory. <br/>
    ///    Caution: This method may be destructive - see DataSize property. <br/>
    ///    Setter for DataSize property.
    ///  </summary>
    procedure setDataSize( const Value: nativeuint );

    ///  <summary>
    ///    Get or Set the size of the virtual memory buffer in bytes. <br/>
    ///    When setting, the standard implementation attempts to retain any
    ///    existing data in the buffer, up-to the new size in bytes. This may
    ///    vary with non-standard implementations. Check the implementation
    ///    documentation / comments to see if altering the size of the buffer
    ///    is destructive.
    ///  </summary>
    property DataSize: nativeuint read getDataSize write setDataSize;

    ///  <summary>
    ///    Returns a pointer to a byte as specified by 'ByteIndex'.
    ///  </summary>
    property DataPtr[ const ByteIndex: nativeuint ]: pbyte read getDataPtr; default;
  end;

{$endregion}

{$region ' IBytecode '}

type

  ///  <summary>
  ///    An implementation of IBytecode represents a window into an IVirtualMemory
  ///    instance in order to edit byte-code instructions. <br/>
  ///  </summary>
  IBytecode = interface
    ['{0BC9900E-9BE3-4283-BEC1-C7FEF757AEBF}']

    ///  <summary>
    ///    Returns a reference to the virtual memory being edited
    ///    through this byte-code window. <br/>
    ///    Getter for VirtualMemory property.
    ///  </summary>
    function getVirtualMemory: IVirtualMemory;

    ///  <summary>
    ///    Returns the location of the byte-code window as an offset within
    ///    virtual memory. <br/>
    ///    Getter for the Cursor property.
    ///  </summary>
    function getCursor: nativeuint;

    ///  <summary>
    ///    Sets the location of the byte-code window as an offset within
    ///    virtual memory. <br/>
    ///    If an attempt is made to position the cursor outside the
    ///    virutal memory space, the cursor will simply be positioned
    ///    at the last byte of virtual memory. <br/>
    ///    Setter for the Cursor property.
    ///  </summary>
    procedure setCursor( const value: nativeuint );

    ///  <summary>
    ///    Writes a number of bytes to virtual memory buffer at the location
    ///    specified by the cursor property. <br/>
    ///    If the top of the virtual memory buffer is reached while there
    ///    are still bytes remaining to be written, writing will simply
    ///    stop to prevent buffer overrun, and this method will return
    ///    the number of bytes actually written to the buffer.
    ///  </summary>
    function Write( const Bytes: array of uint8 ): nativeuint;

    ///  <summary>
    ///    Appends the bytes provided in the bytes array to the
    ///    byte-code buffer at the location specified by the cursor
    ///    property, and progresses the cursor beyond the written bytes. <br/>
    ///    This method will expand the virutal buffer size if required.
    ///  </summary>
    procedure Append( const Bytes: array of uint8 );

    ///  <summary>
    ///    A reference to the virtual memory assigned to this byte-code window. <br/>
    ///  </summary>
    property VirtualMemory: IVirtualMemory read getVirtualMemory;

    ///  <summary>
    ///    Get/Set the cursor offset within the IVirtualMemory buffer. <br/>
    ///    When setting the cursor beyond the bounds of the virtual memory buffer,
    ///    the property will simply be set to point at the last byte in the buffer.
    ///    Care must be taken to check that the cursor was actually set to the
    ///    intended location if this can not be assured before setting.
    ///  </summary>
    property Cursor: nativeuint read getCursor write setCursor;

  end;

{$endregion}


implementation

constructor EVirtualMemoryUnassigned.Create;
begin
  inherited Create('Bytecode Window or VirtualCPU does not have virtual memory assigned.');
end;

end.
