{$ifdef license}
(*
  Copyright 2020 ChapmanWorld LLC ( https://chapmanworld.com )
  Redistribution and use in source and binary forms, with or without modification,
  are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation and/or
     other materials provided with the distribution.

  3. Neither the name of the copyright holder nor the names of its contributors may be
     used to endorse or promote products derived from this software without specific prior
     written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)
{$endif}
unit cwVirtualMachine.VirtualCPU.Chappie;
{$ifdef fpc}{$mode delphiunicode}{$endif}

interface
uses
  cwTypes
, cwVirtualMachine
, cwVirtualMachine.Chappie
;

type
  ///  <summary>
  ///    An enumeration of the op-codes available to the chappie CPU.
  ///  </summary>
  ///  Note: {$Z4} directive to set minimum enum size to four bytes.
  ///        A check will be performed during initialization to ensure that
  ///        this type does not exceed four bytes, making TOpCode compatible with uint32;
  {$Z4}
  TOpcode = (
    opHalt  = $00,
    opNop   = $01,
    opAlert = $02,
    opLoad  = $03,
    opSave  = $04,
    opAdd   = $05
  );


{$region ' CPU State '}

type
  // Represents the state of a chappie CPU.
  TChappieState = record

    (* Reference to our virtual memory space. *)
    Memory: IVirtualMemory;

    (* The HALT instruction sets this to false. *)
    ContinueRunning: boolean;

    (* Actual CPU Registeres *)
    ProgramCounter  : TRegisterValue;
    Accumulator     : TRegisterValue;

  end;
  pChappieState = ^TChappieState;

{$endregion}

type
  TChappieCPU = class( TInterfacedObject, IVirtualCPU )
  private
    fState: TChappieState;
    fOnInstructionCycle: TOnInstructionCycle;
  strict private //- IVirtualCPU -//
    procedure setOnInstructionCycle( const Handler: TOnInstructionCycle );
    function getOnInstructionCycle: TOnInstructionCycle;
    procedure Reset( const EntryPoint: nativeuint );
    function Run: TVirtualCPUState;
  public
    constructor Create( const Memory: IVirtualMemory ); reintroduce;
    destructor Destroy; override;
  end;

implementation

{$region ' Chappie Instruction Handlers'}

procedure HandleNop( var State: TChappieState );
begin
  Writeln('Nop');
end;

procedure HandleHalt( var State: TChappieState );
begin
  State.ContinueRunning := False;
  Writeln('HALT');
end;

procedure HandleAlert( var State: TChappieState );
begin
  Writeln('Alert!!!!!');
end;

procedure HandleLoad( var State: TChappieState );
begin
  State.Accumulator := pRegisterValue( State.Memory[ State.ProgramCounter ] )^;
  inc( State.ProgramCounter, Sizeof( TChappieOperand ) );
end;

procedure HandleSave( var State: TChappieState );
begin
  pRegisterValue( State.Memory[ State.ProgramCounter ] )^ := State.Accumulator;
  inc( State.ProgramCounter, Sizeof( TChappieOperand ) );
end;

procedure HandleAdd( var State: TChappieState );
begin
  State.Accumulator := State.Accumulator + pRegisterValue( State.Memory[ State.ProgramCounter ] )^;
  inc( State.ProgramCounter, Sizeof( TChappieOperand ) );
  Writeln( State.Accumulator.AsString );
end;

{$endregion}

{$region ' Chappie Instruction Set'}

type
  // A callback function to handle execution of an instruction.
  TVMInstructionHandler = procedure( var State: TChappieState );

const
  //- Our instruction set defined as an array of named handlers.
  cInstructionSet: array[ Low(TOpCode) .. High(TOpCode) ] of TVMInstructionHandler = (
    HandleHalt,
    HandleNop,
    HandleAlert,
    HandleLoad,
    HandleSave,
    HandleAdd
  );

{$endregion}

{$region ' TChappieCPU class implementation'}

function TChappieCPU.Run: TVirtualCPUState;
type
  pOpCode = ^TOpCode;
var
  Handler: TVMInstructionHandler;
begin
  //- In case of unexpected fault, set return status accordingly.
  Result := TVirtualCPUState.Unknown;

  fState.ContinueRunning := TRUE;
  repeat

    //- Do not allow overrun of byte-code
    if fState.ProgramCounter >= fState.Memory.DataSize then begin
      Result := TVirtualCPUState.MemoryOverrun;
      exit;
    end;

    //- Check for valid op-code
    if pOpCode( nativeuint( fState.Memory[ 0 ] ) + fState.ProgramCounter )^ > High(TOpCode) then begin
      Result := TVirtualCPUState.InvalidOpCode;
      exit;
    end;

    //- Fetch handler
    Handler := cInstructionSet[ pOpCode( nativeuint( fState.Memory[ 0 ] ) + fState.ProgramCounter )^ ];

    //- Increment program counter
    fState.ProgramCounter := fState.ProgramCounter + Sizeof(TOpCode);

    //- Execute Instruction;
    try
      try
        Handler( fState );
      except exit;
      end;
    finally
      //- If there is an OnInstructionCycle call-back, call it.
      if assigned(fOnInstructionCycle) then begin
        fOnInstructionCycle( Self, fState.Memory );
      end;
    end;

  until not fState.ContinueRunning;

  //- If we didn't exit, we're able to resume operations.
  Result := Success;
end;

procedure TChappieCPU.setOnInstructionCycle(const Handler: TOnInstructionCycle);
begin
  fOnInstructionCycle := Handler;
end;

constructor TChappieCPU.Create( const Memory: IVirtualMemory );
begin
  inherited Create;
  fState.Memory := Memory;
  fOnInstructionCycle := nil;
  Reset( 0 );
end;

destructor TChappieCPU.Destroy;
begin
  fState.Memory := nil;
  inherited Destroy;
end;

function TChappieCPU.getOnInstructionCycle: TOnInstructionCycle;
begin
  Result := fOnInstructionCycle;
end;

procedure TChappieCPU.Reset(const EntryPoint: nativeuint);
begin
  if not assigned(fState.Memory) then begin
    fState.ProgramCounter := 0;
    exit;
  end;
  fState.ProgramCounter := Entrypoint;
end;

{$endregion}

end.
