{$ifdef license}
(*
  Copyright 2020 ChapmanWorld LLC ( https://chapmanworld.com )
  Redistribution and use in source and binary forms, with or without modification,
  are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation and/or
     other materials provided with the distribution.

  3. Neither the name of the copyright holder nor the names of its contributors may be
     used to endorse or promote products derived from this software without specific prior
     written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)
{$endif}
program Hello6502;
{$ifdef fpc}{$mode delphiunicode}{$H+}{$endif}
uses
  cwTypes
, cwVirtualMachine
, cwVirtualMachine.Factory
, cwVirtualMachine.Mos6502
;


type
  (* An example of writing a simple (incredibly simple) debugger for the
     6502 implementation of IVirtualCPU *)
  T6502Example = class
  private
    fMemory: I6502VirtualMemory;
    fCPU: I6502CPU;
  private
    procedure Prompt( const Message: string );
    procedure WriteRegisters;
    procedure HandleInstructionCycle( const VirtualCPU: IVirtualCPU; const VirtualMemory: IVirtualMemory );
  public
    procedure Run;
  end;

procedure T6502Example.Prompt( const Message: string );
begin
  Writeln(Message);
  Readln;
end;

procedure T6502Example.WriteRegisters;
begin
  Writeln( '   PC  SR AC XR YT SP - ChapmanWorld Rocks!');
  Writeln( '; '+
           fCPU.PC.AsHex+' '+
           fCPU.SR.AsHex+' '+
           fCPU.A.AsHex+' '+
           fCPU.X.AsHex+' '+
           fCPU.Y.AsHex+' '+
           fCPU.SP.AsHex );
end;

procedure T6502Example.HandleInstructionCycle( const VirtualCPU: IVirtualCPU; const VirtualMemory: IVirtualMemory );
begin
  WriteRegisters;
  Prompt('Press [return] to continue.');
end;

procedure T6502Example.Run;
var
  ByteCode: I6502Bytecode;
begin
  //- Create some memory for our virtual machine to work with.
  //- Defaults to the minimum of 64KB
  fMemory := T6502VirtualMemory.Create;

  //- Create an instruction code window into memory, specific to the 6502
  ByteCode := T6502Bytecode.Create( fMemory );
  ByteCode.Cursor := 0;

  //- Create a program which adds $02 to $05
  ByteCode.LDA_imm( $05 );
  ByteCode.ADC_imm( $02 );
  ByteCode.BRK;

  //- Create an instance of the 6502 to do the work.
  fCPU := T6502CPU.Create( fMemory );

  //- TRUE is the default, setting here for the sake of this
  //- comment! The 6502 processor does not have a HALT instruction
  //- because the 6502, like most physical processors, does not
  //- expect to stop running until power is discontinued.
  //- The 6502 does however have a BRK (break) instruction which
  //- is intended to be used with on-device debugging tools.
  //- The BRK instruction is used to interrupt execution of
  //- program instructions, and pass execution to the debugging
  //- code.
  //- In the case of our emulator. when BRKTerminates is set to TRUE
  //- the BRK instruction behaves as a HALT instruction, terminating
  //- execution of our emulated 6502 (returning from the Run() method).
  fCPU.BRKTerminates := True;

  //- Set an event handler to hook each instruction for debugging.
  fCPU.OnInstructionCycle := HandleInstructionCycle;
  if fCPU.Run<>Success then begin
    Prompt('ERROR: The CPU entered an error state.');
  end else begin
    Prompt('Program completed successfully.');
  end;
end;

begin
  with T6502Example.Create do begin
    Run;
    Free;
  end;
end.

