﻿
///  <license>
///
///  Copyright 2021 ChapmanWorld LLC ( https://chapmanworld.com )
///
///  Redistribution and use in source and binary forms, with or without modification,
///  are permitted provided that the following conditions are met:
///
///  1. Redistributions of source code must retain the above copyright notice,
///     this list of conditions and the following disclaimer.
///
///  2. Redistributions in binary form must reproduce the above copyright notice,
///     this list of conditions and the following disclaimer in the documentation and/or
///     other materials provided with the distribution.
///
///  3. Neither the name of the copyright holder nor the names of its contributors may be
///     used to endorse or promote products derived from this software without specific prior
///     written permission.
///
///  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
///  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
///  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
///  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
///  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
///  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
///  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
///  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///
///  </license>
/// <summary>
///   Chapmanworld Logging system library.
/// </summary>
/// <remarks>
///   NOTE: ** 16-bit half precision float type thanks to Marek Mauder with permission. <br/>
///   https://galfar.vevb.net/wp/2011/16bit-half-float-in-pascaldelphi/
/// </remarks>
unit cwTypes;
{$ifdef fpc}
  {$mode delphiunicode}
  {$MODESWITCH NESTEDPROCVARS}
  {$MODESWITCH ADVANCEDRECORDS}
  {$MODESWITCH TYPEHELPERS}
  {$ifdef CPU64}{$define CPU64BITS}{$endif}
  {$ifdef CPU32}{$define CPU32BITS}{$endif}
  {$WARN 6058 off : Call to subroutine "$1" marked as inline is not inlined}
  {$notes off}
{$else}
  {$ifdef CPU386} {$define CPU32BITS} {$endif}
  {$ifdef CPUX64} {$define CPU64BITS} {$endif}
{$endif}

interface
uses
  sysutils //- For Supports ( can be removed later )
, cwUnicodeCodec
;

const
  /// <summary>
  ///   A constant representing a carriage return character from the ascii set.
  /// </summary>
  CR = #13;

  /// <summary>
  ///   A constant representing a line feed character from the ascii set. <br />
  /// </summary>
  LF = #10;

  /// <summary>
  ///   A constant representing a tab feed character from the ascii set.
  /// </summary>
  TAB = #9;

  ///  <summary>
  ///    A constant representing a carrage return character followed by a line
  ///    feed character from the ascii set.
  ///  </summary>
  CRLF: string = CR + LF;

  ///  <summary>
  ///    A GUID containing nothing but zeros.
  ///  </summary>
  NullGUID: TGUID = '{00000000-0000-0000-0000-000000000000}';

const
  stSuccess                : TGUID = '{00000000-0000-0000-0000-000000000000}'; // nothing went wrong
  stUnknown                : TGUID = '{01010101-0101-0101-0101-010101010101}'; // something went wrong, don't know what.
  stUTFUnknownNotSupported : TGUID = '{2C9E1261-B5B9-4C3B-B014-DA0A3DE4A3F5}'; // unicode format utfUnknown cannot be used for encoding or decoding.
  stUnicodeEncodingFailed  : TGUID = '{AFDCE160-7449-4D11-B000-93E4CCBDA3B3}';
  stUnicodeDecodingFailed  : TGUID = '{988BDC44-209E-4346-91B8-465859A6A235}';
  stLowercaseFailed        : TGUID = '{97F4FAF1-08C8-4AA1-BE89-5B05728E9562}';
  stUppercaseFailed        : TGUID = '{48AB0089-E5CC-4786-93BF-CAB209F12306}';
  stStrToSingleError       : TGUID = '{A35B9073-0817-41BC-9DA5-D4CDA3375153}';
  stStrToDoubleError       : TGUID = '{C8629BE0-D891-4DF3-927C-1A19F6F58F22}';
  stStrToExtendedError     : TGUID = '{E07B0120-AB72-4A9D-9F69-1C11795CB299}';
  stStrToInt8Error         : TGUID = '{44FC35AC-5665-4C4D-A797-ED0EA7D9D021}';
  stStrToInt16Error        : TGUID = '{DEB3C139-C041-417C-BF16-6D3DA03192C4}';
  stStrToInt32Error        : TGUID = '{73BF83B5-9422-4A6C-8D1A-A8703FEE77A2}';
  stStrToInt64Error        : TGUID = '{2D17C748-9831-4D06-86F6-B2F96EED3CFF}';
  stStrToUInt8Error        : TGUID = '{673130F2-B48C-4900-8F34-D17B084B6FF3}';
  stStrToUInt16Error       : TGUID = '{71EDEFE8-DB73-4D92-81F0-FABA29C94AE8}';
  stStrToUInt32Error       : TGUID = '{3B06E54F-2195-4826-A0E2-7793EE4E5650}';
  stStrToUInt64Error       : TGUID = '{E5FE18D0-31C0-42BD-A040-B843CB63DAF1}';
  stStringToHexFailed      : TGUID = '{EBF2CA32-9002-4B5B-8825-9649B68E69E0}';
  stInvalidYear            : TGUID = '{CAD6BCC5-0003-4FD7-B3A9-5B0C1C3E0D72}';
  stInvalidMonth           : TGUID = '{4F10C822-E66A-4077-9677-363C61649946}';
  stInvalidDate            : TGUID = '{DB36BE0D-C62A-4261-BDAE-BC24F88218B3}';
  stGUIDConversionError    : TGUID = '{D0AF95A5-B00B-45F9-9EB0-D9A3F9B4F53C}';
  stCreateMutexFailed      : TGUID = '{63BE6A3F-BC0B-4134-9DC6-6B853ED8E74F}';
  stVariantTypeCoersion    : TGUID = '{2E979248-B104-493B-921C-3A67E07ABEF6}';

type
  {$if defined(ANDROID) or defined(IOS)}
  AnsiChar = uint8;
  pAnsiChar = ^AnsiChar;
  {$endif}

  ///  <summary>
  ///    A simple alias for dynamic array of string.
  ///    This type is used as the return type for the TStringHelper.Explode()
  ///    method.
  ///  </summary>
  TArrayOfString = array of string;

  ///  <summary>
  ///    Array of characters.
  ///  </summary>
  TArrayOfChar = array of char;

  ///  <summary>
  ///    Character set type.
  ///  </summary>
  TSetOfChar = TSysCharSet;

  /// <summary>
  ///   Specifies a supported unicode text format.
  /// </summary>
  TUnicodeFormat = cwUnicodeCodec.TUnicodeFormat;

  ///  <summary>
  ///    Used to specify the case sensitivity for utfCodepoint and utfString
  ///    operations.
  ///  </summary>
  TCaseSensitivity = (
    caseSensitive,
    caseInsensitive
  );

type
  ///  <summary>
  ///    Represents a single unicode codepoint (character) which may be
  ///    assigned to and from a char, ansichar or widechar. <br/>
  ///    Note that assignment to AnsiChar will fail when the code-point falls
  ///    outside the range of ansi characters, and will return (#0) in this case.
  ///  </summary>
  utfCodepoint = record
  public
    Value: TUnicodeCodepoint;
  public
    class operator Implicit          ( const a: AnsiChar     ): utfCodepoint;
    class operator Implicit          ( const a: utfCodepoint ): AnsiChar;
    class operator Implicit          ( const a: WideChar     ): utfCodepoint;
    class operator Implicit          ( const a: utfCodepoint ): WideChar;
    class operator Explicit          ( const a: AnsiChar     ): utfCodepoint;
    class operator Explicit          ( const a: utfCodepoint ): AnsiChar;
    class operator Explicit          ( const a: WideChar     ): utfCodepoint;
    class operator Explicit          ( const a: utfCodepoint ): WideChar;
    class operator GreaterThan       ( const a: utfCodepoint;   const b: utfCodepoint ): boolean;
    class operator GreaterThan       ( const a: utfCodepoint;   const b: AnsiChar     ): boolean;
    class operator GreaterThan       ( const a: utfCodepoint;   const b: WideChar     ): boolean;
    class operator GreaterThan       ( const a: AnsiChar;       const b: utfCodepoint ): boolean;
    class operator GreaterThan       ( const a: WideChar;       const b: utfCodepoint ): boolean;
    class operator GreaterThanOrEqual( const a: utfCodepoint;   const b: utfCodepoint ): boolean;
    class operator GreaterThanOrEqual( const a: utfCodepoint;   const b: AnsiChar     ): boolean;
    class operator GreaterThanOrEqual( const a: utfCodepoint;   const b: WideChar     ): boolean;
    class operator GreaterThanOrEqual( const a: AnsiChar;       const b: utfCodepoint ): boolean;
    class operator GreaterThanOrEqual( const a: WideChar;       const b: utfCodepoint ): boolean;
    class operator LessThan          ( const a: utfCodepoint;   const b: utfCodepoint ): boolean;
    class operator LessThan          ( const a: utfCodepoint;   const b: AnsiChar     ): boolean;
    class operator LessThan          ( const a: utfCodepoint;   const b: WideChar     ): boolean;
    class operator LessThan          ( const a: AnsiChar;       const b: utfCodepoint ): boolean;
    class operator LessThan          ( const a: WideChar;       const b: utfCodepoint ): boolean;
    class operator LessThanOrEqual   ( const a: utfCodepoint;   const b: utfCodepoint ): boolean;
    class operator LessThanOrEqual   ( const a: utfCodepoint;   const b: AnsiChar     ): boolean;
    class operator LessThanOrEqual   ( const a: utfCodepoint;   const b: WideChar     ): boolean;
    class operator LessThanOrEqual   ( const a: AnsiChar;       const b: utfCodepoint ): boolean;
    class operator LessThanOrEqual   ( const a: WideChar;       const b: utfCodepoint ): boolean;
    class operator Equal             ( const a: utfCodepoint;   const b: utfCodepoint ): boolean;
    class operator Equal             ( const a: utfCodepoint;   const b: AnsiChar     ): boolean;
    class operator Equal             ( const a: utfCodepoint;   const b: WideChar     ): boolean;
    class operator Equal             ( const a: AnsiChar;       const b: utfCodepoint ): boolean;
    class operator Equal             ( const a: WideChar;       const b: utfCodepoint ): boolean;
    class operator NotEqual          ( const a: utfCodepoint;   const b: utfCodepoint ): boolean;
    class operator NotEqual          ( const a: utfCodepoint;   const b: AnsiChar     ): boolean;
    class operator NotEqual          ( const a: utfCodepoint;   const b: WideChar     ): boolean;
    class operator NotEqual          ( const a: AnsiChar;       const b: utfCodepoint ): boolean;
    class operator NotEqual          ( const a: WideChar;       const b: utfCodepoint ): boolean;
  public

    ///  <summary>
    ///    Returns the size of this codepoint in bytes, when converted
    ///    to the specified unicode format.
    ///  </summary>
    function Size( const UnicodeFormat: TUnicodeFormat ): nativeuint;

    ///  <summary>
    ///    Writes the codepoint to the target of the pointer 'P' in
    ///    the format specified, and advances P beyond the codepoint.
    ///  </summary>
    procedure WriteMem( var P: pointer; const UnicodeFormat: TUnicodeFormat );

    ///  <summary>
    ///    Reads the codepoint from the target of pointer 'P' in the
    ///    format specified, and advances P beyond the codepoint.
    ///  </summary>
    procedure ReadMem( var P: pointer; const UnicodeFormat: TUnicodeFormat );

    ///  <summary>
    ///    Returns the uppercase of this code point.
    ///    This method relies on the target-operating system for locale sensitive
    ///    case conversion, and will therefore conform to the constraints of
    ///    the target.
    ///  </summary>
    function Uppercase: utfCodepoint;

    ///  <summary>
    ///    Returns the lowercase of this code point.
    ///    This method relies on the target-operating system for locale sensitive
    ///    case conversion, and will therefore conform to the constraints of
    ///    the target.
    ///  </summary>
    function Lowercase: utfCodepoint;
  end;

  TArrayOfCodepoint = array of utfCodepoint;

  TArrayOfUInt8 = array of uint8;

type
  ///  <summary>
  ///    A utility for allocating and freeing zero-terminated utf-8 strings for
  ///    interacting with ansi/utf-8 across module boundtry.
  ///  </summary>
  TUtf8String = record
    ///  <summary>
    ///    Allocates space to store the srcStr as converted to utf-8 with
    ///    a zero-terminator, and converts the srcStr into that buffer. <br/>
    ///  </summary>
    class function Alloc( const srcStr: string ): pointer; static;

    ///  <summary>
    ///    Returns the size (in bytes) of the srcStr measured as a utf-8 string
    ///    including a byte for the zero-terminator.
    ///  </summary>
    class function Size( const srcStr: pointer ): nativeuint; static;

    ///  <summary>
    ///    Frees the memory used to store a zero-terminated utf-8 string as
    ///    previously allocated with a call to TUF8String.Alloc.
    ///  </summary>
    class procedure Free( const srcStr: pointer ); static;

    ///  <summary>
    ///    Returns a zero-terminated utf-8 string converted back into a pascal
    ///    unicode UTF16-LE string.
    ///  </summary>
    class function AsString( const srcStr: pointer ): string; static;
  end;

type
  ///  <summary>
  ///    A utility for allocating and freeing zero-terminated utf-16LE strings for
  ///    interacting with utf-16LE apis.
  ///  </summary>
  TUtf16LEString = record
    ///  <summary>
    ///    Allocates space to store the srcStr as converted to utf-16LE with
    ///    a zero-terminator, and converts the srcStr into that buffer. <br/>
    ///  </summary>
    class function Alloc( const srcStr: string ): pointer; static;

    ///  <summary>
    ///    Returns the size (in bytes) of the srcStr measured as a utf-16LE string
    ///    including a byte for the zero-terminator.
    ///  </summary>
    class function Size( const srcStr: pointer ): nativeuint; static;

    ///  <summary>
    ///    Frees the memory used to store a zero-terminated utf-16LE string as
    ///    previously allocated with a call to TUF8String.Alloc.
    ///  </summary>
    class procedure Free( const srcStr: pointer ); static;

    ///  <summary>
    ///    Returns a zero-terminated utf-16LE string converted back into a pascal
    ///    unicode UTF-16LE string.
    ///  </summary>
    class function AsString( const srcStr: pointer ): string; static;
  end;

///  <summary>
///    Returns the current executing thread ID.
///  </summary>
function getCurrentThreadID: TThreadID;

function Supports( const Instance: IInterface;  const InterfaceID: TGUID;  out Iface ): Boolean; overload;
function Supports( const Instance: TObject;     const InterfaceID: TGUID;  out Iface ): boolean; overload;
function Supports( const Instance: IInterface;  const InterfaceID: TGUID             ): boolean; overload;
function Supports( const Instance: TObject;     const InterfaceID: TGUID             ): boolean; overload;
function Supports( const _Class  : TClass;      const InterfaceID: TGUID             ): boolean; overload;


type
  ///  <summary>
  ///    Used to return success/failure codes from methods as an alternative
  ///    to using integer or enumerated return types. Status messages which
  ///    are registered (TStatus.Register from TStatusHelper) offer translation
  ///    and parameterization features for use with cwLog.
  ///  <summary>
  TStatus = record
  private
    GUID: TGUID;
  public
    class operator Implicit   ( const a: TStatus                   ): TGUID;
    class operator Explicit   ( const a: TStatus                   ): TGUID;
    class operator Implicit   ( const a: TGUID                     ): TStatus;
    class operator Explicit   ( const a: TGUID                     ): TStatus;
    class operator Implicit   ( const a: TStatus                   ): Boolean;
    class operator Explicit   ( const a: TStatus                   ): Boolean;
    class operator LogicalNot ( const a: TStatus                   ): Boolean;
    class operator Equal      ( const a: TStatus; const b: TStatus ): boolean;
    class operator NotEqual   ( const a: TStatus; const b: TStatus ): boolean;
    class operator Equal      ( const a: TGUID;   const b: TStatus ): boolean;
    class operator Equal      ( const a: TStatus; const b: TGUID   ): boolean;
    class operator NotEqual   ( const a: TGUID;   const b: TStatus ): boolean;
    class operator NotEqual   ( const a: TStatus; const b: TGUID   ): boolean;
  end;

type
  /// <summary>
  ///   Exception raised when the TStatus.Raize method is called.
  /// </summary>
  EStatus = class( Exception )
  private
    fStatus: TStatus;
  private
    function getStatus: TStatus;
  public
    constructor Create( const Status: TStatus; const Parameters: array of string ); reintroduce; overload;
    constructor Create( const Status: TStatus ); reintroduce; overload;
  public
    property Status: TStatus read getStatus;
  end;

type
  TStatusHeler = record helper for TStatus
  public

    ///  <summary>
    ///    Returns the registered text for the status code,
    ///    with parameter place-holders populated if there are any.
    ///  </summary>
    function AsString(const Parameters: array of string): string; overload;
    function AsString: string; overload;

    ///  <summary>
    ///    Simply returns the 'Unknown' status guid
    ///    ['{01010101-0101-0101-0101-010101010101}']
    ///  </summary>
    class function Unknown: TStatus; static;

    ///  <summary>
    ///    Simply returns the 'Success' status guid
    ///    ['{00000000-0000-0000-0000-000000000000}']
    ///  </summary>
    class function Success: TStatus; static;

    ///  <summary>
    ///    Raises an EStatus exception if this TStatus is
    ///    anything other than the success status. <br/>
    ///    If the GUID value has been registered as a
    ///    status message, the exception will be populated
    ///    with the message text. <br/>
    ///  </summary>
    procedure Raize; overload;

    ///  <summary>
    ///    Raies an EStatus exception if this TStatus is
    ///    anything other than the success status. <br/>
    ///    If the GUID value has been registered as a
    ///    status message, the exception will be populated
    ///    with the message text. <br/>
    ///    This overload will also populate parameter
    ///    place-holders in the message if there are any.
    ///  </summary>
    procedure Raize( const Parameters: array of string ); overload;

    ///  <summary>
    ///    Registers a piece of text against this status
    ///    which is used to populate exceptions or log
    ///    entries. <br/>
    ///    The text may contain placeholders for
    ///    parameters in order that it may be translated
    ///    for language support.
    ///  </summary>
    procedure Register(const a: string);
  end;

type
  TStringHelper = record helper for string
  public

    ///  <summary>
    ///    Returns the length of the string.
    ///  </summary>
    function Len: nativeuint;

    {$if defined(fpc) or defined(MSWINDOWS)}
    ///  <summary>
    ///    Returns the string as an ansistring.
    ///    Getter for AsAnsiString property.
    ///  <summary>
    function getAsAnsiString: ansiString;
    {$endif}

    {$if defined(fpc) or defined(MSWINDOWS)}
    ///  <summary>
    ///    Sets the string as an ansistring.
    ///    Setter for AsAnsiString property.
    ///  <summary>
    procedure setAsAnsiString( const value: ansistring );
    {$endif}

    ///  <summary>
    ///    Sets the string using a pansichar / pchar.
    ///  </summary>
    procedure SetAsPChar( const value: pointer );

    ///  <summary>
    ///    Pads the left side of the string with the ch character until the
    ///    length of the string matches CharCount. (Unlike PadLeft which
    ///    pre-fixes the string).
    ///  </summary>
    function LeftPad( const CharCount: nativeuint; const ch: char ): string;

    ///  <summary>
    ///    Pads the right side of the string with the ch character until
    ///    the length of the string matches CharCount. (Unlike PadRight which
    ///    post-fixes the string).
    ///  </summary>
    function RightPad( const CharCount: nativeuint; const ch: char ): string;

    ///  <summary>
    ///    Separates the string for each instance of cDelimiter found.
    ///  </summary>
    function Explode( const cDelimiter: string ): TArrayOfString;

    ///  <summary>
    ///    Combines the 'Parts' into a single string, separated by cDelimiter.
    ///  </summary>
    function Combine( const cDelimiter: string; const Parts: TArrayOfString; const StartIdx: nativeuint = 0): string;

    /// <summary>
    ///   Trims whitepsace characters from the left of the string. Whitespace
    ///   characters may be provided in the optional array parameter, or else
    ///   the default whitespace characters will be used. <br />( Space, Tab,
    ///   CR, LF )
    /// </summary>
    /// <param name="Whitespace">
    ///   Optional array of characters to consider as white space.
    /// </param>
    function TrimLeft(const Whitespace: array of char): string; overload;
    ///<exclude/>
    function TrimLeft: string; overload;

    /// <summary>
    ///   Trims whitepsace characters from the right of the string. Whitespace
    ///   characters may be provided in the optional array parameter, or else
    ///   the default whitespace characters will be used. <br />( Space, Tab,
    ///   CR, LF )
    /// </summary>
    /// <param name="Whitespace">
    ///   Optional array of characters to consider as white space.
    /// </param>
    function TrimRight(const Whitespace: array of char): string; overload;
    ///<exclude/>
    function TrimRight: string; overload;

    /// <summary>
    ///   Trims whitepsace characters from both the left and the right of the
    ///   string. Whitespace characters may be provided in the optional array
    ///   parameter, or else the default whitespace characters will be used. <br />
    ///   ( Space, Tab, CR, LF )
    /// </summary>
    /// <param name="Whitespace">
    ///   Optional array of characters to consider as white space.
    /// </param>
    function Trim( const Whitespace: array of char ): string; overload;
    ///  <exclude/>
    function Trim: string; overload;

    ///  <summary>
    ///    Replaces lower-case characters with upper-case characters.
    ///  </summary>
    function Uppercase: string;

    /// <summary>
    ///   Performs both Uppercase() and Trim() using the default whitepsace
    ///   characters.
    /// </summary>
    function UppercaseTrim: string;

    ///  <summary>
    ///    Replaces upper-case characters with lower-case charaacters
    ///  </summary>
    function Lowercase: string;

    /// <summary>
    ///   Performs both Lowercase() and Trim() using the default whitepsace
    ///   characters.
    /// </summary>
    function LowercaseTrim: string;

    /// <summary>
    ///   Returns a new string which is a copy of this one, but with the first
    ///   found instance of OldParttern replaced by NewPattern. <br />You may
    ///   optionally specify the case sensitivity of the search.
    /// </summary>
    /// <param name="OldPattern">
    ///   Test to be replaced.
    /// </param>
    /// <param name="NewPattern">
    ///   The replacement text.
    /// </param>
    /// <param name="CaseSensitivity">
    ///   (Optional) Specify case seneitivity of search.
    /// </param>
    function Replace(const OldPattern, NewPattern: string; const CaseSensitivity: TCaseSensitivity = TCaseSensitivity.caseSensitive  ): string;

    /// <summary>
    ///   Returns a new string which is a copy of this one, but with the all
    ///   found instances of OldParttern replaced by NewPattern. <br />You may
    ///   optionally specify the case sensitivity of the search.
    /// </summary>
    /// <param name="OldPattern">
    ///   Test to be replaced.
    /// </param>
    /// <param name="NewPattern">
    ///   The replacement text.
    /// </param>
    /// <param name="CaseSensitivity">
    ///   (Optional) Specify case seneitivity of search.
    /// </param>
    function ReplaceAll(const OldPattern, NewPattern: string; const CaseSensitivity: TCaseSensitivity = TCaseSensitivity.caseSensitive ): string;

    ///  <summary>
    ///  </summary>
    function Format ( const args: Array of const ): string;

    ///  <summary>
    ///    Returns 'ACount' characters from the left of the string.
    ///  </summary>
    function Left(const ACount: nativeuint): string;

    ///  <summary>
    ///    Returns 'ACount' characters from the right of the string.
    ///  </summary>
    function Right(const ACount: nativeuint): string;

    ///  <summary>
    ///    Returns 'ACount' characters from the position 'AStart' within
    ///    the string.
    ///  </summary>
    function Copy(const AStart, ACount: nativeuint): string;

    /// <summary>
    ///   Attempts to locate the pattern 'SubString' within this string, and if
    ///   successful, returns the start and stop character indices in the out
    ///   parameters. <br />Optionally, the search may be performed with case
    ///   insensitivity.
    /// </summary>
    /// <param name="SubString">
    ///   Pattern to search for.
    /// </param>
    /// <param name="Start">
    ///   If successfully found, this is the index of the first character of
    ///   the search pattern within the string.
    /// </param>
    /// <param name="Stop">
    ///   If succesfully found, this is the indes of the last character of the
    ///   search pattern within the string.
    /// </param>
    /// <param name="CaseSensitivity">
    ///   (Optional) You may specify the case sensitivity of the search using
    ///   the enum TCaseSensitivity.
    /// </param>
    /// <returns>
    ///   Returns TRUE if the sub-string pattern is found, else returns FALSE.
    /// </returns>
    function Find( const SubString: string; out Start: nativeuint; out Stop: nativeuint; const CaseSensitivity: TCaseSensitivity = TCaseSensitivity.caseSensitive ): boolean;

    ///  <summary>
    ///    Returns TRUE if this string contains the provided sub-string.
    ///  </summary>
    function Contains( const SubString: string; const CaseSensitivity: TCaseSensitivity = TCaseSensitivity.caseSensitive ): boolean;

    ///  <summary>
    ///    Inserts 'Padding' instances of 'PaddingChar' at the left of the
    ///    string (prefix}
    ///  </summary>
    function Padleft(const Padding: uint32; const PaddingChar: char): string;

    ///  <summary>
    ///    Inserts 'Padding' instances of 'PaddignChar' at the right of the
    ///    string (postfix)
    ///  </summary>
    function PadRight(const Padding: uint32; const PaddingChar: char): string;

    ///  <summary>
    ///    Returns the string as a boolean.
    ///    The return of this funciton is true if the string matches the
    ///    positive parameter, otherwise returns false.
    ///  </summary>
    function AsBoolean(const CaseSensitive: boolean = False; const positive: string = 'TRUE' ): boolean;

    ///  <summary>
    ///    Returns the string as an unsigned int8
    ///  </summary>
    function AsUInt8: uint8;

    ///  <summary>
    ///    Returns the string as an unsigned int16
    ///  </summary>
    function AsUInt16: uint16;

    ///  <summary>
    ///    Returns the string as an unsigned int32
    ///  </summary>
    function AsUint32: uint32;

    ///  <summary>
    ///    Returns the string as an insigned int64
    ///  </summary>
    function AsUint64: uint64;

    ///  <summary>
    ///    Returns the string as a signed int8
    ///  </summary>
    function AsInt8: int8;

    ///  <summary>
    ///    Returns the string as a signed int16
    ///  </summary>
    function AsInt16: int16;

    ///  <summary>
    ///    Returns the string as a signed int32
    ///  </summary>
    function AsInt32: int32;

    ///  <summary>
    ///    Returns the string as a signed int64
    ///  </summary>
    function AsInt64: int64;

    ///  <summary>
    ///    Returns the string as a single-precision floating point.
    ///  </summary>
    function AsSingle: single;

    ///  <summary>
    ///    Returns the string as a double-precision floating point.
    ///  </summary>
    function AsDouble: double;

    ///  <summary>
    ///    Returns the string as an extended-precision floating point.
    ///  </summary>
    function AsExtended: extended;

    {$if defined(fpc) or defined(MSWINDOWS)}
    ///  <summary>
    ///    Get or Set this string as an ansi string, for compatibility with
    ///    older API's.
    ///  </summary>
    property AsAnsiString: ansistring read getAsAnsiString write setAsAnsiString;
    {$endif}

  end;

type
  {$if defined(fpc) or defined(MSWINDOWS)}
  TAnsiStringHelper = record helper for ansistring
    ///  <summary>
    ///    Returns the ansi string as a regular string for interation with
    ///    ansi API's.
    ///    Getter for AsString property.
    ///  </summary>
    function getAsString: string;

    ///  <summary>
    ///    Returns the ansi string as a regular string for interation with
    ///    ansi API's.
    ///  </summary>
    property AsString: string read getAsString;
  end;
  {$endif}

  TCharHelper = record helper for Char
  public
    ///  <summary>
    ///    Returns the character as a hex representation of the unicode codepoint.
    ///  </summary>
    function AsHex: string;

    ///  <summary>
    ///    Returns true if the ch char can be found in the char array.
    ///  </summary>
    function CharInArray( const Characters: array of char ): boolean;

    ///  <summary>
    ///    Returns true if the ch char can be found in the char set.
    ///  </summary>
    function CharInSet( const Characters: TSetOfChar ): boolean;
  end;

  TNativeUIntHelper = record helper for nativeuint
  private
    function getAsString: string;
    procedure setAsString( const value: string );
  public
    /// <summary>
    ///   Sets the nativeuint as a pointer.
    /// </summary>
    class function FromPointer( const value: pointer ): nativeuint; static;

    ///  <summary>
    ///    Casts the nativeuint as a pointer
    ///  </summary>
    function AsPointer: pointer;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;
  end;


  TNativeIntHelper = record helper for nativeint
  private
    function getAsString: string;
    procedure setAsString( const value: string );

  public
    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;
  end;

  TBooleanHelper = record helper for boolean
  private
    function getAsString: string;
    procedure setAsString(const Value: string);
  public

    property AsString: string read getAsString write setAsString;
  end;

  TUInt8Helper = record helper for uint8
  private
    function getAsHex: string;
    procedure setAsHex( const value: string );
    function getAsString: string;
    procedure setAsString( const value: string );

  public

    /// <summary>
    ///   Get / Set the value as a series of hex digits.
    /// </summary>
    /// <remarks>
    ///   Returns two hex digits, and expects two or fewer hex digits. Setting
    ///   to null-string will set the value to zero.
    /// </remarks>
    property AsHex: string read getAsHex write setAsHex;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;
  end;

  TUInt16Helper = record helper for uint16
  private
    function getAsHex: string;
    procedure setAsHex( const value: string );
    function getAsString: string;
    procedure setAsString( const value: string );

  public
    /// <summary>
    ///   Get / Set the value as a series of hex digits.
    /// </summary>
    /// <remarks>
    ///   Returns four hex digits, and expects four or fewer hex digits.
    ///   Setting to null-string will set the value to zero.
    /// </remarks>
    property AsHex: string read getAsHex write setAsHex;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;
  end;

  TUInt32Helper = record helper for uint32
  private
    function getAsHex: string;
    procedure setAsHex( const value: string );
    function getAsString: string;
    procedure setAsString( const value: string );

  public

    /// <summary>
    ///   Get / Set the value as a series of hex digits.
    /// </summary>
    /// <remarks>
    ///   Returns eight hex digits, and expects eight or fewer hex digits.
    ///   Setting to null-string will set the value to zero.
    /// </remarks>
    property AsHex: string read getAsHex write setAsHex;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;
  end;


  TUInt64Helper = record helper for uint64
  private
    function getAsHex: string;
    procedure setAsHex( const value: string );
    function getAsString: string;
    procedure setAsString( const value: string );

  public

    /// <summary>
    ///   Get / Set the value as a series of hex digits.
    /// </summary>
    /// <remarks>
    ///   Returns sixteen hex digits, and expects sixteen or fewer hex digits.
    ///   Setting to null-string will set the value to zero.
    /// </remarks>
    property AsHex: string read getAsHex write setAsHex;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;
  end;


  TInt8Helper = record helper for int8
  private
    function getAsHex: string;
    procedure setAsHex( const value: string );
    function getAsString: string;
    procedure setAsString( const value: string );

  public

    /// <summary>
    ///   Get / Set the value as a series of hex digits.
    /// </summary>
    /// <remarks>
    ///   Returns two hex digits, and expects two or fewer hex digits. Setting
    ///   to null-string will set the value to zero. <br /><br />On conversion
    ///   to or from hex strings, the sign is ignored, negative numbers are
    ///   simply treated as the reflection of their twos compliment value.
    /// </remarks>
    property AsHex: string read getAsHex write setAsHex;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;
  end;

  TInt16Helper = record helper for int16
  private
    function getAsHex: string;
    procedure setAsHex( const value: string );
    function getAsString: string;
    procedure setAsString( const value: string );

  public

    /// <summary>
    ///   Get / Set the value as a series of hex digits.
    /// </summary>
    /// <remarks>
    ///   Returns two hex digits, and expects two or fewer hex digits. Setting
    ///   to null-string will set the value to zero. <br /><br />On conversion
    ///   to or from hex strings, the sign is ignored, negative numbers are
    ///   simply treated as the reflection of their twos compliment value.
    /// </remarks>
    property AsHex: string read getAsHex write setAsHex;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;
  end;


  TInt32Helper = record helper for int32
  private
    function getAsHex: string;
    procedure setAsHex( const value: string );
    function getAsString: string;
    procedure setAsString( const value: string );

  public

    /// <summary>
    ///   Get / Set the value as a series of hex digits.
    /// </summary>
    /// <remarks>
    ///   Returns two hex digits, and expects two or fewer hex digits. Setting
    ///   to null-string will set the value to zero. <br /><br />On conversion
    ///   to or from hex strings, the sign is ignored, negative numbers are
    ///   simply treated as the reflection of their twos compliment value.
    /// </remarks>
    property AsHex: string read getAsHex write setAsHex;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;
  end;


  TInt64Helper = record helper for int64
  private
    function getAsHex: string;
    procedure setAsHex( const value: string );
    function getAsString: string;
    procedure setAsString( const value: string );

  public

    /// <summary>
    ///   Get / Set the value as a series of hex digits.
    /// </summary>
    /// <remarks>
    ///   Returns two hex digits, and expects two or fewer hex digits. Setting
    ///   to null-string will set the value to zero. <br /><br />On conversion
    ///   to or from hex strings, the sign is ignored, negative numbers are
    ///   simply treated as the reflection of their twos compliment value.
    /// </remarks>
    property AsHex: string read getAsHex write setAsHex;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;
  end;


  TSingleHelper = record helper for single
  private
    function getAsString: string;
    procedure setAsString( const value: string );

  public

    ///  <summary>
    ///    Returns true if this single is equal to Value using
    ///    Precision as a margin for error both above and below
    ///    the value.
    ///  </summary>
    function Approximates( const Value: single; const Precision: single = 0.01 ): boolean;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;

  end;

  TDoubleHelper = record helper for double
  private
    function getAsString: string;
    procedure setAsString( const value: string );

  public

    ///  <summary>
    ///    Returns true if this double is equal to Value using
    ///    Precision as a margin for error both above and below
    ///    the value.
    ///  </summary>
    function Approximates( const Value: double; const Precision: double = 0.01 ): boolean;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;

  end;

  TExtendedHelper = record helper for extended
  private
    function getAsString: string;
    procedure setAsString( const value: string );

  public

    ///  <summary>
    ///    Returns true if this extended is equal to Value using
    ///    Precision as a margin for error both above and below
    ///    the value.
    ///  </summary>
    function Approximates( const Value: extended; const Precision: extended = 0.01 ): boolean;

    ///  <summary>
    ///    Get or Set the value as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;
  end;


  /// <summary>
  ///   A helper for working with TDateTime.
  /// </summary>
  TDateTimeHelper = record helper for TDateTime
  private
    function getDate: TDateTime;
    procedure setDate( const Value: TDateTime );
    function getTime: TDateTime;
    procedure setTime( const Value: TDateTime );
  public

    ///  <summary>
    ///    Returns the current date & time as a TDateTime;
    ///  </summary>
    class function Now: TDateTime; static;

    /// <summary>
    ///   This overload of Encode allows the TDateTime to be set using discrete
    ///   components of Year, Month, Day, Hour, Minute, Second and Millisecond.
    /// </summary>
    class function Encode( const Year, Month, Day, Hour, Minute, Second, Millisecond: uint16 ): TDateTime; overload; static;

    /// <summary>
    ///   This overload of Encode allows the TDateTime to be set using discrete
    ///   components of Year, Month, and Day. The time will be set to 00:00:00
    ///   at 0 ms.
    /// </summary>
    class function Encode( const Year, Month, Day: uint16 ): TDateTime; overload; static;

    /// <summary>
    ///   This overload of Encode allows the TDateTime to be set using discrete
    ///   components of Hour, Minute, Second and Millisecond. The date will be
    ///   set to 0000-00-00.
    /// </summary>
    class function Encode( const Hour, Minute, Second, Millisecond: uint16 ): TDateTime; overload; static;

    /// <summary>
    ///   This overload of Decode extracts the Year, Month, Day, Hour, Minute,
    ///   Second and Millisecond components from the TDateTime.
    /// </summary>
    procedure Decode( out Year, Month, Day, Hour, Minute, Second, Millisecond: uint16 ); overload;

    /// <summary>
    ///   This overload of Decode extracts only the Year, Month and Day
    ///   components of the TDateTime, ignoring the Hour, Minute, Second and
    ///   Millisecond components.
    /// </summary>
    procedure Decode( out Year, Month, Day: uint16 ); overload;

    /// <summary>
    ///   This overload of Decode extracts the Hour, Minute, Second and
    ///   Millisecond components of the TDateTime, ignoring the Year, Month and
    ///   Day.
    /// </summary>
    procedure Decode( out Hour, Minute, Second, Millisecond: uint16 ); overload;

    /// <summary>
    ///   This property returns or sets the Date components of the TDateTime,
    ///   without affecting the Time components.
    /// </summary>
    property Date: TDateTime read getDate write setDate;

    /// <summary>
    ///   This property returns or sets the Time componets of the TDateTime,
    ///   without affecting the date components.
    /// </summary>
    property Time: TDateTime read getTime write setTime;

  end;


  TArrayOfStringHelper = record helper for TArrayOfString
  public
    procedure TrimAll;
    procedure UppercaseAll;
    procedure LowercaseAll;
    procedure AssignArray(const value: array of string );
    function Find( const value: string; out FoundIdx: nativeuint; const CaseInsensitive: boolean = False ): boolean;
    function RemoveDuplicates( const CaseInsensitive: boolean = false ): TArrayOfString;
    procedure Remove( const item: string; const CaseInsensitive: boolean = false );
  end;


type
  half = record
    value: word;
    class operator Implicit(const a: single): half;
    class operator Implicit(const a: half): single;
    class operator Implicit(const a: double): half;
    class operator Implicit(const a: half): double;
    {$if defined(LINUX) and defined(FPC)}
    {$if (not defined(CPUAARCH64)) and (not defined(CPUARM))}
    class operator Implicit(const a: extended): half;
    class operator Implicit(const a: half): extended;
    {$endif}
    {$endif}
    {$ifndef CPU64BITS}
    {$warnings off}
    class operator Explicit(const a: single): half;
    class operator Explicit(const a: half): single;
    class operator Explicit(const a: double): half;
    class operator Explicit(const a: half): double;
    {$warnings on}
    {$endif}
    class operator Add(const a, b: half): half;
    class operator Subtract(const a, b: half): half;
    class operator Multiply(const a, b: half): half;
    class operator Divide(const a, b: half): half;
    class operator GreaterThan(const a: half; const b: half): boolean;
    class operator GreaterThan(const a: half; const b: single): boolean;
    class operator GreaterThan(const a: half; const b: double): boolean;

    class operator GreaterThanOrEqual(const a: half; const b: half): boolean;
    class operator GreaterThanOrEqual(const a: half; const b: single): boolean;
    class operator GreaterThanOrEqual(const a: half; const b: double): boolean;

    {$ifndef CPU64BITS}
    {$if (not defined(CPUAARCH64)) and (not defined(CPUARM))}
    class operator GreaterThan(const a: half; const b: extended): boolean;
    class operator GreaterThanOrEqual(const a: half; const b: extended): boolean;
    {$endif}
    {$endif}
    class operator GreaterThan(const a: single; const b: half): boolean;
    class operator GreaterThan(const a: double; const b: half): boolean;
    class operator GreaterThanOrEqual(const a: single; const b: half): boolean;
    class operator GreaterThanOrEqual(const a: double; const b: half): boolean;

    {$ifndef CPU64BITS}
    {$if (not defined(CPUAARCH64)) and (not defined(CPUARM))}
    class operator GreaterThan(const a: extended; const b: half): boolean;
    class operator GreaterThanOrEqual(const a: extended; const b: half): boolean;
    {$endif}
    {$endif}
    class operator LessThan(const a: half; const b: half): boolean;
    class operator LessThan(const a: half; const b: single): boolean;
    class operator LessThan(const a: half; const b: double): boolean;
    class operator LessThanOrEqual(const a: half; const b: half): boolean;
    class operator LessThanOrEqual(const a: half; const b: single): boolean;
    class operator LessThanOrEqual(const a: half; const b: double): boolean;
    {$ifndef CPU64BITS}
    {$if (not defined(CPUAARCH64)) and (not defined(CPUARM))}
    class operator LessThan(const a: half; const b: extended): boolean;
    class operator LessThanOrEqual(const a: half; const b: extended): boolean;
    {$endif}
    {$endif}
    class operator LessThan(const a: single; const b: half): boolean;
    class operator LessThan(const a: double; const b: half): boolean;
    class operator LessThanOrEqual(const a: single; const b: half): boolean;
    class operator LessThanOrEqual(const a: double; const b: half): boolean;
    {$ifndef CPU64BITS}
    {$if (not defined(CPUAARCH64)) and (not defined(CPUARM))}
    class operator LessThan(const a: extended; const b: half): boolean;
    class operator LessThanOrEqual(const a: extended; const b: half): boolean;
    {$endif}
    {$endif}
    class operator Equal(const a, b: half): boolean;

  end;

const
  HalfMin:     Single = 5.96046448e-08; // Smallest positive half
  HalfMinNorm: Single = 6.10351562e-05; // Smallest positive normalized half
  HalfMax:     Single = 65504.0;        // Largest positive half
  // Smallest positive e for which half (1.0 + e) != half (1.0)
  HalfEpsilon: Single = 0.00097656;
  HalfNaN:     half = ( value: 65535 );
  HalfPosInf:  half = ( value: 31744 );
  HalfNegInf:  half = ( value: 64512 );


type
  TGUIDHelper = record helper for TGUID
  private
    function getAsString: string;
    procedure setAsString( const value: string );

  public

    ///  <summary>
    ///    Compare B to this GUID, return TRUE if they are the same.
    ///  </summary>
    function EqualTo( const B: TGUID ): boolean;

    ///  <summary>
    ///    Set the GUID to a newly generated value.
    ///  </summary>
    class function New: TGUID; static;

    ///  <summary>
    ///    Get/Set the GUID as a string.
    ///  </summary>
    property AsString: string read getAsString write setAsString;

  end;

type
  ///  <summary>
  ///    A simple interface for storing a value of
  ///    generic type. Useful for making a passable
  ///    interface of any type. <br/>
  ///    Note: When specializing as a reference type
  ///    such as an interface, the reference will
  ///    be set nil when freeing the generic.
  ///  </summary>
  IGenericValue<T> = interface
    ['{39D7BB35-168B-4AF1-ACF5-D223A29D3E80}']

    ///  <summary>
    ///    Returns the value. (getter for 'Value' property)
    ///  </summary>
    function getValue: T;

    ///  <summary>
    ///    Sets the value. (setter for 'Value' property)
    ///  </summary>
    procedure setValue( const Value: T );

    ///  <summary>
    ///    Get/Set the value.
    ///  </summary>
    property Value: T read getValue write setValue;
  end;

  ///  <summary>
  ///    A class implementing IGenericValue<T>
  ///  </summary>
  TGenericValue<T> = class( TInterfacedObject, IGenericValue<T> )
  private
    fValue: T;
  strict private //- IGenericValue<T> -//
    function getValue: T;
    procedure setValue( const Value: T );
  public
    constructor Create( const Value: T );
    destructor Destroy; override;
  end;

type
  ///  <summary>
  ///    Enumeration of data types supported
  ///    by the TVariant record type.
  ///  </summary>
  TVariantType = (
    vt_null,
    vt_uint8,
    vt_uint16,
    vt_uint32,
    vt_uint64,
    vt_int8,
    vt_int16,
    vt_int32,
    vt_int64,
    vt_char,
    vt_string,
    vt_half,
    vt_single,
    vt_double,
    vt_extended,
    vt_object,
    vt_interface,
    vt_pointer
  );


  ///  <summary>
  ///    Provides a variant data type using
  ///    operator overloads to make it assignable
  ///    with other types.
  ///  </summary>
  TVariant = record
  private
    fStorage: IInterface;
    fDataType: TVariantType;
  private
    class function setter( const a: uint8      ): TVariant; overload; static;
    class function setter( const a: uint16     ): TVariant; overload; static;
    class function setter( const a: uint32     ): TVariant; overload; static;
    class function setter( const a: uint64     ): TVariant; overload; static;
    class function setter( const a: int8       ): TVariant; overload; static;
    class function setter( const a: int16      ): TVariant; overload; static;
    class function setter( const a: int32      ): TVariant; overload; static;
    class function setter( const a: int64      ): TVariant; overload; static;
    class function setter( const a: half       ): TVariant; overload; static;
    class function setter( const a: single     ): TVariant; overload; static;
    class function setter( const a: double     ): TVariant; overload; static;
    {$ifdef CPU32} class function setter( const a: extended   ): TVariant; overload; static; {$endif}
    class function setter( const a: char       ): TVariant; overload; static;
    class function setter( const a: string     ): TVariant; overload; static;
    class function setter( const a: TObject    ): TVariant; overload; static;
    class function setter( const a: IInterface ): TVariant; overload; static;
    class function setter( const a: pointer    ): TVariant; overload; static;
    procedure getter( out a: uint8      ); overload;
    procedure getter( out a: uint16     ); overload;
    procedure getter( out a: uint32     ); overload;
    procedure getter( out a: uint64     ); overload;
    procedure getter( out a: int8       ); overload;
    procedure getter( out a: int16      ); overload;
    procedure getter( out a: int32      ); overload;
    procedure getter( out a: int64      ); overload;
    procedure getter( out a: half       ); overload;
    procedure getter( out a: single     ); overload;
    procedure getter( out a: double     ); overload;
    {$ifdef CPU32}  procedure getter( out a: extended   ); overload; {$endif}
    procedure getter( out a: char       ); overload;
    procedure getter( out a: string     ); overload;
    procedure getter( out a: TObject    ); overload;
    procedure getter( out a: IInterface ); overload;
    procedure getter( out a: pointer    ); overload;
  public
    procedure Clear;
    function DataType: TVariantType;
    class operator Implicit( const a: uint8      ): TVariant;
    class operator Implicit( const a: uint16     ): TVariant;
    class operator Implicit( const a: uint32     ): TVariant;
    class operator Implicit( const a: uint64     ): TVariant;
    class operator Implicit( const a: int8       ): TVariant;
    class operator Implicit( const a: int16      ): TVariant;
    class operator Implicit( const a: int32      ): TVariant;
    class operator Implicit( const a: int64      ): TVariant;
    class operator Implicit( const a: half       ): TVariant;
    class operator Implicit( const a: single     ): TVariant;
    class operator Implicit( const a: double     ): TVariant;
    {$ifdef CPU32}  class operator Implicit( const a: extended   ): TVariant; {$endif}
    class operator Implicit( const a: char       ): TVariant;
    class operator Implicit( const a: string     ): TVariant;
    class operator Implicit( const a: TObject    ): TVariant;
    class operator Implicit( const a: IInterface ): TVariant;
    class operator Implicit( const a: pointer    ): TVariant;
    class operator Explicit( const a: uint8      ): TVariant;
    class operator Explicit( const a: uint16     ): TVariant;
    class operator Explicit( const a: uint32     ): TVariant;
    class operator Explicit( const a: uint64     ): TVariant;
    class operator Explicit( const a: int8       ): TVariant;
    class operator Explicit( const a: int16      ): TVariant;
    class operator Explicit( const a: int32      ): TVariant;
    class operator Explicit( const a: int64      ): TVariant;
    class operator Explicit( const a: half       ): TVariant;
    class operator Explicit( const a: single     ): TVariant;
    class operator Explicit( const a: double     ): TVariant;
    {$ifdef CPU32}  class operator Explicit( const a: extended   ): TVariant; {$endif}
    class operator Explicit( const a: char       ): TVariant;
    class operator Explicit( const a: string     ): TVariant;
    class operator Explicit( const a: TObject    ): TVariant;
    class operator Explicit( const a: IInterface ): TVariant;
    class operator Explicit( const a: pointer    ): TVariant;

    class operator Implicit( const a: TVariant ): uint8;
    class operator Implicit( const a: TVariant ): uint16;
    class operator Implicit( const a: TVariant ): uint32;
    class operator Implicit( const a: TVariant ): uint64;
    class operator Implicit( const a: TVariant ): int8;
    class operator Implicit( const a: TVariant ): int16;
    class operator Implicit( const a: TVariant ): int32;
    class operator Implicit( const a: TVariant ): int64;
    class operator Implicit( const a: TVariant ): half;
    class operator Implicit( const a: TVariant ): single;
    class operator Implicit( const a: TVariant ): double;
    {$ifdef CPU32} class operator Implicit( const a: TVariant ): extended; {$endif}
    class operator Implicit( const a: TVariant ): char;
    class operator Implicit( const a: TVariant ): string;
    class operator Implicit( const a: TVariant ): TObject;
    class operator Implicit( const a: TVariant ): IInterface;
    class operator Implicit( const a: TVariant ): pointer;
    class operator Explicit( const a: TVariant ): uint8;
    class operator Explicit( const a: TVariant ): uint16;
    class operator Explicit( const a: TVariant ): uint32;
    class operator Explicit( const a: TVariant ): uint64;
    class operator Explicit( const a: TVariant ): int8;
    class operator Explicit( const a: TVariant ): int16;
    class operator Explicit( const a: TVariant ): int32;
    class operator Explicit( const a: TVariant ): int64;
    class operator Explicit( const a: TVariant ): half;
    class operator Explicit( const a: TVariant ): single;
    class operator Explicit( const a: TVariant ): double;
    {$ifdef CPU32} class operator Explicit( const a: TVariant ): extended; {$endif}
    class operator Explicit( const a: TVariant ): char;
    class operator Explicit( const a: TVariant ): string;
    class operator Explicit( const a: TVariant ): TObject;
    class operator Explicit( const a: TVariant ): IInterface;
    class operator Explicit( const a: TVariant ): pointer;
  end;


type
  ///  <summary>
  ///    An ILifeCycleDelegate may apply to any system which has
  ///    initialization and finalization events. For example, an
  ///    implementation of cwThreading.IThreadDelegate might also
  ///    implement ILifeCycleDelegate to be notified when the
  ///    controlling thread starts running, and when it shuts down.
  ///  </summary>
  ILifeCycleDelegate = interface
    ['{AFC78561-866C-4DDE-9820-DB3A776EEBD5}']

    ///  <summary>
    ///    Executed when the controlling system
    ///    starts up.
    ///  </summary>
    procedure StartUp;

    ///  <summary>
    ///    Executed when the controlling system
    ///    shuts down.
    ///  </summary>
    procedure ShutDown;
  end;

  ///  <summary>
  ///    An IPauseResumeDelegate may apply to any system which
  ///    pauses processing. For example, an implementation of
  ///    cwThreading.IThreadDelegate might also implement
  ///    IPauseResumeDelegate to be notified when a thread
  ///    is paused (put to sleep), or resumed (woken).
  ///  </summary>
  IPauseResumeDelegate = interface
    ['{C021BCAB-01F6-427B-B92F-3276EE65CD5F}']

    ///  <summary>
    ///    Called when the controlling system enters
    ///    a paused state.
    ///  </summary>
    procedure Pause;

    ///  <summary>
    ///    Called when the controlling system enters
    ///    a resumed state.
    ///  </summary>
    procedure Resume;
  end;

{$region ' Callbacks for delegate types '}

  {$ifdef fpc}
  /// <exclude/>
  TEnumerateCallback<T>                     = procedure( const Item: T );
  /// <exclude/>
  TEnumerateObjectCallback<T>               = procedure( const Item: T ) of object;
  /// <exclude/>
  TEnumerateNestedCallback<T>               = procedure( const Item: T ) is nested;
  /// <exclude/>
  TEnumerateBreakoutCallback<T>             = procedure( const Item: T; var BreakOut: boolean );
  /// <exclude/>
  TEnumerateBreakoutObjectCallback<T>       = procedure( const Item: T; var BreakOut: boolean ) of object;
  /// <exclude/>
  TEnumerateBreakoutNestedCallback<T>       = procedure( const Item: T; var BreakOut: boolean ) is nested;
  /// <exclude/>
  TEnumeratePairCallback<K,V>               = procedure( const Key: K; const Value: V );
  /// <exclude/>
  TEnumeratePairObjectCallback<K,V>         = procedure( const Key: K; const Value: V ) of object;
  /// <exclude/>
  TEnumeratePairNestedCallback<K,V>         = procedure( const Key: K; const Value: V ) is nested;
  /// <exclude/>
  TEnumeratePairBreakoutCallback<K,V>       = procedure( const Key: K; const Value: V; var BreakOut: boolean );
  /// <exclude/>
  TEnumeratePairBreakoutObjectCallback<K,V> = procedure( const Key: K; const Value: V; var BreakOut: boolean ) of object;
  /// <exclude/>
  TEnumeratePairBreakoutNestedCallback<K,V> = procedure( const Key: K; const Value: V; var BreakOut: boolean ) is nested;
  /// <exclude/>
  TNotifyCallback                           = procedure();
  /// <exclude/>
  TNotifyObjectCallback                     = procedure() of object;
  /// <exclude/>
  TNotifyNestedCallback                     = procedure() is nested;
  /// <exclude/>
  TNotifyWithSenderCallback<T>              = procedure( const Sender: T );
  /// <exclude/>
  TNotifyWithSenderObjectCallback<T>        = procedure( const Sender: T ) of object;
  /// <exclude/>
  TNotifyWithSenderNestedCallback<T>        = procedure( const Sender: T ) is nested;
  /// <exclude/>
  TIndexedLoopCallback                      = procedure( const index: nativeuint; const Parameters: array of TVariant );
  /// <exclude/>
  TIndexedLoopNoParametersCallback          = procedure( const index: nativeuint );
  /// <exclude/>
  TIndexedLoopObjectCallback                = procedure( const index: nativeuint; const Parameters: array of TVariant ) of object;
  /// <exclude/>
  TIndexedLoopObjectNoParametersCallback    = procedure( const index: nativeuint ) of object;
  /// <exclude/>
  TIndexedLoopNestedCallback                = procedure( const index: nativeuint; const Parameters: array of TVariant ) is nested;
  /// <exclude/>
  TIndexedLoopNestedNoParametersCallback    = procedure( const index: nativeuint ) is nested;
  /// <exclude/>
  TBoundedLoopCallback                      = procedure( const start: nativeuint; const stop: nativeuint; const Parameters: array of TVariant );
  /// <exclude/>
  TBoundedLoopNoParametersCallback          = procedure( const start: nativeuint; const stop: nativeuint );
  /// <exclude/>
  TBoundedLoopObjectCallback                = procedure( const start: nativeuint; const stop: nativeuint; const Parameters: array of TVariant ) of object;
  /// <exclude/>
  TBoundedLoopObjectNoParametersCallback    = procedure( const start: nativeuint; const stop: nativeuint ) of object;
  /// <exclude/>
  TBoundedLoopNestedCallback                = procedure( const start: nativeuint; const stop: nativeuint; const Parameters: array of TVariant ) is nested;
  /// <exclude/>
  TBoundedLoopNestedNoParametersCallback    = procedure( const start: nativeuint; const stop: nativeuint ) is nested;
  {$else}
  /// <exclude/>
  TEnumerateCallback<T>                     = reference to procedure( const Item: T );
  /// <exclude/>
  TEnumerateBreakoutCallback<T>             = reference to procedure( const Item: T; var BreakOut: boolean );
  /// <exclude/>
  TEnumeratePairCallback<K,V>               = reference to procedure( const Key: K; const Value: V );
  /// <exclude/>
  TEnumeratePairBreakoutCallback<K,V>       = reference to procedure( const Key: K; const Value: V; var BreakOut: boolean );
  /// <exclude/>
  TNotifyCallback                           = reference to procedure();
  /// <exclude/>
  TNotifyWithSenderCallback<T>              = reference to procedure( const Sender: T );
  /// <exclude/>
  TIndexedLoopCallback                      = reference to procedure( const index: nativeuint; const Parameters: array of TVariant );
  /// <exclude/>
  TIndexedLoopNoParametersCallback          = reference to procedure( const index: nativeuint );
  /// <exclude/>
  TBoundedLoopCallback                      = reference to procedure( const start: nativeuint; const stop: nativeuint; const Parameters: array of TVariant );
  /// <exclude/>
  TBoundedLoopNoParametersCallback          = reference to procedure( const start: nativeuint; const stop: nativeuint );
  {$endif}

{$endregion}


  TNotifyDelegate = record
  private
    fNotify: TNotifyCallback;
    {$ifdef fpc}
    fNotifyObject: TNotifyObjectCallback;
    fNotifyNested: TNotifyNestedCallback;
    {$endif}
  public
    function IsAssigned: boolean;
    /// <exclude/>
    class operator Implicit( const Value: TNotifyCallback ): TNotifyDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TNotifyCallback ): TNotifyDelegate;
    {$ifdef fpc}
    /// <exclude/>
    class operator Implicit( const Value: TNotifyObjectCallback ): TNotifyDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TNotifyObjectCallback ): TNotifyDelegate;
    /// <exclude/>
    class operator Implicit( const Value: TNotifyNestedCallback ): TNotifyDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TNotifyNestedCallback ): TNotifyDelegate;
    {$endif}
  public
    class function null: TNotifyDelegate; static;
    procedure Execute;
  end;

  TNotifyWithSenderDelegate<T> = record
  private
    fNotify: TNotifyWithSenderCallback<T>;
    {$ifdef fpc}
    fNotifyObject: TNotifyWithSenderObjectCallback<T>;
    fNotifyNested: TNotifyWithSenderNestedCallback<T>;
    {$endif}
  public
    /// <exclude/>
    class operator Implicit( const Value: TNotifyWithSenderCallback<T> ): TNotifyWithSenderDelegate<T>;
    /// <exclude/>
    class operator Explicit( const Value: TNotifyWithSenderCallback<T> ): TNotifyWithSenderDelegate<T>;
    {$ifdef fpc}
    /// <exclude/>
    class operator Implicit( const Value: TNotifyWithSenderObjectCallback<T> ): TNotifyWithSenderDelegate<T>;
    /// <exclude/>
    class operator Explicit( const Value: TNotifyWithSenderObjectCallback<T> ): TNotifyWithSenderDelegate<T>;
    /// <exclude/>
    class operator Implicit( const Value: TNotifyWithSenderNestedCallback<T> ): TNotifyWithSenderDelegate<T>;
    /// <exclude/>
    class operator Explicit( const Value: TNotifyWithSenderNestedCallback<T> ): TNotifyWithSenderDelegate<T>;
    {$endif}
  public
     class function null: TNotifyWithSenderDelegate<T>; static;
    procedure Execute( const Sender: T );
  end;

  TLoopDelegate = record
  private
    fIndexedLoopCallback: TIndexedLoopCallback;
    fIndexedLoopNoParametersCallback: TIndexedLoopNoParametersCallback;
    fBoundedLoopCallback: TBoundedLoopCallback;
    fBoundedLoopNoParametersCallback: TBoundedLoopNoParametersCallback;
    {$ifdef fpc}
    fBoundedLoopObjectCallback: TBoundedLoopObjectCallback;
    fBoundedLoopObjectNoParametersCallback: TBoundedLoopObjectNoParametersCallback;
    fBoundedLoopNestedCallback: TBoundedLoopNestedCallback;
    fBoundedLoopNestedNoParametersCallback: TBoundedLoopNestedNoParametersCallback;
    fIndexedLoopObjectCallback: TIndexedLoopObjectCallback;
    fIndexedLoopObjectNoParametersCallback: TIndexedLoopObjectNoParametersCallback;
    fIndexedLoopNestedCallback: TIndexedLoopNestedCallback;
    fIndexedLoopNestedNoParametersCallback: TIndexedLoopNestedNoParametersCallback;
    {$endif}
  public
    /// <exclude/>
    class operator Implicit( const Value: TIndexedLoopCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TIndexedLoopCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Implicit( const Value: TIndexedLoopNoParametersCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TIndexedLoopNoParametersCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Implicit( const Value: TBoundedLoopCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TBoundedLoopCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Implicit( const Value: TBoundedLoopNoParametersCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TBoundedLoopNoParametersCallback ): TLoopDelegate;
    {$ifdef fpc}
    /// <exclude/>
    class operator Implicit( const Value: TIndexedLoopObjectCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TIndexedLoopObjectCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Implicit( const Value: TIndexedLoopObjectNoParametersCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TIndexedLoopObjectNoParametersCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Implicit( const Value: TIndexedLoopNestedCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TIndexedLoopNestedCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Implicit( const Value: TIndexedLoopNestedNoParametersCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TIndexedLoopNestedNoParametersCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Implicit( const Value: TBoundedLoopObjectCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TBoundedLoopObjectCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Implicit( const Value: TBoundedLoopObjectNoParametersCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TBoundedLoopObjectNoParametersCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Implicit( const Value: TBoundedLoopNestedCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TBoundedLoopNestedCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Implicit( const Value: TBoundedLoopNestedNoParametersCallback ): TLoopDelegate;
    /// <exclude/>
    class operator Explicit( const Value: TBoundedLoopNestedNoParametersCallback ): TLoopDelegate;
    {$endif}
  public
    class function null: TLoopDelegate; static;
    procedure Execute( const start: nativeuint; const stop: nativeuint; const Parameters: array of TVariant );
  end;


  ///  <summary>
  ///    A utility for creating delegate callbacks of any procedure, function or method appropriate to the delegate type. <br/>
  ///    *** NOTE: FPC must be in {$mode delphiunicode} in order to pass methods correctly, otherwise internal error may occur.
  ///  </summary>
  MakeDelegate = record
    {$ifdef fpc}
    class function From      ( const Callback: TNotifyCallback                           ): TNotifyDelegate;               overload; static;
    class function From      ( const Callback: TNotifyObjectCallback                     ): TNotifyDelegate;               overload; static;
    class function From      ( const Callback: TNotifyNestedCallback                     ): TNotifyDelegate;               overload; static;
    class function From<T>   ( const Callback: TNotifyWithSenderCallback<T>              ): TNotifyWithSenderDelegate<T>;  overload; static;
    class function From<T>   ( const Callback: TNotifyWithSenderObjectCallback<T>        ): TNotifyWithSenderDelegate<T>;  overload; static;
    class function From<T>   ( const Callback: TNotifyWithSenderNestedCallback<T>        ): TNotifyWithSenderDelegate<T>;  overload; static;
    class function From      ( const Callback: TIndexedLoopCallback                      ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TIndexedLoopNoParametersCallback          ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TIndexedLoopObjectCallback                ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TIndexedLoopObjectNoParametersCallback    ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TIndexedLoopNestedCallback                ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TIndexedLoopNestedNoParametersCallback    ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TBoundedLoopCallback                      ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TBoundedLoopNoParametersCallback          ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TBoundedLoopObjectCallback                ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TBoundedLoopObjectNoParametersCallback    ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TBoundedLoopNestedCallback                ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TBoundedLoopNestedNoParametersCallback    ): TLoopDelegate;                 overload; static;
    {$else}
    class function From      ( const Callback: TNotifyCallback                           ): TNotifyDelegate;               overload; static;
    class function From<T>   ( const Callback: TNotifyWithSenderCallback<T>              ): TNotifyWithSenderDelegate<T>;  overload; static;
    class function From      ( const Callback: TIndexedLoopCallback                      ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TIndexedLoopNoParametersCallback          ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TBoundedLoopCallback                      ): TLoopDelegate;                 overload; static;
    class function From      ( const Callback: TBoundedLoopNoParametersCallback          ): TLoopDelegate;                 overload; static;
    {$endif}
  end;

implementation
uses
{$ifdef MSWINDOWS}
  cwWin32
{$else}
  cwPosix
{$endif}
;

const
  ///  <summary>
  ///    The default whitepsace character array used when trimming strings.
  ///  </summary>
  cDefaultWhitespace: array of char = [ ' ', TAB, CR, LF ];

  ///  <summary>
  ///    Array of characters which are valid as part of a hex number.
  ///  </summary>
  cHexChars: array of char = [ '0','1','2','3','4','5','6','7','8','9', 'a','b','c','d','e','f','A','B','C','D','E','F' ];

function getCurrentThreadID: TThreadID;
begin
  {$ifdef fpc}
  Result := system.GetCurrentThreadID;
  {$else}
  {$ifdef MSWINDOWS}
  Result := cwWin32.GetCurrentThreadID;
  {$else}
  Result := pthread_self();
  {$endif}
  {$endif}
end;

function HexToString( const value; const bytecount: nativeuint ): string;
var
  CP: utfCodepoint;
  pbyte: ^uint8;
  idx: nativeuint;
  lsn: uint8;
  msn: uint8;
begin
  Result := '';
  if bytecount=0 then exit;
  {$hints off}
  pbyte := pointer( nativeuint(@value) + pred(bytecount) );
  {$hints on}
  for idx := 0 to pred(bytecount) do begin
    lsn := (pbyte^ and $F0) shr 4;
    msn := pbyte^ and $0F;
    CP.Value := lsn;
    if (CP.Value<$0A) then
      CP.value := CP.value + $30 else //- + ascii/utf '0'
      CP.value := CP.value + $37;     //- + ascii/utf 'A' (sub 10 for '0'..'9')
    Result := Result + char( CP );
    CP.Value := msn;
    if (CP.Value<$0A) then
      CP.value := CP.value + $30 else //- + ascii/utf '0'
      CP.value := CP.value + $37;     //- + ascii/utf 'A' (sub 10 for '0'..'9')
    Result := Result + char( CP );
    {$hints off}
    pbyte := pointer( nativeuint(pbyte) - sizeof(uint8) );
    {$hints on}
  end;
end;

procedure StringToHex( const src: string; var Buffer; const BufferSize: nativeuint );

  function CalcNibble( const Codepoint: uint32 ): uint8; inline;
  begin
    Result := 0;
    if (Codepoint>=$30) and (Codepoint<=$39) then begin // utf/ascii 0..9
      Result := CodePoint-$30;
    end else if (Codepoint>=$41) and (Codepoint<=$46) then begin // utf/ascii 'A'..'F'
      Result := Codepoint-$37;
    end else if (Codepoint>=$61) and (Codepoint<=$66) then begin // utf/ascii 'a'..'f'
      Result := Codepoint-$57;
    end else begin
      TStatus( stStringToHexFailed ).Raize( [ char(utfCodepoint(Codepoint)) ] );
    end;
  end;
var
  CP: utfCodepoint;
  DecodedBytes: nativeuint;
  Stridx: nativeuint;
  Strtop: nativeuint;
  pByte: ^uint8;
  lsn: uint8; //- least significant nibble, going to toggle 0/4 for shl.
begin
  DecodedBytes := 0;
  {$ifdef NEXTGEN}
  stridx := 0;
  strtop := pred(Length(src));
  {$else}
  stridx := 1;
  strtop := Length(src);
  {$endif}
  {$hints off}
  pByte := pointer( nativeuint( @Buffer ) + pred(BufferSize) );
  {$hints on}
  lsn := 4;
  while (stridx<=strtop) and (DecodedBytes<=BufferSize) do begin
    CP := src[stridx];
    if lsn=4 then begin
      pByte^ := CalcNibble(CP.Value);
      pByte^ := pByte^shl lsn;
      lsn := 0;
    end else begin
      pByte^ := pByte^ or CalcNibble(CP.Value);
      inc( DecodedBytes );
      {$hints off}
      pByte := pointer( nativeuint(pByte) - sizeof(uint8) );
      {$hints on}
      lsn := 4;
    end;
    inc(stridx);
  end;
end;

function IsEqualGUID( const A: TGUID; const B: TGUID ): boolean; inline;
type
  compareGUID = record
    A: uint32;
    B: uint32;
    C: uint32;
    D: uint32;
  end;
begin
  Result := (compareGUID(A).A = compareGUID(B).A)
        and (compareGUID(A).B = compareGUID(B).B)
        and (compareGUID(A).C = compareGUID(B).C)
        and (compareGUID(A).D = compareGUID(B).D);
end;

function DaysInMonth( const Year, Month: uint16 ): uint16; inline;
begin
  case Month of
    1,3,5,7,8,10,12: Result := 31;
    4,6,9,11: Result := 30;
    else if (Year mod 4)=0 then Result := 29 else Result := 28;
  end;
end;

function DaysInYear( const Year: uint16 ): uint16; inline;
var
  idx: uint32;
begin
  Result := 0;
  for idx := 1 to 12 do Result := Result + DaysInMonth( Year, idx );
end;

function EncodeDate( const Year, Month, Day: uint16 ): TDateTime; inline;
// Epoch for TDateTime = 1899/12/30
const
  cEpochDays = 693974; // 1899/12/30 in days.
var
  idx: uint16;
  TotalDays: nativeuint;
begin
  if (Year>9999) then TStatus(stInvalidYear).Raize( [ Year.AsString ] );
  if (Month<1) or (Month>12) then TStatus(stInvalidMonth).Raize( [ Month.AsString ] );
  if (Day<1) or (Day>DaysInMonth(Year,Month)) then TStatus(stInvalidDate).Raize( [ Day.AsString, Month.AsString, Year.AsString] );
  TotalDays := 0;
  if Year>0  then for idx := 0 to pred(Year) do TotalDays := TotalDays + DaysInYear(idx);
  if Month>1 then for idx := 1 to pred(Month) do TotalDays := TotalDays + DaysInMonth( Year, Idx );
  {$warnings off} TotalDays := TotalDays + pred(Day); {$warnings on}
  TotalDays := TotalDays - cEpochDays;
  Double(Result) := TotalDays;
end;

procedure DecodeDate( const DateTime: TDateTime; out Year, Month, Day: uint16 ); inline;
const
  cEpochDays = 693974; // 1899/12/30 in days.
var
  Days: uint16;
  TotalDays: nativeuint;
begin
  TotalDays := Trunc(DateTime) + cEpochDays + 1;
  Year := 0;
  repeat
    Days := DaysInYear(Year);
    if TotalDays>=Days then begin
      inc(Year);
      TotalDays := TotalDays - Days;
    end;
  until TotalDays<Days;
  Month := 1;
  repeat
    Days := DaysInMonth(Year,Month);
    if TotalDays>=Days then begin
      inc(month);
      TotalDays := TotalDays - Days;
    end;
  until TotalDays<Days;
  Day := TotalDays;
end;

function EncodeTime( const Hour, Minute, Second, Millisecond: uint16 ): TDateTime; inline;
const
  cMillisecondsPerSecond = 1000;
  cMillisecondsPerMinute = cMillisecondsPerSecond * 60;
  cMillisecondsPerHour = cMillisecondsPerMinute * 60;
  cMillisecondsPerDay = cMillisecondsPerHour * 24;
  cUnitaryDay = 1 / cMillisecondsPerDay;
begin
  Double(Result) := (
                      (Hour * cMillisecondsPerHour) +
                      (Minute*cMillisecondsPerMinute) +
                      (Second * cMillisecondsPerSecond) +
                      (Millisecond)
                    ) * cUnitaryDay;
end;

procedure DecodeTime( const DateTime: TDateTime; out Hour, Minute, Second, Millisecond: uint16 ); inline;
const
  cMillisecondsPerSecond = 1000;
  cMillisecondsPerMinute = cMillisecondsPerSecond * 60;
  cMillisecondsPerHour = cMillisecondsPerMinute * 60;
  cMillisecondsPerDay = cMillisecondsPerHour * 24;
var
  F: nativeint;
begin
  F := Trunc( (Frac(DateTime) * cMillisecondsPerDay ) );
  Hour := (F div cMillisecondsPerHour);
  F := F mod cMillisecondsPerHour;
  Minute := (F div cMillisecondsPerMinute);
  F := F mod cMillisecondsPerMinute;
  Second := (F div cMillisecondsPerSecond);
  Millisecond := F mod cMillisecondsPerSecond;
end;

function EncodeDateTime( const Year, Month, Day, Hour, Minute, Second, Millisecond: uint16 ): TDateTime; inline;
begin
  Result := EncodeDate( Year, Month, Day ) + EncodeTime( Hour, Minute, Second, Millisecond );
end;

procedure DecodeDateTime( const DateTime: TDateTime; out Year, Month, Day, Hour, Minute, Second, Millisecond: uint16 ); inline;
begin
  DecodeDate( DateTime, Year, Month, Day );
  DecodeTime( DateTime, Hour, Minute, Second, Millisecond );
end;

function IntegerToString( const Value: int64 ): string; inline;
begin
  {$warnings off} Str(value,Result); {$warnings on}
end;

function UnsignedIntegerToString( const Value: uint64 ): string; inline;
begin
  {$warnings off} Str(value,Result); {$warnings on}
end;

function FloatToString( const Value: extended ): string; inline;
begin
  {$warnings off} Str(value:10:10,Result); {$warnings on}
end;

procedure StringToInteger8( const src: string; out value: int8 ); inline;
var
  Code: int32;
begin
  Val(src,value,code);
  if Code<>0 then TStatus( stStrToInt8Error ).Raize( [src] );
end;

procedure StringToInteger16( const src: string; out value: int16 ); inline;
var
  Code: int32;
begin
  Val(src,value,code);
  if Code<>0 then TStatus( stStrToInt16Error ).Raize( [src] );
end;

procedure StringToInteger32( const src: string; out value: int32 ); inline;
var
  Code: int32;
begin
  Val(src,value,code);
  if Code<>0 then TStatus( stStrToInt32Error ).Raize( [src] );
end;

procedure StringToInteger64( const src: string; out value: int64 ); inline;
var
  Code: int32;
begin
  Val(src,value,code);
  if Code<>0 then TStatus( stStrToInt64Error ).Raize( [src] );
end;

procedure StringToUnsignedInteger8( const src: string; out value: uint8 ); inline;
var
  Code: int32;
begin
  Val(src,value,code);
  if Code<>0 then TStatus( stStrToUInt8Error ).Raize( [src] );
end;

procedure StringToUnsignedInteger16( const src: string; out value: uint16 ); inline;
var
  Code: int32;
begin
  Val(src,value,code);
  if Code<>0 then TStatus( stStrToUInt16Error ).Raize( [src] );
end;

procedure StringToUnsignedInteger32( const src: string; out value: uint32 ); inline;
var
  Code: int32;
begin
  Val(src,value,code);
  if Code<>0 then TStatus( stStrToUInt32Error ).Raize( [src] );
end;

procedure StringToUnsignedInteger64( const src: string; out value: uint64 ); inline;
var
  Code: int32;
begin
  Val(src,value,code);
  if Code<>0 then TStatus( stStrToUInt64Error ).Raize( [src] );
end;

procedure StringToSingle( const src: string; out value: single ); inline;
var
  Code: int32;
begin
  Val(src,value,code);
  if Code<>0 then TStatus( stStrToSingleError ).Raize( [src] );
end;

procedure StringToDouble( const src: string; out value: double ); inline;
var
  Code: int32;
begin
  Val(src,value,code);
  if Code<>0 then TStatus( stStrToDoubleError ).Raize( [src] );
end;

{$ifdef CPU32BITS}
procedure StringToExtended( const src: string; out value: extended ); inline;
var
  Code: int32;
begin
  Val(src,value,code);
  if Code<>0 then TStatus( stStrToExtendedError ).Raize( [src] );
end;
{$endif}

function Supports(const Instance: IInterface; const InterfaceID: TGUID; out Iface): boolean; overload;
begin
  Result := sysutils.Supports( Instance, InterfaceID, IFace );
end;

function Supports(const Instance: TObject; const InterfaceID: TGUID; out Iface): boolean; overload;
begin
  Result := sysutils.Supports( Instance, InterfaceID, IFace );
end;

function Supports(const Instance: IInterface; const InterfaceID: TGUID): boolean; overload;
begin
  Result := sysutils.Supports( Instance, InterfaceID );
end;

function Supports(const Instance: TObject; const InterfaceID: TGUID): boolean; overload;
begin
  Result := sysutils.Supports( Instance, InterfaceID );
end;

function Supports(const _Class: TClass; const InterfaceID: TGUID): boolean; overload;
begin
  Result := sysutils.Supports( _Class, InterfaceID );
end;


class operator utfCodepoint.Implicit( const a: AnsiChar ): utfCodepoint;
begin
  if not TUnicodeCodec.AnsiDecode(A,Result.Value) then begin
    TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utfANSI ) ] );
  end;
end;

class operator utfCodepoint.Implicit( const a: utfCodepoint ): AnsiChar;
var
  throwaway: uint8;
begin
  Result := Default( AnsiChar );
  if not TUnicodeCodec.AnsiEncode(a.Value,Result,throwaway) then begin
    TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utfANSI ) ] );
  end;
end;

class operator utfCodepoint.Implicit( const a: WideChar ): utfCodepoint;
begin
  if not TUnicodeCodec.UTF16LEDecode(A,Result.Value) then begin
    TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utf16LE ) ] );
  end;
end;

class operator utfCodepoint.Implicit( const a: utfCodepoint ): WideChar;
type
  TSurrogatePair = record
    a: uint16;
    b: uint16;
  end;
var
  bytes: TSurrogatePair;
  count: uint8;
begin
  Result := Default( WideChar );
  bytes := Default( TSurrogatePair );
  if not TUnicodeCodec.UTF16LEEncode( a.value, bytes, count ) then begin
    TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utf16LE ) ] );
  end;
  Result := WideChar(bytes.a);
end;

class operator utfCodepoint.Explicit( const a: AnsiChar ): utfCodepoint;
begin
  if not TUnicodeCodec.AnsiDecode( a, Result.Value ) then begin
    TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utfANSI ) ] );
  end;
end;

class operator utfCodepoint.Explicit( const a: utfCodepoint ): AnsiChar;
var
  throwaway: uint8;
begin
  Result := Default( AnsiChar );
  if not TUnicodeCodec.AnsiEncode( a.Value, Result, throwaway ) then begin
    TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utfANSI ) ] );
  end;
end;

class operator utfCodepoint.Explicit( const a: WideChar ): utfCodepoint;
begin
  if not TUnicodeCodec.UTF16LEDecode( a, Result.Value ) then begin
    TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utf16LE ) ] );
  end;
end;

class operator utfCodepoint.Explicit( const a: utfCodepoint ): WideChar;
type
  TSurrogatePair = record
    a: uint16;
    b: uint16;
  end;
var
  bytes: TSurrogatePair;
  count: uint8;
begin
  Result := Default( WideChar );
  Bytes := Default( TSurrogatePair );
  if not TUnicodeCodec.UTF16LEEncode( a.Value, bytes, count ) then begin
    TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utf16LE ) ] );
  end;
  Result := WideChar(bytes.a);
end;

class operator utfCodepoint.GreaterThan( const a: utfCodepoint; const b: utfCodepoint ): boolean;
begin
  Result := a.Value > b.Value;
end;

class operator utfCodepoint.GreaterThan( const a: utfCodepoint; const b: AnsiChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value > cp.Value;
end;

class operator utfCodepoint.GreaterThan( const a: utfCodepoint; const b: WideChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value > cp.Value;
end;

class operator utfCodepoint.GreaterThan( const a: AnsiChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value > b.Value;
end;

class operator utfCodepoint.GreaterThan( const a: WideChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value > b.Value;
end;

class operator utfCodepoint.GreaterThanOrEqual( const a: utfCodepoint; const b: utfCodepoint ): boolean;
begin
  Result := a.Value >= b.Value;
end;

class operator utfCodepoint.GreaterThanOrEqual( const a: utfCodepoint; const b: AnsiChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value >= cp.Value;
end;

class operator utfCodepoint.GreaterThanOrEqual( const a: utfCodepoint; const b: WideChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value >= cp.Value;
end;

class operator utfCodepoint.GreaterThanOrEqual( const a: AnsiChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value >= b.Value;
end;

class operator utfCodepoint.GreaterThanOrEqual( const a: WideChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value >= b.Value;
end;

class operator utfCodepoint.LessThan( const a: utfCodepoint; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value < CP.Value;
end;

class operator utfCodepoint.LessThan( const a: utfCodepoint; const b: AnsiChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value < CP.Value;
end;

class operator utfCodepoint.LessThan( const a: utfCodepoint; const b: WideChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value < CP.Value;
end;

class operator utfCodepoint.LessThan( const a: AnsiChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value < b.Value;
end;

class operator utfCodepoint.LessThan( const a: WideChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value < b.Value;
end;

class operator utfCodepoint.LessThanOrEqual( const a: utfCodepoint; const b: utfCodepoint ): boolean;
begin
  Result := a.Value <= b.Value;
end;

class operator utfCodepoint.LessThanOrEqual( const a: utfCodepoint; const b: AnsiChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value <= CP.Value;
end;

class operator utfCodepoint.LessThanOrEqual( const a: utfCodepoint; const b: WideChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value <= CP.Value;
end;

class operator utfCodepoint.LessThanOrEqual( const a: AnsiChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value <= b.Value;
end;

class operator utfCodepoint.LessThanOrEqual( const a: WideChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value <= b.Value;
end;

class operator utfCodepoint.Equal( const a: utfCodepoint; const b: utfCodepoint ): boolean;
begin
  Result := a.Value = b.Value;
end;

class operator utfCodepoint.Equal( const a: utfCodepoint; const b: AnsiChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value = CP.Value;
end;

class operator utfCodepoint.Equal( const a: utfCodepoint; const b: WideChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value = CP.Value;
end;

class operator utfCodepoint.Equal( const a: AnsiChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value = b.Value;
end;

class operator utfCodepoint.Equal( const a: WideChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value = b.Value;
end;

class operator utfCodepoint.NotEqual( const a: utfCodepoint; const b: utfCodepoint ): boolean;
begin
  Result := a.Value <> b.Value;
end;

class operator utfCodepoint.NotEqual( const a: utfCodepoint; const b: AnsiChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value <> CP.Value;
end;

class operator utfCodepoint.NotEqual( const a: utfCodepoint; const b: WideChar ): boolean;
var
  CP: utfCodepoint;
begin
  CP := b;
  Result := a.Value <> CP.Value;
end;

class operator utfCodepoint.NotEqual( const a: AnsiChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value <> b.Value;
end;


class operator utfCodepoint.NotEqual( const a: WideChar; const b: utfCodepoint ): boolean;
var
  CP: utfCodepoint;
begin
  CP := a;
  Result := CP.Value <> b.Value;
end;

procedure utfCodepoint.ReadMem(var P: pointer; const UnicodeFormat: TUnicodeFormat);
begin
  case UnicodeFormat of
    utfUnknown: TStatus( stUTFUnknownNotSupported ).Raize;
       utfANSI: if not TUnicodeCodec.AnsiDecode( P^, Self.Value )    then TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
          utf8: if not TUnicodeCodec.UTF8Decode( P^, Self.Value )    then TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
       utf16LE: if not TUnicodeCodec.UTF16LEDecode( P^, Self.Value ) then TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
       utf16BE: if not TUnicodeCodec.UTF16BEDecode( P^, Self.Value ) then TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
       utf32LE: if not TUnicodeCodec.UTF32LEDecode( P^, Self.Value ) then TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
       utf32BE: if not TUnicodeCodec.UTF32BEDecode( P^, Self.Value ) then TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
  end;
  {$hints off}
  P := pointer( nativeuint(p) + Self.Size( UnicodeFormat ) );
  {$hints on}
end;

function utfCodepoint.Size( const UnicodeFormat: TUnicodeFormat ): nativeuint;
var
  CP: TUnicodeCodePoint;
  B: uint8;
begin
  CP := Default( TUnicodeCodePoint );
  case UnicodeFormat of
    utfUnknown : TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utfUnknown ) ] );
    utfANSI    : B := sizeof(uint8);
    utf8       : if not    TUnicodeCodec.UTF8Encode( self.Value, CP, B ) then TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utf8    ) ] );
    utf16LE    : if not TUnicodeCodec.UTF16LEEncode( self.Value, CP, B ) then TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utf16LE ) ] );
    utf16BE    : if not TUnicodeCodec.UTF16BEEncode( self.Value, CP, B ) then TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( TUnicodeFormat.utf16BE ) ] );
    utf32LE,
    utf32BE    : B := sizeof(uint32);
  end;
  Result := B;
end;

function utfCodepoint.Uppercase: utfCodepoint;
var
  {$ifdef fpc}
  ch: unicodechar;
  {$else}
  ch: char;
  {$endif}
  s: string;
  size_encoded: uint8;
begin
  {$ifdef fpc}
  ch := Default( unicodechar );
  {$else}
  ch := Default( char );
  {$endif}
  if not TUnicodeCodec.UTF16LEEncode( Self.Value, ch, size_encoded ) then begin
    TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( utf16le ) ] );
  end;
  s := ch;
  s := SysUtils.Uppercase( s );
  {$ifdef NEXTGEN}
  ch := s[0];
  {$else}
  ch := s[1];
  {$endif}
  if not TUnicodeCodec.UTF16LEDecode( ch, Result.Value ) then begin
    TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( utf16le ) ] );
  end;
end;

procedure utfCodepoint.WriteMem(var P: pointer; const UnicodeFormat: TUnicodeFormat);
var
  B: uint8;
begin
  case UnicodeFormat of
    utfUnknown: TStatus( stUTFUnknownNotSupported ).Raize;
       utfANSI: if not TUnicodeCodec.AnsiEncode(Self.Value,P^,B)    then TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
          utf8: if not TUnicodeCodec.UTF8Encode(Self.Value,P^,B)    then TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
       utf16LE: if not TUnicodeCodec.UTF16LEEncode(Self.Value,P^,B) then TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
       utf16BE: if not TUnicodeCodec.UTF16BEEncode(Self.Value,P^,B) then TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
       utf32LE: if not TUnicodeCodec.UTF32LEEncode(Self.Value,P^,B) then TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
       utf32BE: if not TUnicodeCodec.UTF32BEEncode(Self.Value,P^,B) then TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( UnicodeFormat ) ] );
  end;
  {$hints off}
  P := pointer( nativeuint( P ) + B );
  {$hints on}
end;

function utfCodepoint.Lowercase: utfCodepoint;
var
  {$ifdef fpc}
  ch: unicodechar;
  {$else}
  ch: char;
  {$endif}
  s: string;
  size_encoded: uint8;
begin
  {$ifdef fpc}
  ch := Default( unicodechar );
  {$else}
  ch := Default( char );
  {$endif}
  if not TUnicodeCodec.UTF16LEEncode( Self.Value, ch, size_encoded ) then begin
    TStatus( stUnicodeEncodingFailed ).Raize( [ TUnicodeCodec.FormatToString( utf16le ) ] );
  end;
  s := ch;
  s := SysUtils.Lowercase( s );
  {$ifdef NEXTGEN}
  ch := s[0];
  {$else}
  ch := s[1];
  {$endif}
  if not TUnicodeCodec.UTF16LEDecode( ch, Result.Value ) then begin
    TStatus( stUnicodeDecodingFailed ).Raize( [ TUnicodeCodec.FormatToString( utf16le ) ] );
  end;
end;


class function TUtf8String.Alloc(const srcStr: string): pointer;
var
  idx: nativeuint;
  L: nativeuint;
  CP: utfCodepoint;
  P: pointer;
begin
  Result := nil;
  if srcStr='' then exit;
  // Measure source string in utf-8 and allocate space to store it.
  L := 0;
  {$ifdef NEXTGEN}
  for idx := 0 to pred(Length(srcStr)) do begin
  {$else}
  for idx := 1 to Length(srcStr) do begin
  {$endif}
    CP := srcStr[idx];
    L := L + CP.Size(utf8)
  end;
  inc(L); // space for zero terminator.
  {$ifndef fpc}
  GetMem(Result,L);
  {$else}
  Result := GetMem(L);
  {$endif}
  FillChar(Result^,L,0);
  //- Convert the string to UTF-8
  P := Result;
  {$ifdef NEXTGEN}
  for idx := 0 to pred(Length(srcStr)) do begin
  {$else}
  for idx := 1 to Length(srcStr) do begin
  {$endif}
    CP := srcStr[idx];
    CP.WriteMem( P, utf8 );
  end;
end;

class function TUtf8String.Size(const srcStr: pointer): nativeuint;
var
  L: nativeuint;
  CP: utfCodepoint;
  P: pointer;
begin
  Result := 0;
  if srcStr=nil then exit;
  L := 0;
  P := srcStr;
  while uint8(P^)<>0 do begin
    CP.ReadMem(P,utf8);
    L := L + CP.Size( utf8 );
  end;
  Result := succ(L); // for zero term.
end;

class procedure TUtf8String.Free(const srcStr: pointer);
begin
  if srcStr=nil then exit;
  FreeMem(srcStr,TUtf8String.Size(srcStr));
end;

class function TUtf8String.AsString(const srcStr: pointer): string;
var
  P: pointer;
  CP: utfCodepoint;
begin
  Result := '';
  if srcStr=nil then exit;
  P := srcStr;
  while uint8(P^)<>0 do begin
    CP.ReadMem( P, utf8 );
    Result := Result + char( CP );
  end;
end;

class function TUtf16LEString.Alloc(const srcStr: string): pointer;
var
  idx: nativeuint;
  L: nativeuint;
  CP: utfCodepoint;
  P: pointer;
begin
  Result := nil;
  if srcStr='' then exit;
  // Measure source string in utf-16 and allocate space to store it.
  L := 0;
  {$ifdef NEXTGEN}
  for idx := 0 to pred(Length(srcStr)) do begin
  {$else}
  for idx := 1 to Length(srcStr) do begin
  {$endif}
    CP := srcStr[idx];
    L := L + CP.Size(utf16LE)
  end;
  inc(L,2); // space for zero terminator.
  {$ifndef fpc}
  GetMem(Result,L);
  {$else}
  Result := GetMem(L);
  {$endif}
  FillChar(Result^,L,0);
  //- Convert the string to UTF-16
  P := Result;
  {$ifdef NEXTGEN}
  for idx := 0 to pred(Length(srcStr)) do begin
  {$else}
  for idx := 1 to Length(srcStr) do begin
  {$endif}
    CP := srcStr[idx];
    CP.WriteMem( P, utf16LE );
  end;
end;

class function TUtf16LEString.Size(const srcStr: pointer): nativeuint;
var
  L: nativeuint;
  CP: utfCodepoint;
  P: pointer;
begin
  Result := 0;
  if srcStr=nil then exit;
  L := 0;
  P := srcStr;
  while uint16(P^)<>0 do begin
    CP.ReadMem(P,utf16LE);
    L := L + CP.Size( utf16LE );
  end;
  inc(L,2); // for zero term.
  Result := L;
end;

class procedure TUtf16LEString.Free(const srcStr: pointer);
begin
  if srcStr=nil then exit;
  FreeMem(srcStr,TUtf16LEString.Size(srcStr));
end;

class function TUtf16LEString.AsString(const srcStr: pointer): string;
var
  P: pointer;
  CP: utfCodepoint;
begin
  Result := '';
  if srcStr=nil then exit;
  P := srcStr;
  while uint16(P^)<>0 do begin
    CP.ReadMem( P, utf16LE );
    Result := Result + char( CP );
  end;
end;

const
  cStatusMessageGranularity = 32;

type
  TStatusMessage = record
    GUID: TGUID;
    Text: string;
  private
    class procedure ExtractPlaceholders( const SourceString: string; var Placeholders: TArrayOfString ); static;
  public
    class procedure AddOrUpdateEntry( const GUID: TGUID; const Text: string ); static;
    class function Parameterize( const GUID: TGUID; const Parameters: array of string ): string; static;
  end;

(* Need to thread-protect this global singleton
   without creating a dependency on cwThreading,
   so doing own inclusion of OS critical section
*)

type
  TStatusMessages = record
  private
    Mutex: {$ifdef MSWINDOWS} SRWLOCK; {$else} pthread_mutex_t; {$endif}
    Messages: array of TStatusMessage;
    MessageCount: nativeuint;
  private
    procedure Acquire;
    procedure Release;
  public
    function FindByGUID( const GUID: TGUID; out FoundIdx: nativeuint ): boolean;
    function ItemByIndex( const idx: nativeuint ): string;
    procedure AddOrUpdateEntry( const GUID: TGUID; const Text: string );
  public
    class function Create: TStatusMessages; static;
    procedure Free;
  end;

var
  xStatusMessages: TStatusMessages;

function TStatusMessages.FindByGUID( const GUID: TGUID; out FoundIdx: nativeuint ): boolean;
var
  idx: nativeuint;
begin
  Result := False;
  if MessageCount=0 then exit;
  for idx := 0 to pred(MessageCount) do begin
    if IsEqualGUID(Messages[idx].GUID,GUID) then begin
      FoundIdx := idx;
      Result := True;
      exit;
    end
  end;
end;

function TStatusMessages.ItemByIndex(const idx: nativeuint): string;
begin
  Result := Messages[idx].Text;
end;

procedure TStatusMessages.AddOrUpdateEntry( const GUID: TGUID; const Text: string );
var
  EntryIdx: nativeuint;
begin
  Acquire;
  try
    if xStatusMessages.FindByGUID( GUID, EntryIdx ) then begin
      Messages[EntryIdx].Text := Text;
    end else begin
      {$warnings off}
      if MessageCount = Length( Messages ) then begin
      {$warnings on}
        SetLength( Messages, Length(Messages) + cStatusMessageGranularity );
      end;
      Messages[MessageCount].GUID := GUID;
      Messages[MessageCount].Text := Text;
      inc(MessageCount);
    end;
  finally
    Release;
  end;
end;

class function TStatusMessages.Create: TStatusMessages;
begin
  {$warnings off} SetLength( Result.Messages,0 ); {$warnings on}
  Result.MessageCount := 0;
  {$ifdef MSWINDOWS}
  InitializeSRWLock(Result.Mutex);
  {$else}
  if pthread_mutex_init(@Result.Mutex, nil)<>0 then TStatus(stCreateMutexFailed).Raize([GetLastError().AsString]);
  {$endif}
end;

procedure TStatusMessages.Free;
begin
  {$ifndef MSWINDOWS}
  pthread_mutex_destroy(@Mutex);
  {$endif}
  SetLength( Messages,0 );
  MessageCount := 0;
end;

procedure TStatusMessages.Acquire;
begin
  {$ifdef MSWINDOWS}
  AcquireSRWLockExclusive(Mutex);
  {$else}
  pthread_mutex_lock(@Mutex);
  {$endif}
end;

procedure TStatusMessages.Release;
begin
  {$ifdef MSWINDOWS}
  ReleaseSRWLockExclusive(Mutex);
  {$else}
  pthread_mutex_unlock(@Mutex)
  {$endif}
end;

class procedure TStatusMessage.ExtractPlaceholders( const SourceString: string; var Placeholders: TArrayOfString );
type
  TParseStage = ( psCounting, psExtracting, psDone );
  TParseState = ( psNormal, psBra, psCollecting, psClosing );
var
  ltSourceString: string;
  ParseStage: TParseStage;
  ParseState: TParseState;
  PlaceholderCount: nativeuint;
  CurrentName: string;
  idx: nativeuint;
  top: nativeuint;
begin
  ltSourceString := SourceString.LowercaseTrim;
  {$ifdef NEXTGEN} top := pred(Length(ltSourceString)); {$else} top := Length(ltSourceString); {$endif}

  //* We're going to use a state machine to parse and extract the placeholders.
  //* The state machine has two stages, first psCounting which counts the placeholders,
  //* and second psExtracting which extracts the placeholders into an array.
  ParseStage := psCounting;
  repeat

    // Initialize parsing stage.
    {$ifdef NEXTGEN} idx := 0; {$else} idx := 1; {$endif}
    ParseState := psNormal;
    PlaceholderCount := 0;

    // Each stage loops idx to top.
    while (idx<=top) do begin
      case ParseState of

        psNormal: begin
          if ltSourceString[idx]='(' then begin
            ParseState := psBra;
          end;
        end;

        psBra: begin
          if ltSourceString[idx]='%' then begin
            ParseState := psCollecting;
            if ParseStage=psExtracting then begin
              CurrentName := '';
            end;
          end else begin
            if ltSourceString[idx]<>'(' then begin
              ParseState := psNormal;
            end;
          end;
        end;

        psCollecting: begin
          if ltSourceString[idx]='%' then begin
            ParseState := psClosing;
          end else if ParseStage=psExtracting then begin
              CurrentName := CurrentName + ltSourceString[idx];
          end;
        end;

        psClosing: begin
          if ltSourceString[idx]=')' then begin
            if ParseStage=psExtracting then begin
              Placeholders[PlaceholderCount] := CurrentName;
            end;
            inc(PlaceholderCount);
            ParseState := psNormal;
          end else begin
            if ParseStage=psExtracting then begin
              CurrentName := CurrentName + '%';
            end;
            ParseState := psCollecting;
          end;
        end;
      end;
      inc(idx);
    end;

    //- Advance the stage.
    case ParseStage of

      psCounting: begin
        SetLength(Placeholders,PlaceholderCount);
        if PlaceholderCount=0 then begin
          ParseStage := psDone;
        end else begin
          ParseStage := psExtracting;
        end;
      end;

      psExtracting: ParseStage := psDone;
    end;

  until ParseStage=psDone;
end;

class function TStatusMessage.Parameterize( const GUID: TGUID; const Parameters: array of string ): string;
var
  FoundIdx: nativeuint;
  Count: nativeuint;
  idx: nativeuint;
  Placeholders: TArrayOfString;
begin
  Result := '???';
  //- Locate the entry text.
  xStatusMessages.Acquire;
  try
    if not xStatusMessages.FindByGUID( GUID, FoundIdx ) then exit;
    Result := xStatusMessages.ItemByIndex(FoundIdx);
  finally
    xStatusMessages.Release;
  end;
  if Length(Parameters)=0 then exit;
  //- Extract the placeholders.
  {$hints off}
  ExtractPlaceholders( Result, Placeholders );
  {$hints on}
  if Length(Placeholders)=0 then exit;
  //- Loop the number of placeholders / number of parameters whichever is the lesser.
  Count := Length(Placeholders);
  {$warnings off}
  if Length( Parameters ) < Count then Count := Length(Parameters);
  {$warnings on}
  for idx := 0 to pred(Count) do begin
    Result := Result.ReplaceAll( '(%' + Placeholders[ idx ] + '%)', Parameters[ idx ], TCaseSensitivity.caseInsensitive );
  end;
end;

class procedure TStatusMessage.AddOrUpdateEntry( const GUID: TGUID; const Text: string );
begin
  xStatusMessages.AddOrUpdateEntry( GUID, Text );
end;

function EStatus.getStatus: TStatus;
begin
  Result := fStatus;
end;

constructor EStatus.Create( const Status: TStatus );
var
  S: string;
begin
  S := Status.AsString;
  S.ReplaceAll('"','\"');
  {$ifdef fpc}
  inherited Create( S.AsAnsiString );
  {$else}
  inherited Create( S );
  {$endif}
  fStatus := Status;
end;

constructor EStatus.Create( const Status: TStatus; const Parameters: array of string );
var
  S: string;
begin
  S := Status.AsString( Parameters );
  S.ReplaceAll('"','\"');
  {$ifdef fpc}
  inherited Create( S.AsAnsiString );
  {$else}
  inherited Create( S );
  {$endif}
  fStatus := Status;
end;

class operator TStatus.Implicit(const a: TStatus): TGUID;
begin
  Result := a.GUID;
end;

class operator TStatus.Explicit(const a: TStatus): TGUID;
begin
  Result := a.GUID;
end;

class operator TStatus.Implicit(const a: TGUID): TStatus;
begin
  Result.GUID := a;
end;

class operator TStatus.Explicit(const a: TGUID): TStatus;
begin
  Result.GUID := a;
end;

class operator TStatus.Implicit(const a: TStatus): Boolean;
begin
  Result := IsEqualGUID(stSuccess,a.GUID);
end;

class operator TStatus.Explicit(const a: TStatus): Boolean;
begin
  Result := IsEqualGUID(stSuccess,a.GUID);
end;

class operator TStatus.LogicalNot(const a: TStatus): Boolean;
begin
  Result := not IsEqualGUID(a.GUID,stSuccess);
end;

class operator TStatus.Equal(const a: TStatus; const b: TStatus): boolean;
begin
  Result := IsEqualGUID(a.GUID,b.GUID);
end;

class operator TStatus.NotEqual(const a: TStatus; const b: TStatus): boolean;
begin
  Result := not IsEqualGUID(a.GUID,b.GUID);
end;

class operator TStatus.Equal(const a: TGUID; const b: TStatus): boolean;
begin
  Result := IsEqualGUID(a,b.GUID);
end;

class operator TStatus.Equal(const a: TStatus; const b: TGUID): boolean;
begin
  Result := IsEqualGUID(a.GUID,b);
end;

class operator TStatus.NotEqual(const a: TGUID; const b: TStatus): boolean;
begin
  Result := not IsEqualGUID(a,b.GUID);
end;

class operator TStatus.NotEqual(const a: TStatus; const b: TGUID): boolean;
begin
  Result := not IsEqualGUID(a.GUID,b);
end;

class function TStatusHeler.Unknown: TStatus;
begin
  Result.GUID := stUnknown;
end;

class function TStatusHeler.Success: TStatus;
begin
  Result.GUID := stSuccess;
end;

function TStatusHeler.AsString( const Parameters: array of string ): string;
begin
  Result := TStatusMessage.Parameterize(Self.GUID,Parameters);
end;

function TStatusHeler.AsString: string;
begin
  Result := Self.AsString([]);
end;

procedure TStatusHeler.Raize(const Parameters: array of string);
begin
  {$hints off} if IsEqualGUID(GUID,stSuccess) then exit; {$hints on}
  {$ifdef fpc}
  raise
    EStatus.Create(Self,Parameters);
  {$else}
  raise
    EStatus.Create(Self,Parameters);
  {$endif}
end;

procedure TStatusHeler.Raize;
begin
  Raize([]);
end;

procedure TStatusHeler.Register(const a: string);
begin
  TStatusMessage.AddOrUpdateEntry( Self.GUID, a );
end;

function TBooleanHelper.getAsString: string;
begin
   if Self then begin
     Result := 'TRUE';
   end else begin
     Result := 'FALSE';
   end;
end;

procedure TBooleanHelper.setAsString(const Value: string);
begin
   if Value.UppercaseTrim='TRUE' then begin
     Self := True;
   end else begin
     Self := False;
   end;
end;

function TStringHelper.AsBoolean(const CaseSensitive: boolean; const positive: string): boolean;
begin
  if CaseSensitive then begin
    Result := self = positive;
  end else begin
    Result := self.Uppercase = positive.Uppercase;
  end;
end;

function TStringHelper.AsDouble: double;
begin
  StringToDouble( Self, Result );
end;

function TStringHelper.AsExtended: extended;
{$ifdef CPU64BITS}
var
  D: double;
{$endif}
begin
  {$ifdef CPU32BITS}
    StringToExtended( Self, Result );
  {$else}
    {$ifdef CPU64BITS}
    StringToDouble( Self, D );
    Result := D;
    {$else}
      .ErrorBuilding-how many bits is the target cpu?
    {$endif}
  {$endif}
end;

function TStringHelper.AsInt16: int16;
begin
  StringToInteger16( Self, Result );
end;

function TStringHelper.AsInt32: int32;
begin
  StringToInteger32( Self, Result );
end;

function TStringHelper.AsInt64: int64;
begin
  StringToInteger64( Self, Result );
end;

function TStringHelper.AsInt8: int8;
begin
  StringToInteger8( Self, Result );
end;

function TStringHelper.AsSingle: single;
begin
  StringToSingle( Self, Result );
end;

function TStringHelper.AsUInt16: uint16;
begin
  StringToUnsignedInteger16( Self, Result );
end;

function TStringHelper.AsUint32: uint32;
begin
  StringToUnsignedInteger32( Self, Result );
end;

function TStringHelper.AsUint64: uint64;
begin
  StringToUnsignedInteger64( Self, Result );
end;

function TStringHelper.AsUInt8: uint8;
begin
  StringToUnsignedInteger8( Self, Result );
end;

function TStringHelper.Combine(const cDelimiter: string; const Parts: TArrayOfString; const StartIdx: nativeuint = 0): string;
var
  idx: nativeuint;
  Len: nativeuint;
begin
  Self := '';
  Result := '';
  Len := Length(Parts);
  if Len=0 then begin
    Exit;
  end;
  if StartIdx>pred(Len) then begin
    Exit;
  end;
  for idx := StartIdx to pred(Len) do begin
    Self := Self + Parts[idx];
    if idx<pred(Len) then begin
      Self := Self + cDelimiter;
    end;
  end;
  Result := Self;
end;

function TStringHelper.Contains(const SubString: string; const CaseSensitivity: TCaseSensitivity): boolean;
var
  DiscardStart: nativeuint;
  DiscardStop: nativeuint;
begin
  Result := Self.Find(SubString,DiscardStart,DiscardStop,CaseSensitivity);
end;

function TStringHelper.Explode(const cDelimiter: string ): TArrayOfString;
const
  cGranularity = 8;
var
  s: string;
  start,stop: nativeuint;
  i: nativeuint;
begin
  s:= Self;
  i:= 0;
  {$warnings off} SetLength(Result,0); {$warnings on}
  try
    repeat
      // Ensure there is space for the next item.
      {$warnings off}
      if i>=Length( Result ) then begin
      {$warnings on}
        SetLength(Result,Length(Result)+cGranularity);
      end;
      //- Find and add the next item
      if s.Find(cDelimiter,start,stop) then begin
        Result[ i ] := S.Left(pred(start));
        S := S.Right( S.Len - stop );
      end else begin
        Result[ i ] := S;
        break;
      end;
      inc(i);
    until False;
  finally
    SetLength(Result,succ(i));
  end;
end;

{$if defined(fpc) or defined(MSWINDOWS)}
function TStringHelper.getAsAnsiString: ansiString;
var
  Cursor: int32;
  Stop: int32;
  CP: TUnicodeCodepoint;
  B: uint8;
  S: uint8;
begin
  Result := '';
  if Length(Self)=0 then begin
    exit;
  end;
  Cursor := 1;
  Stop := succ(Length(Self));
  {$ifdef NEXTGEN}{$ifndef LINUX}
  Cursor := 0;
  Stop := Length(Self));
  {$endif}{$endif}
  CP := 0;
  S := 0;
  B := 0;
  while (Cursor<Stop) do begin
    if not TUnicodeCodec.DecodeCodepointFromString(CP,Self,Cursor) then begin
      exit;
    end;
    if not TUnicodeCodec.AnsiEncode(CP,B,S) then begin
      exit;
    end;
    Result := Result + ansichar(B);
  end;
end;
{$endif}

function TStringHelper.LeftPad( const CharCount: nativeuint; const ch: char ): string;
begin
  Result := Self;
  {$warnings off}
  if CharCount <= Length( Result ) then exit;
  while Length( Result ) < CharCount do begin
  {$warnings on}
    Result := ch + Result;
  end;
end;

function TStringHelper.Left(const ACount: nativeuint): string;
var
  idx: nativeuint;
  top: nativeuint;
begin
  Result := '';
  if Length(Self)=0 then exit;
  {$ifdef NEXTGEN}
  idx := 0;
  top := pred(Length(Self));
  {$else}
  idx := 1;
  top := Length(Self);
  {$endif}
  while (idx<=top) and {$ifdef NEXTGEN}(idx<aCount){$else}(idx<=aCount){$endif} do begin
    Result := Result + Self[idx];
    inc(idx);
  end;
end;

function TStringHelper.Len: nativeuint;
begin
  Result := Length(Self);
end;

function TStringHelper.Lowercase: string;
var
  idx: nativeuint;
  CP: utfCodepoint;
begin
  Result := '';
  if Length(Self)=0 then exit;
  {$ifdef NEXTGEN}
  for idx := 0 to pred(Length(Self)) do begin
  {$else}
  for idx := 1 to Length(Self) do begin
  {$endif}
    CP := Self[idx];
    Result := Result + char( CP.Lowercase );
  end;
end;

function TStringHelper.LowercaseTrim: string;
begin
  Result := Self.Lowercase.Trim;
end;

function TStringHelper.Copy(const AStart, ACount: nativeuint): string;
var
  charPos: nativeuint;
  L: nativeuint;
begin
  Result := '';
  L := Length(Self);
  if L<AStart+ACount then begin
    exit;
  end;
  CharPos := AStart;
  while CharPos<(AStart+ACount) do begin
    Result := Result + Self[CharPos];
    inc(CharPos);
  end;
end;

function TStringHelper.Padleft(const Padding: uint32; const PaddingChar: char): string;
var
  StrPadding: string;
  idx: uint32;
begin
  if Padding=0 then begin
    Result := Self;
    Exit;
  end;
  StrPadding := '';
  for idx := 0 to pred(Padding) do begin
    StrPadding := StrPadding + PaddingChar;
  end;
  Result := StrPadding + Self;
end;

function TStringHelper.PadRight(const Padding: uint32; const PaddingChar: char): string;
var
  StrPadding: string;
  idx: uint32;
begin
  if Padding=0 then begin
    Result := Self;
    Exit;
  end;
  StrPadding := '';
  for idx := 0 to pred(Padding) do begin
    StrPadding := StrPadding + PaddingChar;
  end;
  Result := Self + StrPadding;
end;

function TStringHelper.RightPad( const CharCount: nativeuint; const ch: char ): string;
begin
  Result := Self;
  {$warnings off}
  if CharCount <= Length( Result ) then exit;
  while Length( Result ) < CharCount do begin
  {$warnings on}
    Result := Result + ch;
  end;
end;

function TStringHelper.Right(const ACount: nativeuint): string;
var
  idx: nativeuint;
  top: nativeuint;
begin
  Result := '';
  if Length(Self)=0 then exit;
  {$ifdef NEXTGEN}
  idx := 0;
  top := pred(Length(Self));
  {$else}
  idx := 1;
  top := Length(Self);
  {$endif}
  while (idx<=top) and {$ifdef NEXTGEN}(idx<aCount){$else}(idx<=aCount){$endif} do begin
    Result := Result + Self[top-(ACount-idx)];
    inc(idx);
  end;
end;

{$if defined(fpc) or defined(MSWINDOWS)}
procedure TStringHelper.setAsAnsiString(const value: ansistring);
var
  idx: uint32;
  CP: TUnicodeCodepoint;
  B: uint8;
  L: uint32;
begin
  Self := '';
  L := Length(Value);
  if L=0 then begin
    exit;
  end;
  CP := 0;
  for idx := 1 to L do begin
    B := ord(value[idx]);
    if not TUnicodeCodec.AnsiDecode(B,CP) then begin
      exit;
    end;
    if not TUnicodeCodec.EncodeCodepointToString(CP,Self) then begin
      exit;
    end;
  end;
end;
{$endif}

procedure TStringHelper.SetAsPChar(const value: pointer);
var
  p: ^byte;
  slen: uint64;
  CP: TUnicodeCodePoint;
begin
  if Value=nil then begin
    Self := '';
    exit;
  end;
  CP := 0;
  p := Value;
  slen := 0;
  while (slen<65535) and (p^<>0) do begin
    TUnicodeCodec.AnsiDecode(p^,CP);
    TUnicodeCodec.EncodeCodepointToString(CP,Self);
    inc(p);
    inc(slen);
  end;
end;

function TStringHelper.Find(const SubString: string; out Start: nativeuint; out Stop: nativeuint; const CaseSensitivity: TCaseSensitivity): boolean;
var
  idx: nativeuint;
  top: nativeuint;
  pidx: nativeuint;
  ptop: nativeuint;
  InPattern: boolean;
  searchString: string;
  sourceString: string;
begin
  Result := False;
  Start := 0;
  Stop := 0;
  if Length(SubString)=0 then exit;
  if Length(Self)=0 then exit;
  if CaseSensitivity=TCaseSensitivity.caseInsensitive then begin
    searchString := SubString.Uppercase;
    sourceString := Self.Uppercase;
  end else begin
    searchString := SubString;
    sourceString := Self;
  end;
  {$ifdef NEXTGEN}
  idx := 0;
  top := pred(Length(sourceString));
  pidx := 0;
  ptop := pred(Length(searchString));
  {$else}
  idx := 1;
  top := Length(sourceString);
  pidx := 1;
  ptop := Length(searchString);
  {$endif}
  InPattern := False;
  while (not Result) and (idx<=top) do begin
    if InPattern then begin
      //- We're in the search pattern, so cancel if appropriate.
      if sourceString[idx]<>searchString[pidx] then begin
        Start := 0;
        {$ifdef NEXTGEN}
        pidx := 0;
        {$else}
        pidx := 1;
        {$endif}
        InPattern := False;
      end else begin
        inc(pidx);
        if pidx>ptop then begin
          Stop := idx;
          Result := True;
        end;
      end;
    end else begin
      //- We're not in the pattern so check to see if the pattern can start.
      if sourceString[idx]=searchstring[pidx] then begin
        Start := idx;
        if Length(searchString)=1 then begin
          Stop := Start;
          Result := True;
        end;
        inc(pidx);
        InPattern := True;
      end;
    end;
    inc(idx);
  end;
end;

function TStringHelper.Replace( const OldPattern, NewPattern: string; const CaseSensitivity: TCaseSensitivity = TCaseSensitivity.caseSensitive ): string;
var
  Start: nativeuint;
  Stop: nativeuint;
  L: string;
  R: string;
begin
  if Self.Find(OldPattern,Start,Stop,CaseSensitivity) then begin
    L := Self.Left( pred( Start ) );
    {$warnings off} R := Self.Right( Length( Self ) - Stop ); {$warnings on}
    Self := L + NewPattern + R;
  end;
  Result := Self;
end;

function TStringHelper.ReplaceAll(const OldPattern, NewPattern: string; const CaseSensitivity: TCaseSensitivity): string;
begin
  if CaseSensitivity = caseInsensitive then begin
    {$ifdef fpc}
    Self := SysUtils.StringReplace(Self.AsAnsiString,OldPattern.AsAnsiString,NewPattern.AsAnsiString,[rfReplaceAll,rfIgnoreCase]).AsString;
    {$else}
    Self := SysUtils.StringReplace(Self,OldPattern,NewPattern,[rfReplaceAll,rfIgnoreCase]);
    {$endif}
  end else begin
    {$ifdef fpc}
    Self := SysUtils.StringReplace(Self.AsAnsiString,OldPattern.AsAnsiString,NewPattern.AsAnsiString,[rfReplaceAll]).AsString;
    {$else}
    Self := SysUtils.StringReplace(Self,OldPattern,NewPattern,[rfReplaceAll]);
    {$endif}
  end;
  Result := Self;
end;

function TStringHelper.Format(const args: array of const): string;
begin
  {$ifdef fpc}
  Self := SysUtils.Format(Self.AsAnsiString,args).AsString;
  {$else}
  Self := SysUtils.Format(Self,args);
  {$endif}
  Result := Self;
end;

function TStringHelper.TrimLeft( const Whitespace: array of char ): string;
var
  idx: nativeuint;
  Count: nativeuint;
begin
  Result := '';
  if Length(Self)=0 then exit;
  Result := Self;
  //- count whitespace characters left.
  Count := 0;
  {$ifdef NEXTGEN}
  for idx := 0 to pred(Length(Result)) do begin
  {$else}
  for idx := 1 to Length(Result) do begin
  {$endif}
    if not (Result[idx].CharInArray(Whitespace)) then break;
    inc(Count);
  end;
  {$warnings off}
  if Count=Length(Result) then begin
  {$warnings on}
    Result := '';
    exit;
  end else begin
    {$warnings off} Result := Result.Right(Length(Result)-Count); {$warnings on}
  end;
end;

function TStringHelper.TrimLeft: string;
begin
  Result := Self.TrimLeft( cDefaultWhitespace );
end;

function TStringHelper.TrimRight( const Whitespace: array of char ): string;
var
  idx: nativeuint;
  Count: nativeuint;
begin
  Result := '';
  if Length(Self)=0 then exit;
  Result := Self;
  //- count whitespace characters left.
  Count := 0;
  {$ifdef NEXTGEN}
  for idx := pred(Length(Result)) downto 0 do begin
  {$else}
  for idx := Length(Result) downto 1 do begin
  {$endif}
    if not (Result[idx].CharInArray(Whitespace)) then break;
    inc(Count);
  end;
  {$warnings off}
  if Count = Length( Result ) then begin
  {$warnings on}
    Result := '';
    exit;
  end else begin
    {$warnings off} Result := Result.Left(Length(Result)-Count); {$warnings on}
  end;
end;

function TStringHelper.TrimRight: string;
begin
  Result := Self.TrimRight( cDefaultWhitespace );
end;


function TStringHelper.Trim(const Whitespace: array of char): string;
begin
  Result := Self.TrimLeft(Whitespace).TrimRight(WhiteSpace);
end;

function TStringHelper.Trim: string;
begin
  Result := Self.TrimLeft.TrimRight;
end;

function TStringHelper.Uppercase: string;
var
  idx: nativeuint;
  CP: utfCodepoint;
begin
  Result := '';
  if Length(Self)=0 then exit;
  {$ifdef NEXTGEN}
  for idx := 0 to pred(Length(Self)) do begin
  {$else}
  for idx := 1 to Length(Self) do begin
  {$endif}
    CP := Self[idx];
    Result := Result + char( CP.Uppercase );
  end;
end;

function TStringHelper.UppercaseTrim: string;
begin
  Result := Self.Uppercase.Trim;
end;

{$if defined(fpc) or defined(MSWINDOWS)}
function TAnsiStringHelper.getAsString: string;
begin
  Result := '';
  Result.AsAnsiString := Self;
end;
{$endif}


function TCharHelper.AsHex: string;
var
  Codepoint: TUnicodeCodepoint;
begin
  TUnicodeCodec.UTF16LEDecode( Self, Codepoint );
  Result := Codepoint.AsHex;
end;

function TCharHelper.CharInArray( const Characters: array of char ): boolean;
var
  idx: uint32;
begin
  Result := False;
  if Length(Characters)=0 then begin
    exit;
  end;
  for idx := 0 to pred(length(Characters)) do begin
    if self = Characters[idx] then begin
      Result := True;
      Exit;
    end;
  end;
end;

function TCharHelper.CharInSet( const Characters: TSetOfChar ): boolean;
begin
  Result := SysUtils.CharInSet( Self, Characters );
end;

class function TNativeUIntHelper.FromPointer( const value: pointer ): nativeuint;
begin
  {$hints off} Result := nativeuint( value ); {$hints on} // FPC Compiler warns that conversion between pointers and ordinals is not portable, but this is intentional.
end;

function TNativeUIntHelper.AsPointer: pointer;
begin
  {$hints off} Result := pointer(self); {$hints on} // FPC Compiler warns that conversion between pointers and ordinals is not portable, but this is intentional.
end;

procedure TNativeUIntHelper.setAsString( const value: string );
{$ifndef fpc}
var
  R: {$ifdef CPU32BITS} uint32; {$else} uint64; {$endif}
{$endif}
begin
  {$ifdef CPU32BITS}
  StringToUnsignedInteger32( value, {$ifdef fpc} Self {$else} R {$endif} );
  {$else}
  StringToUnsignedInteger64( value, {$ifdef fpc} Self {$else} R {$endif} );
  {$endif}
  {$ifndef fpc} Self := R; {$endif}
end;

function TNativeUIntHelper.getAsString: string;
begin
  Result := UnsignedIntegerToString( Self );
end;


procedure TNativeIntHelper.setAsString( const value: string );
{$ifndef fpc}
var
  R: {$ifdef CPU32BITS} int32; {$else} int64; {$endif}
{$endif}
begin
  {$ifdef CPU32BITS}
  StringToInteger32( value, {$ifdef fpc} Self {$else} R {$endif} );
  {$else}
  StringToInteger64( value, {$ifdef fpc} Self {$else} R {$endif} );
  {$endif}
  {$ifndef fpc} Self := R; {$endif}
end;


function TNativeIntHelper.getAsString: string;
begin
  Result := IntegerToString( Self )
end;

function TUInt32Helper.getAsHex: string;
begin
  Result := HexToString( Self, sizeof(uint32) );
end;

procedure TUInt32Helper.setAsHex( const value: string );
begin
  StringToHex( value, self, sizeof(uint32) );
end;

procedure TUInt32Helper.setAsString( const value: string );
begin
  StringToUnsignedInteger32( value, Self );
end;

function TUInt32Helper.getAsString: string;
begin
  Result := UnsignedIntegerToString(Self);
end;


function TUInt64Helper.getAsHex: string;
begin
  Result := HexToString( Self, sizeof(uint64) );
end;

procedure TUInt64Helper.setAsHex(const value: string);
begin
  StringToHex( value, self, sizeof(uint64) );
end;

procedure TUInt64Helper.setAsString( const value: string );
begin
  StringToUnsignedInteger64( value, Self );
end;

function TUInt64Helper.getAsString: string;
begin
  Result := UnsignedIntegerToString(Self);
end;


function TInt32Helper.getAsHex: string;
begin
  Result := HexToString( Self, sizeof(uint32) );
end;

procedure TInt32Helper.setAsHex(const value: string);
begin
  StringToHex( value, self, sizeof(uint32) );
end;

procedure TInt32Helper.setAsString( const value: string );
begin
  StringToInteger32( value, Self );
end;

function TInt32Helper.getAsString: string;
begin
  Result := IntegerToString(Self);
end;


function TInt64Helper.getAsHex: string;
begin
  Result := HexToString( Self, sizeof(uint64) );
end;

procedure TInt64Helper.setAsHex(const value: string);
begin
  StringToHex( value, self, sizeof(uint64) );
end;

procedure TInt64Helper.setAsString( const value: string );
begin
  StringToInteger64( value, Self );
end;

function TInt64Helper.getAsString: string;
begin
  Result := IntegerToString(Self);
end;


function TSingleHelper.Approximates(const Value, Precision: single): boolean;
begin
  Result := (Self>(Value-Precision)) and (Self<(Value+Precision));
end;

procedure TSingleHelper.setAsString( const value: string );
begin
  StringToSingle( value, Self );
end;

function TSingleHelper.getAsString: string;
begin
  Result := FloatToString( Self ).Trim(['0']);
end;


function TDoubleHelper.Approximates(const Value, Precision: double): boolean;
begin
  Result := (Self>(Value-Precision)) and (Self<(Value+Precision));
end;

procedure TDoubleHelper.setAsString( const value: string );
begin
  StringToDouble( value, Self );
end;

function TDoubleHelper.getAsString: string;
begin
  Result := FloatToString( Self ).Trim(['0']);
end;

function TExtendedHelper.Approximates(const Value, Precision: extended): boolean;
begin
  Result := (Self>(Value-Precision)) and (Self<(Value+Precision));
end;

procedure TExtendedHelper.setAsString( const value: string );
{$ifdef CPU64BITS}
var
  D: double;
{$endif}
begin
  {$ifdef CPU32BITS}
  StringToExtended( value, Self );
  {$else}
    {$ifdef CPU64BITS}
    StringToDouble( value, D );
    Self := D;
    {$else}
      .ErrorBuilding = How many bits is the target?
    {$endif}
  {$endif}
end;

function TExtendedHelper.getAsString: string;
begin
  Result := FloatToString( Self ).Trim(['0']).Trim(['.']);
end;

class function TDateTimeHelper.Now: TDateTime;
var
{$ifdef MSWINDOWS}
  SysTime: SystemTime;
{$else}
  SysTime: TimeSpec;
  pTime: ptm;
const
  cNanosecondsPerMillisecond = 1000000;
{$endif}
begin
  {$ifdef MSWINDOWS}
  GetSystemTime(SysTime);
  Result := EncodeDateTime( SysTime.wYear, SysTime.wMonth, SysTime.wDay, SysTime.wHour, SysTime.wMinute, SysTime.wSecond, SysTime.wMilliseconds );
  {$else}
  clock_gettime(CLOCK_REALTIME, SysTime);
  ptime := localtime(SysTime.tv_sec);
  Result := EncodeDateTime( ptime^.tm_year, ptime^.tm_mon, ptime^.tm_mday, ptime^.tm_hour, ptime^.tm_min, ptime^.tm_sec, SysTime.tv_nsec div cNanosecondsPerMillisecond );
  {$endif}
end;

function TDateTimeHelper.getDate: TDateTime;
var
  D: double;
begin
  D := Trunc(Double(Self));
  Result := TDateTime( D );
end;

procedure TDateTimeHelper.setDate( const Value: TDateTime );
var
  D: double;
begin
  D := Trunc(Double(Value)) + + Frac(Double(Self));
  Self := TDateTime( D );
end;

function TDateTimeHelper.getTime: TDateTime;
begin
  Result := Frac(Double(Self));
end;

procedure TDateTimeHelper.setTime( const Value: TDateTime );
var
  D: double;
begin
  D := Trunc(Double(Self)) + Frac(Double(Value));
  Self := TDateTime( D );
end;

class function TDateTimeHelper.Encode( const Year, Month, Day, Hour, Minute, Second, Millisecond: uint16 ): TDateTime;
begin
  Result := EncodeDateTime( Year, Month, Day, Hour, Minute, Second, Millisecond );
end;

class function TDateTimeHelper.Encode( const Year, Month, Day: uint16 ): TDateTime;
begin
  Result := EncodeDate( Year, Month, Day );
end;

class function TDateTimeHelper.Encode( const Hour, Minute, Second, Millisecond: uint16 ): TDateTime;
begin
  Result := EncodeTime( Hour, Minute, Second, Millisecond ).getTime;
end;

procedure TDateTimeHelper.Decode( out Year, Month, Day, Hour, Minute, Second, Millisecond: uint16 );
begin
  DecodeDateTime( Self, Year, Month, Day, Hour, Minute, Second, Millisecond );
end;

procedure TDateTimeHelper.Decode( out Year, Month, Day: uint16 );
begin
  DecodeDate( Self, Year, Month, Day );
end;

procedure TDateTimeHelper.Decode( out Hour, Minute, Second, Millisecond: uint16 );
begin
  DecodeTime( Self, Hour, Minute, Second, Millisecond );
end;

function TUInt8Helper.getAsHex: string;
begin
  Result := HexToString( Self, sizeof(uint8) );
end;

procedure TUInt8Helper.setAsHex(const value: string);
begin
  StringToHex( value, self, sizeof(uint8) );
end;

procedure TUInt8Helper.setAsString( const value: string );
begin
  StringToUnsignedInteger8( value, self );
end;

function TUInt8Helper.getAsString: string;
begin
  Result := UnsignedIntegerToString( Self );
end;

function TUInt16Helper.getAsHex: string;
begin
  Result := HexToString( Self, sizeof(uint16) );
end;

procedure TUInt16Helper.setAsHex( const Value: string );
begin
  StringToHex( value, self, sizeof(uint16) );
end;

procedure TUInt16Helper.setAsString( const value: string );
begin
  StringToUnsignedInteger16( value, Self );
end;

function TUInt16Helper.getAsString: string;
begin
  Result := UnsignedIntegerToString( Self );
end;

function TInt8Helper.getAsHex: string;
begin
  Result := HexToString( Self, sizeof(uint8) );
end;

procedure TInt8Helper.setAsHex(const value: string);
begin
  StringToHex( value, self, sizeof(uint8) );
end;

procedure TInt8Helper.setAsString( const value: string );
begin
  StringToInteger8( value, Self );
end;

function TInt8Helper.getAsString: string;
begin
  Result := IntegerToString( Self );
end;

function TInt16Helper.getAsHex: string;
begin
  Result := HexToString( Self, sizeof(uint16) );
end;

procedure TInt16Helper.setAsHex( const value: string );
begin
  StringToHex( value, self, sizeof(uint16) );
end;

procedure TInt16Helper.setAsString( const value: string );
begin
  StringToInteger16( value, Self );
end;

function TInt16Helper.getAsString: string;
begin
  Result := IntegerToString( Self );
end;

procedure TArrayOfStringHelper.TrimAll;
var
  idx: nativeuint;
begin
  if Length(Self)=0 then exit;
  for idx := 0 to pred(length(Self)) do begin
    Self[idx] := Self[idx].Trim;
  end;
end;

procedure TArrayOfStringHelper.UppercaseAll;
var
  idx: nativeuint;
begin
  if Length(Self)=0 then exit;
  for idx := 0 to pred(length(Self)) do begin
    Self[idx] := Self[idx].Uppercase;
  end;
end;

procedure TArrayOfStringHelper.LowercaseAll;
var
  idx: nativeuint;
begin
  if Length(Self)=0 then exit;
  for idx := 0 to pred(length(Self)) do begin
    Self[idx] := Self[idx].Lowercase;
  end;
end;

procedure TArrayOfStringHelper.AssignArray(const value: array of string );
var
  idx: nativeuint;
begin
  SetLength(Self,Length(Value));
  if Length(Self)=0 then begin
    exit;
  end;
  for idx := 0 to pred(Length(Self)) do begin
    Self[idx] := Value[idx];
  end;
end;

function TArrayOfStringHelper.Find(const value: string; out FoundIdx: nativeuint; const CaseInsensitive: boolean): boolean;
var
  idx: nativeuint;
  SearchStr: string;
begin
  Result := False;
  if Length(Self)=0 then begin
    exit;
  end;
  SearchStr := value;
  if CaseInsensitive then begin
    SearchStr := SearchStr.Uppercase;
  end;
  for idx := 0 to pred(Length(self)) do begin
    if CaseInsensitive then begin
      if Self[idx].Uppercase=SearchStr then begin
        Result := True;
        FoundIdx := Idx;
        exit;
      end;
    end else begin
      if Self[idx]=SearchStr then begin
        Result := True;
        FoundIdx := Idx;
        exit;
      end;
    end;
  end;
end;

function TArrayOfStringHelper.RemoveDuplicates( const CaseInsensitive: boolean = false ): TArrayOfString;
var
  idx: nativeuint;
  Counter: nativeuint;
  FoundIdx: nativeuint;
  InterMediate: TArrayOfString;
begin
  if Length(Self)=0 then begin
    {$warnings off}
    SetLength(Result,Length(Self));
    {$warnings on}
    exit;
  end;
  {$hints off}
  SetLength(InterMediate,Length(Self));
  {$hints on}
  try
    Counter := 0;
    for idx := 0 to pred(Length(Self)) do begin
      if not Intermediate.Find(Self[idx],FoundIdx,CaseInsensitive) then begin
        Intermediate[counter] := Self[idx];
        inc(Counter);
      end;
    end;
    SetLength(Result,Counter);
    if Counter>0 then begin
      for idx := 0 to pred(Counter) do begin
        Result[idx] := Intermediate[idx];
      end;
    end;
  finally
    SetLength(Intermediate,0);
  end;
end;

procedure TArrayOfStringHelper.Remove(const item: string; const CaseInsensitive: boolean);
var
  idx: nativeuint;
  Counter: nativeuint;
  InterMediate: TArrayOfString;
  utItem: string;
begin
  if Length(Self)=0 then begin
    exit;
  end;
  utItem := item.Uppercase;
  {$hints off}
  SetLength(InterMediate,Length(Self));
  {$hints on}
  try
    Counter := 0;
    for idx := 0 to pred(Length(Self)) do begin
      if not ((Self[idx]=item) or ((CaseInsensitive) and (self[idx].Uppercase=utItem))) then begin
        Intermediate[counter] := Self[idx];
        inc(Counter);
      end;
    end;
    SetLength(Self,Counter);
    if Counter>0 then begin
      for idx := 0 to pred(Counter) do begin
        Self[idx] := Intermediate[idx];
      end;
    end;
  finally
    SetLength(Intermediate,0);
  end;
end;

function FloatToHalf(Float: Single): word;
var
  Src: LongWord;
  Sign, Exp, Mantissa: LongInt;
begin
  Src := PLongWord(@Float)^;
  // Extract sign, exponent, and mantissa from Single number
  Sign := Src shr 31;
  Exp := LongInt((Src and $7F800000) shr 23) - 127 + 15;
  Mantissa := Src and $007FFFFF;

  if (Exp > 0) and (Exp < 30) then
  begin
    // Simple case - round the significand and combine it with the sign and exponent
    Result := (Sign shl 15) or (Exp shl 10) or ((Mantissa + $00001000) shr 13);
  end
  else if Src = 0 then
  begin
    // Input float is zero - return zero
    Result := 0;
  end
  else
  begin
    // Difficult case - lengthy conversion
    if Exp <= 0 then
    begin
      if Exp < -10 then
      begin
        // Input float's value is less than HalfMin, return zero
         Result := 0;
      end
      else
      begin
        // Float is a normalized Single whose magnitude is less than HalfNormMin.
        // We convert it to denormalized half.
        Mantissa := (Mantissa or $00800000) shr (1 - Exp);
        // Round to nearest
        if (Mantissa and $00001000) > 0 then
          Mantissa := Mantissa + $00002000;
        // Assemble Sign and Mantissa (Exp is zero to get denormalized number)
        Result := (Sign shl 15) or (Mantissa shr 13);
      end;
    end
    else if Exp = 255 - 127 + 15 then
    begin
      if Mantissa = 0 then
      begin
        // Input float is infinity, create infinity half with original sign
        Result := (Sign shl 15) or $7C00;
      end
      else
      begin
        // Input float is NaN, create half NaN with original sign and mantissa
        Result := (Sign shl 15) or $7C00 or (Mantissa shr 13);
      end;
    end
    else
    begin
      // Exp is > 0 so input float is normalized Single

      // Round to nearest
      if (Mantissa and $00001000) > 0 then
      begin
        Mantissa := Mantissa + $00002000;
        if (Mantissa and $00800000) > 0 then
        begin
          Mantissa := 0;
          Exp := Exp + 1;
        end;
      end;

      if Exp > 30 then
      begin
        // Exponent overflow - return infinity half
        Result := (Sign shl 15) or $7C00;
      end
      else
        // Assemble normalized half
        Result := (Sign shl 15) or (Exp shl 10) or (Mantissa shr 13);
    end;
  end;
end;

function HalfToFloat(Half: word): Single;
var
  Dst, Sign, Mantissa: LongWord;
  Exp: LongInt;
begin
  // Extract sign, exponent, and mantissa from half number
  Sign := Half shr 15;
  Exp := (Half and $7C00) shr 10;
  Mantissa := Half and 1023;

  if (Exp > 0) and (Exp < 31) then
  begin
    // Common normalized number
    Exp := Exp + (127 - 15);
    Mantissa := Mantissa shl 13;
    Dst := (Sign shl 31) or (LongWord(Exp) shl 23) or Mantissa;
    // Result := Power(-1, Sign) * Power(2, Exp - 15) * (1 + Mantissa / 1024);
  end
  else if (Exp = 0) and (Mantissa = 0) then
  begin
    // Zero - preserve sign
    Dst := Sign shl 31;
  end
  else if (Exp = 0) and (Mantissa <> 0) then
  begin
    // Denormalized number - renormalize it
    while (Mantissa and $00000400) = 0 do
    begin
      Mantissa := Mantissa shl 1;
      Dec(Exp);
    end;
    Inc(Exp);
    Mantissa := Mantissa and not $00000400;
    // Now assemble normalized number
    Exp := Exp + (127 - 15);
    Mantissa := Mantissa shl 13;
    Dst := (Sign shl 31) or (LongWord(Exp) shl 23) or Mantissa;
    // Result := Power(-1, Sign) * Power(2, -14) * (Mantissa / 1024);
  end
  else if (Exp = 31) and (Mantissa = 0) then
  begin
    // +/- infinity
    Dst := (Sign shl 31) or $7F800000;
  end
  else //if (Exp = 31) and (Mantisa <> 0) then
  begin
    // Not a number - preserve sign and mantissa
    Dst := (Sign shl 31) or $7F800000 or (Mantissa shl 13);
  end;

  // Reinterpret LongWord as Single
  Result := PSingle(@Dst)^;
end;

{$ifndef CPU64BITS}
class operator half.Explicit(const a: single): half;
begin
  Result.Value := FloatToHalf(a);
end;
{$endif}

class operator half.GreaterThan(const a: double; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a>sB;
end;

class operator half.GreaterThanOrEqual(const a: double; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a>=sB;
end;


{$ifndef CPU64BITS}
{$if (not defined(CPUAARCH64)) and (not defined(CPUARM))}
class operator half.GreaterThan(const a: extended; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a>sB;
end;

class operator half.GreaterThanOrEqual(const a: extended; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a>=sB;
end;

{$endif}
{$endif}

class operator half.GreaterThan(const a: half; const b: double): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA>b;
end;

class operator half.GreaterThanOrEqual(const a: half; const b: double): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA>=b;
end;


{$ifndef CPU64BITS}
{$if (not defined(CPUAARCH64)) and (not defined(CPUARM))}
class operator half.GreaterThan(const a: half; const b: extended): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA>b;
end;

class operator half.GreaterThanOrEqual(const a: half; const b: extended): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA>=b;
end;
{$endif}
{$endif}

class operator half.GreaterThan(const a: half; const b: half): boolean;
var
  sA: single;
  sB: single;
begin
  sA := a;
  sB := b;
  Result := sA>sB;
end;

class operator half.GreaterThanOrEqual(const a: half; const b: half): boolean;
var
  sA: single;
  sB: single;
begin
  sA := a;
  sB := b;
  Result := sA>=sB;
end;

class operator half.GreaterThan(const a: half; const b: single): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA>b;
end;

class operator half.GreaterThanOrEqual(const a: half; const b: single): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA>=b;
end;

class operator half.GreaterThan(const a: single; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a>sB;
end;

class operator half.GreaterThanOrEqual(const a: single; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a>=sB;
end;

class operator half.LessThan(const a: double; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a<sB;
end;

class operator half.LessThanOrEqual(const a: double; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a<=sB;
end;

{$ifndef CPU64BITS}
{$if (not defined(CPUAARCH64)) and (not defined(CPUARM))}
class operator half.LessThan(const a: extended; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a<sB;
end;

class operator half.LessThanOrEqual(const a: extended; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a<=sB;
end;
{$endif}
{$endif}

class operator half.LessThan(const a: half; const b: double): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA<b;
end;

class operator half.LessThanOrEqual(const a: half; const b: double): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA<=b;
end;


{$ifndef CPU64BITS}
{$if (not defined(CPUAARCH64)) and (not defined(CPUARM))}
class operator half.LessThan(const a: half; const b: extended): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA<b;
end;

class operator half.LessThanOrEqual(const a: half; const b: extended): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA<=b;
end;

{$endif}
{$endif}

class operator half.LessThan(const a: half; const b: half): boolean;
var
  sA: single;
  sB: single;
begin
  sA := a;
  sB := b;
  Result := sA<sB;
end;

class operator half.LessThanOrEqual(const a: half; const b: half): boolean;
var
  sA: single;
  sB: single;
begin
  sA := a;
  sB := b;
  Result := sA<=sB;
end;

class operator half.LessThan(const a: half; const b: single): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA<b;
end;

class operator half.LessThanOrEqual(const a: half; const b: single): boolean;
var
  sA: single;
begin
  sA := a;
  Result := sA<=b;
end;

class operator half.LessThan(const a: single; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a<sB;
end;

class operator half.LessThanOrEqual(const a: single; const b: half): boolean;
var
  sB: single;
begin
  sB := b;
  Result := a<=sB;
end;


{$ifndef CPU64BITS}
class operator half.Explicit(const a: half): single;
begin
  Result := HalfToFloat(a.Value);
end;
{$endif}

class operator half.Add(const a, b: half): half;
begin
  Result.value := FloatToHalf( HalfToFloat(a.value) + HalfToFloat(b.value) );
end;

class operator half.Divide(const a, b: half): half;
begin
  Result.value := FloatToHalf( HalfToFloat(a.value) / HalfToFloat(b.value) );
end;

class operator half.Equal(const a, b: half): boolean;
begin
  Result := single(a)=single(b);
end;

{$ifndef CPU64BITS}
class operator half.Explicit(const a: half): double;
var
  s: single;
begin
  s := HalfToFloat(a.value);
  Result := S;
end;
{$endif}

{$ifndef CPU64BITS}
class operator half.Explicit(const a: double): half;
var
  s: single;
begin
  s := a;
  Result.value := FloatToHalf(s);
end;
{$endif}

class operator half.Implicit(const a: half): single;
begin
  Result := HalfToFloat(a.value);
end;

class operator half.Implicit(const a: single): half;
begin
  Result.value := FloatToHalf(a);
end;

class operator half.Implicit(const a: double): half;
var
  s: single;
begin
  s := a;
  Result.value := FloatToHalf(s);
end;

class operator half.Implicit(const a: half): double;
var
  s: single;
begin
  s := HalfToFloat(a.value);
  Result := s;
end;

{$if defined(LINUX) and defined(FPC)}
{$if (not defined(CPUAARCH64)) and (not defined(CPUARM))}
class operator half.Implicit(const a: extended): half;
var
  d: double;
begin
  d := a;
  Result.value := FloatToHalf(d);
end;
{$endif}
{$endif}

{$if defined(LINUX) and defined(FPC)}
{$if (not defined(CPUAARCH64)) and (not defined(CPUARM))}
class operator half.Implicit(const a: half): extended;
var
  d: double;
begin
  d := HalfToFloat(a.value);
  Result := d;
end;
{$endif}
{$endif}

class operator half.Multiply(const a, b: half): half;
begin
  Result.value := FloatToHalf( HalfToFloat(a.value) * HalfToFloat(b.value) );
end;

class operator half.Subtract(const a, b: half): half;
begin
  Result.value := FloatToHalf( HalfToFloat(a.value) - HalfToFloat(b.value) );
end;

type
  TConvertGUID = record
    P1: uint32;
    P2: uint16;
    P3: uint16;
    P4a: uint8;
    P4b: uint8;
    P5a: uint8;
    P5b: uint8;
    P5c: uint8;
    P5d: uint8;
    P5e: uint8;
    P5f: uint8;
  end;
  pConvertGUID = ^TConvertGUID;

function TGUIDHelper.getAsString: string;
begin
  Result :=
    '{'+
    pConvertGUID(@Self)^.P1.AsHex+'-'+
    pConvertGUID(@Self)^.P2.AsHex+'-'+
    pConvertGUID(@Self)^.P3.AsHex+'-'+
    pConvertGUID(@Self)^.P4a.AsHex+
    pConvertGUID(@Self)^.P4b.AsHex+'-'+
    pConvertGUID(@Self)^.P5a.AsHex+
    pConvertGUID(@Self)^.P5b.AsHex+
    pConvertGUID(@Self)^.P5c.AsHex+
    pConvertGUID(@Self)^.P5d.AsHex+
    pConvertGUID(@Self)^.P5e.AsHex+
    pConvertGUID(@Self)^.P5f.AsHex+
    '}';
end;

procedure TGUIDHelper.setAsString(const value: string);

  procedure ConvertPart( const src: string; var idx: nativeuint; const chars: nativeuint; var R );
  var
    idy: nativeuint;
    P: string;
  begin
    P := '';
    for idy := 0 to pred(chars) do begin
      if not src[idx].CharInArray(cHexChars) then TStatus( stGUIDConversionError ).Raize;
      P := P + src[idx];
      inc(idx);
    end;
    StringToHex( P, R, (chars div 2) );
  end;

  procedure Expect( const src: string; var idx: nativeuint; const c: char ); inline;
  begin
    if src[idx]<>c then TStatus( stGUIDConversionError ).Raize;
    inc(idx);
  end;

var
  idx: nativeuint;
  S: string;
begin
  S := Value.UppercaseTrim;
  if Length(S)<>38 then TStatus( stGUIDConversionError ).Raize;
  {$ifdef NEXTGEN}
  idx := 0;
  {$else}
  idx := 1;
  {$endif}
  Expect( S, idx, '{' );

  ConvertPart( S, idx, 8, pConvertGUID(@Self)^.P1 );
  Expect(S,idx,'-');
  ConvertPart( S, idx, 4, pConvertGUID(@Self)^.P2 );
  Expect(S,idx,'-');
  ConvertPart( S, idx, 4, pConvertGUID(@Self)^.P3 );
  Expect(S,idx,'-');
  ConvertPart( S, idx, 2, pConvertGUID(@Self)^.P4a );
  ConvertPart( S, idx, 2, pConvertGUID(@Self)^.P4b );
  Expect(S,idx,'-');
  ConvertPart( S, idx, 2, pConvertGUID(@Self)^.P5a );
  ConvertPart( S, idx, 2, pConvertGUID(@Self)^.P5b );
  ConvertPart( S, idx, 2, pConvertGUID(@Self)^.P5c );
  ConvertPart( S, idx, 2, pConvertGUID(@Self)^.P5d );
  ConvertPart( S, idx, 2, pConvertGUID(@Self)^.P5e );
  ConvertPart( S, idx, 2, pConvertGUID(@Self)^.P5f );
end;

function TGUIDHelper.EqualTo(const B: TGUID): boolean;
begin
  Result := IsEqualGUID( B, Self );
end;

class function TGUIDHelper.New: TGUID;
begin
  {$ifdef MSWINDOWS}
  CoCreateGuid(@Result);
  {$else}
  uuid_generate_time_safe(Result);
  {$endif}
end;


constructor TGenericValue<T>.Create(const Value: T);
begin
  inherited Create;
  fValue := Value;
end;

destructor TGenericValue<T>.Destroy;
begin
  fValue := Default(T);
  inherited;
end;

function TGenericValue<T>.getValue: T;
begin
  Result := fValue;
end;

procedure TGenericValue<T>.setValue(const Value: T);
begin
  fValue := Value;
end;


procedure TVariant.Clear;
begin
  fStorage := nil;
  fDataType := vt_null;
end;

function TVariant.DataType: TVariantType;
begin
  Result := vt_null;
  if not assigned(fStorage) then exit;
  Result := fDataType;
end;


class operator TVariant.Explicit(const a: int8): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: int16): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: int32): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: int64): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: uint8): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: uint16): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: uint32): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: uint64): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: half): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: string): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: TObject): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: IInterface): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: pointer): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: single): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Explicit(const a: double): TVariant;
begin
  Result := TVariant.setter(a);
end;

{$ifdef CPU32}
class operator TVariant.Explicit(const a: extended): TVariant;
begin
  Result := TVariant.setter(a);
end;
{$endif}

class operator TVariant.Explicit(const a: char): TVariant;
begin
  Result := TVariant.setter(a);
end;

{$ifdef CPU32}
procedure TVariant.getter(out a: extended);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','extended'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','extended'] );
    vt_half      : a := ( fStorage as IGenericValue<half>       ).Value;
    vt_single    : a := ( fStorage as IGenericValue<single>     ).Value;
    vt_double    : a := ( fStorage as IGenericValue<double>     ).Value;
    vt_extended  : a := ( fStorage as IGenericValue<extended>   ).Value;
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','extended'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','extended'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','extended'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','extended'] );
  end;
end;
{$endif}

procedure TVariant.getter(out a: char);
begin
  case DataType of
    vt_uint8     : TStatus( stVariantTypeCoersion ).Raize( ['uint8','char'] );
    vt_uint16    : TStatus( stVariantTypeCoersion ).Raize( ['uint16','char'] );
    vt_uint32    : TStatus( stVariantTypeCoersion ).Raize( ['uint32','char'] );
    vt_uint64    : TStatus( stVariantTypeCoersion ).Raize( ['uint64','char'] );
    vt_int8      : TStatus( stVariantTypeCoersion ).Raize( ['int8','char'] );
    vt_int16     : TStatus( stVariantTypeCoersion ).Raize( ['int16','char'] );
    vt_int32     : TStatus( stVariantTypeCoersion ).Raize( ['uint32','char'] );
    vt_int64     : TStatus( stVariantTypeCoersion ).Raize( ['uint64','char'] );
    vt_char      : a := ( fStorage as IGenericValue<char>       ).Value;
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','char'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','char'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','char'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','char'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','char'] );
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','char'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','char'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','char'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','char'] );
  end;
end;

procedure TVariant.getter(out a: double);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','double'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','double'] );
    vt_half      : a := ( fStorage as IGenericValue<half>       ).Value;
    vt_single    : a := ( fStorage as IGenericValue<single>     ).Value;
    vt_double    : a := ( fStorage as IGenericValue<double>     ).Value;
    vt_extended  : a := ( fStorage as IGenericValue<extended>   ).Value;
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','double'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','double'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','double'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','dobule'] );
  end;
end;

procedure TVariant.getter(out a: single);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','single'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','single'] );
    vt_half      : a := ( fStorage as IGenericValue<half>       ).Value;
    vt_single    : a := ( fStorage as IGenericValue<single>     ).Value;
    vt_double    : a := ( fStorage as IGenericValue<double>     ).Value;
    vt_extended  : a := ( fStorage as IGenericValue<extended>   ).Value;
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','single'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','single'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','single'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','single'] );
  end;
end;

procedure TVariant.getter(out a: pointer);
begin
  case DataType of
    vt_uint8     : TStatus( stVariantTypeCoersion ).Raize( ['uint8','pointer'] );
    vt_uint16    : TStatus( stVariantTypeCoersion ).Raize( ['uint16','pointer'] );
    vt_uint32    : TStatus( stVariantTypeCoersion ).Raize( ['uint32','pointer'] );
    vt_uint64    : TStatus( stVariantTypeCoersion ).Raize( ['uint64','pointer'] );
    vt_int8      : TStatus( stVariantTypeCoersion ).Raize( ['int8','pointer'] );
    vt_int16     : TStatus( stVariantTypeCoersion ).Raize( ['int16','pointer'] );
    vt_int32     : TStatus( stVariantTypeCoersion ).Raize( ['int32','pointer'] );
    vt_int64     : TStatus( stVariantTypeCoersion ).Raize( ['int64','pointer'] );
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','pointer'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','pointer'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','pointer'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','pointer'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','pointer'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','pointer'] );
    vt_object    : a := pointer(( fStorage as IGenericValue<TObject>    ).Value);
    vt_interface : a := pointer( ( fStorage as IGenericValue<IInterface> ).Value );
    vt_pointer   : a := ( fStorage as IGenericValue<pointer>    ).Value;
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','pointer'] );
  end;
end;

procedure TVariant.getter(out a: IInterface);
begin
  case DataType of
    vt_uint8     : TStatus( stVariantTypeCoersion ).Raize( ['uint8','interface'] );
    vt_uint16    : TStatus( stVariantTypeCoersion ).Raize( ['uint16','interface'] );
    vt_uint32    : TStatus( stVariantTypeCoersion ).Raize( ['uint32','interface'] );
    vt_uint64    : TStatus( stVariantTypeCoersion ).Raize( ['uint64','interface'] );
    vt_int8      : TStatus( stVariantTypeCoersion ).Raize( ['int8','interface'] );
    vt_int16     : TStatus( stVariantTypeCoersion ).Raize( ['int16','interface'] );
    vt_int32     : TStatus( stVariantTypeCoersion ).Raize( ['int32','interface'] );
    vt_int64     : TStatus( stVariantTypeCoersion ).Raize( ['int64','interface'] );
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','interface'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','interface'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','interface'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','interface'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','interface'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','interface'] );
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','interface'] );
    vt_interface : a := ( fStorage as IGenericValue<IInterface> ).Value;
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','interface'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','IInterface'] );
  end;
end;

procedure TVariant.getter(out a: TObject);
begin
  case DataType of
    vt_uint8     : TStatus( stVariantTypeCoersion ).Raize( ['uint8','TObject'] );
    vt_uint16    : TStatus( stVariantTypeCoersion ).Raize( ['uint16','TObject'] );
    vt_uint32    : TStatus( stVariantTypeCoersion ).Raize( ['uint32','TObject'] );
    vt_uint64    : TStatus( stVariantTypeCoersion ).Raize( ['uint64','TObject'] );
    vt_int8      : TStatus( stVariantTypeCoersion ).Raize( ['int8','TObject'] );
    vt_int16     : TStatus( stVariantTypeCoersion ).Raize( ['int16','TObject'] );
    vt_int32     : TStatus( stVariantTypeCoersion ).Raize( ['int32','TObject'] );
    vt_int64     : TStatus( stVariantTypeCoersion ).Raize( ['int64','TObject'] );
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','TObject'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','TObject'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','TObject'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','TObject'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','TObject'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','TObject'] );
    vt_object    : a := ( fStorage as IGenericValue<TObject>    ).Value;
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','TObject'] );
    vt_pointer   : a := ( fStorage as IGenericValue<pointer>    ).Value;
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','TObject'] );
  end;
end;

procedure TVariant.getter(out a: string);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value.AsString;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value.AsString;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value.AsString;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value.AsString;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value.AsString;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value.AsString;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value.AsString;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value.AsString;
    vt_char      : a := ( fStorage as IGenericValue<char>       ).Value;
    vt_string    : a := ( fStorage as IGenericValue<string>     ).Value;
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','string'] );
    vt_single    : a := ( fStorage as IGenericValue<single>     ).Value.AsString;
    vt_double    : a := ( fStorage as IGenericValue<double>     ).Value.AsString;
    vt_extended  : a := ( fStorage as IGenericValue<extended>   ).Value.AsString;
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','string'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','string'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','string'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','string'] );
  end;
end;

procedure TVariant.getter(out a: half);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','half'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','half'] );
    vt_half      : a := ( fStorage as IGenericValue<half>       ).Value;
    vt_single    : a := ( fStorage as IGenericValue<single>     ).Value;
    vt_double    : a := ( fStorage as IGenericValue<double>     ).Value;
    vt_extended  : a := ( fStorage as IGenericValue<extended>   ).Value;
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','half'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','half'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','half'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','half'] );
  end;
end;

procedure TVariant.getter(out a: uint64);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','uint64'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','uint64'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','uint64'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','uint64'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','uint64'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','uint64'] );
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','uint64'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','uint64'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','uint64'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','uint64'] );
  end;
end;

procedure TVariant.getter(out a: uint32);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','uint32'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','uint32'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','uint32'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','uint32'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','uint32'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','uint32'] );
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['object','uint32'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','uint32'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','uint32'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','uint32'] );
  end;
end;

procedure TVariant.getter(out a: uint16);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','uint16'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['single','uint16'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','uint16'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','uint16'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','uint16'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','uint16'] );
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','uint16'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','uint16'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','uint16'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','uint16'] );
  end;
end;

procedure TVariant.getter(out a: uint8);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','uint8'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','uint8'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','uint8'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','uint8'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','uint8'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','uint8'] );
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','uint8'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','uint8'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','uint8'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','uint8'] );
  end;
end;

procedure TVariant.getter(out a: int64);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','int64'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','int64'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','int64'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','int64'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','int64'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','int64'] );
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','int64'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','int64'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','int64'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','int64'] );
  end;
end;

procedure TVariant.getter(out a: int32);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','int32'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','int32'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','int32'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','int32'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','int32'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','int32'] );
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','int32'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','int32'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','int32'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','int32'] );
  end;
end;

procedure TVariant.getter(out a: int8);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','int8'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','int8'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','int8'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','int8'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','int8'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','int8'] );
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','int8'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','int8'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','int8'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','int8'] );
  end;
end;

procedure TVariant.getter(out a: int16);
begin
  case DataType of
    vt_uint8     : a := ( fStorage as IGenericValue<uint8>      ).Value;
    vt_uint16    : a := ( fStorage as IGenericValue<uint16>     ).Value;
    vt_uint32    : a := ( fStorage as IGenericValue<uint32>     ).Value;
    vt_uint64    : a := ( fStorage as IGenericValue<uint64>     ).Value;
    vt_int8      : a := ( fStorage as IGenericValue<int8>       ).Value;
    vt_int16     : a := ( fStorage as IGenericValue<int16>      ).Value;
    vt_int32     : a := ( fStorage as IGenericValue<int32>      ).Value;
    vt_int64     : a := ( fStorage as IGenericValue<int64>      ).Value;
    vt_char      : TStatus( stVariantTypeCoersion ).Raize( ['char','int16'] );
    vt_string    : TStatus( stVariantTypeCoersion ).Raize( ['string','int16'] );
    vt_half      : TStatus( stVariantTypeCoersion ).Raize( ['half','int16'] );
    vt_single    : TStatus( stVariantTypeCoersion ).Raize( ['single','int16'] );
    vt_double    : TStatus( stVariantTypeCoersion ).Raize( ['double','int16'] );
    vt_extended  : TStatus( stVariantTypeCoersion ).Raize( ['extended','int16'] );
    vt_object    : TStatus( stVariantTypeCoersion ).Raize( ['TObject','int16'] );
    vt_interface : TStatus( stVariantTypeCoersion ).Raize( ['IInterface','int16'] );
    vt_pointer   : TStatus( stVariantTypeCoersion ).Raize( ['pointer','int16'] );
    vt_null      : TStatus( stVariantTypeCoersion ).Raize( ['null','int16'] );
  end;
end;

class operator TVariant.Implicit(const a: int16): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: int8): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: int32): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: int64): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: uint8): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: uint16): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: uint32): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: uint64): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: string): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: TObject): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: IInterface): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: pointer): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: char): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: half): TVariant;
begin
  Result := TVariant.setter(a);
end;

class operator TVariant.Implicit(const a: single): TVariant;
begin
  Result := TVariant.setter(a);
end;

{$ifdef CPU32}
class operator TVariant.Implicit(const a: extended): TVariant;
begin
  Result := TVariant.setter(a);
end;
{$endif}

class operator TVariant.Implicit(const a: double): TVariant;
begin
  Result := TVariant.setter(a);
end;

{$ifdef CPU32}
class function TVariant.setter(const a: extended): TVariant;
begin
  Result.fStorage := TGenericValue<extended>.Create( a );
  Result.fDataType := vt_extended;
end;
{$endif}

class function TVariant.setter(const a: char): TVariant;
begin
  Result.fStorage := TGenericValue<char>.Create( a );
  Result.fDataType := vt_char;
end;

class function TVariant.setter(const a: double): TVariant;
begin
  Result.fStorage := TGenericValue<double>.Create( a );
  Result.fDataType := vt_double;
end;

class function TVariant.setter(const a: single): TVariant;
begin
  Result.fStorage := TGenericValue<single>.Create( a );
  Result.fDataType := vt_single;
end;

class function TVariant.setter(const a: pointer): TVariant;
begin
  Result.fStorage := TGenericValue<pointer>.Create( a );
  Result.fDataType := vt_pointer;
end;

class function TVariant.setter(const a: IInterface): TVariant;
begin
  Result.fStorage := TGenericValue<IInterface>.Create( a );
  Result.fDataType := vt_interface;
end;

class function TVariant.setter(const a: TObject): TVariant;
begin
  Result.fStorage := TGenericValue<TObject>.Create( a );
  Result.fDataType := vt_object;
end;

class function TVariant.setter(const a: string): TVariant;
begin
  Result.fStorage := TGenericValue<string>.Create( a );
  Result.fDataType := vt_string;
end;

class function TVariant.setter(const a: half): TVariant;
begin
  Result.fStorage := TGenericValue<half>.Create( a );
  Result.fDataType := vt_half;
end;

class function TVariant.setter(const a: uint64): TVariant;
begin
  Result.fStorage := TGenericValue<uint64>.Create( a );
  Result.fDataType := vt_uint64;
end;

class function TVariant.setter(const a: uint32): TVariant;
begin
  Result.fStorage := TGenericValue<uint32>.Create( a );
  Result.fDataType := vt_uint32;
end;

class function TVariant.setter(const a: uint16): TVariant;
begin
  Result.fStorage := TGenericValue<uint16>.Create( a );
  Result.fDataType := vt_uint16;
end;

class function TVariant.setter(const a: uint8): TVariant;
begin
  Result.fStorage := TGenericValue<uint8>.Create( a );
  Result.fDataType := vt_uint8;
end;

class function TVariant.setter(const a: int64): TVariant;
begin
  Result.fStorage := TGenericValue<int64>.Create( a );
  Result.fDataType := vt_int64;
end;

class function TVariant.setter(const a: int32): TVariant;
begin
  Result.fStorage := TGenericValue<int32>.Create( a );
  Result.fDataType := vt_int32;
end;

class function TVariant.setter(const a: int16): TVariant;
begin
  Result.fStorage := TGenericValue<int16>.Create( a );
  Result.fDataType := vt_int16;
end;

class function TVariant.setter(const a: int8): TVariant;
begin
  Result.fStorage := TGenericValue<int8>.Create( a );
  Result.fDataType := vt_int8;
end;

class operator TVariant.Explicit(const a: TVariant): int8;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): int16;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): int32;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): int64;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): uint8;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): uint16;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): uint32;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): uint64;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): half;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): string;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): TObject;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): IInterface;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): pointer;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): single;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): double;
begin
  a.getter( Result );
end;

class operator TVariant.Explicit(const a: TVariant): char;
begin
  a.getter( Result );
end;

{$ifdef CPU32}
class operator TVariant.Explicit(const a: TVariant): extended;
begin
  a.getter( Result );
end;
{$endif}

class operator TVariant.Implicit(const a: TVariant): int16;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): int8;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): int32;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): int64;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): uint8;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): uint16;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): uint32;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): uint64;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): half;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): string;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): TObject;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): IInterface;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): pointer;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): single;
begin
  a.getter( Result );
end;

class operator TVariant.Implicit(const a: TVariant): double;
begin
  a.getter( Result );
end;

{$ifdef CPU32}
class operator TVariant.Implicit(const a: TVariant): extended;
begin
  a.getter( Result );
end;
{$endif}

class operator TVariant.Implicit(const a: TVariant): char;
begin
  a.getter( Result );
end;

procedure TNotifyDelegate.Execute;
begin
  if assigned(fNotify) then begin
    fNotify();
    exit;
  end;
  {$ifdef fpc}
  if assigned(fNotifyObject) then begin
    fNotifyObject();
    exit;
  end;
  if assigned(fNotifyNested) then begin
    fNotifyNested();
    exit;
  end;
  {$endif}
end;

function TNotifyDelegate.IsAssigned: boolean;
begin
  {$ifdef fpc}
    Result := assigned( fNotifyObject ) or assigned( fNotifyNested ) or assigned( fNotify );
  {$else}
    Result := assigned( fNotify );
  {$endif}
end;

class operator TNotifyDelegate.Implicit( const Value: TNotifyCallback ): TNotifyDelegate;
begin
  Result := Default(  TNotifyDelegate );
  Result.fNotify := Value;
end;

class function TNotifyDelegate.null: TNotifyDelegate;
begin
  Result := Default(  TNotifyDelegate );
end;

class operator TNotifyDelegate.Explicit( const Value: TNotifyCallback ): TNotifyDelegate;
begin
  Result := Default(  TNotifyDelegate );
  Result.fNotify := Value;
end;

{$ifdef fpc}
class operator TNotifyDelegate.Implicit( const Value: TNotifyObjectCallback ): TNotifyDelegate;
begin
  Result := Default(  TNotifyDelegate );
  Result.fNotifyObject := Value;
end;

class operator TNotifyDelegate.Explicit( const Value: TNotifyObjectCallback ): TNotifyDelegate;
begin
  Result := Default(  TNotifyDelegate );
  Result.fNotifyObject := Value;
end;

class operator TNotifyDelegate.Implicit( const Value: TNotifyNestedCallback ): TNotifyDelegate;
begin
  Result := Default(  TNotifyDelegate );
  Result.fNotifyNested := Value;
end;

class operator TNotifyDelegate.Explicit( const Value: TNotifyNestedCallback ): TNotifyDelegate;
begin
  Result := Default(  TNotifyDelegate );
  Result.fNotifyNested := Value;
end;

{$endif}

procedure TNotifyWithSenderDelegate<T>.Execute( const Sender: T );
begin
  if assigned(fNotify) then begin
    fNotify( Sender );
    exit;
  end;
  {$ifdef fpc}
  if assigned(fNotifyObject) then begin
    fNotifyObject( Sender );
    exit;
  end;
  if assigned(fNotifyNested) then begin
    fNotifyNested( Sender );
    exit;
  end;
  {$endif}
end;

class function TNotifyWithSenderDelegate<T>.null(): TNotifyWithSenderDelegate<T>;
begin
  Result := Default(  TNotifyWithSenderDelegate<T> );
end;

class operator TNotifyWithSenderDelegate<T>.Implicit( const Value: TNotifyWithSenderCallback<T> ): TNotifyWithSenderDelegate<T>;
begin
  Result := Default(  TNotifyWithSenderDelegate<T> );
  Result.fNotify := Value;
end;

class operator TNotifyWithSenderDelegate<T>.Explicit( const Value: TNotifyWithSenderCallback<T> ): TNotifyWithSenderDelegate<T>;
begin
  Result := Default(  TNotifyWithSenderDelegate<T> );
  Result.fNotify := Value;
end;

{$ifdef fpc}
class operator TNotifyWithSenderDelegate<T>.Implicit( const Value: TNotifyWithSenderObjectCallback<T> ): TNotifyWithSenderDelegate<T>;
begin
  Result := Default(  TNotifyWithSenderDelegate<T> );
  Result.fNotifyObject := Value;
end;

class operator TNotifyWithSenderDelegate<T>.Explicit( const Value: TNotifyWithSenderObjectCallback<T> ): TNotifyWithSenderDelegate<T>;
begin
  Result := Default(  TNotifyWithSenderDelegate<T> );
  Result.fNotifyObject := Value;
end;

class operator TNotifyWithSenderDelegate<T>.Implicit( const Value: TNotifyWithSenderNestedCallback<T> ): TNotifyWithSenderDelegate<T>;
begin
  Result := Default(  TNotifyWithSenderDelegate<T> );
  Result.fNotifyNested := Value;
end;

class operator TNotifyWithSenderDelegate<T>.Explicit( const Value: TNotifyWithSenderNestedCallback<T> ): TNotifyWithSenderDelegate<T>;
begin
  Result := Default(  TNotifyWithSenderDelegate<T> );
  Result.fNotifyNested := Value;
end;

{$endif}

procedure TLoopDelegate.Execute(const start, stop: nativeuint; const Parameters: array of TVariant);
var
  idx: nativeuint;
begin
  if assigned( fIndexedLoopCallback ) then begin
    for idx := start to stop do fIndexedLoopCallback( idx, Parameters );
    exit;
  end;
  if assigned( fIndexedLoopNoParametersCallback ) then begin
    for idx := start to stop do fIndexedLoopNoParametersCallback( idx );
    exit;
  end;
  if assigned( fBoundedLoopCallback ) then begin
    fBoundedLoopCallback( start, stop, Parameters );
    exit;
  end;
  if assigned( fBoundedLoopNoParametersCallback ) then begin
    fBoundedLoopNoParametersCallback( start, stop );
    exit;
  end;
  {$ifdef fpc}
  if assigned( fIndexedLoopObjectCallback ) then begin
    for idx := start to stop do fIndexedLoopObjectCallback( idx, Parameters );
    exit;
  end;
  if assigned( fIndexedLoopObjectNoParametersCallback ) then begin
    for idx := start to stop do fIndexedLoopObjectNoParametersCallback( idx );
    exit;
  end;
  if assigned( fIndexedLoopNestedCallback ) then begin
    for idx := start to stop do fIndexedLoopNestedCallback( idx, Parameters );
    exit;
  end;
  if assigned( fIndexedLoopNestedNoParametersCallback ) then begin
    for idx := start to stop do fIndexedLoopNestedNoParametersCallback( idx );
    exit;
  end;
  if assigned( fBoundedLoopObjectCallback ) then begin
    fBoundedLoopObjectCallback( start, stop, Parameters );
    exit;
  end;
  if assigned( fBoundedLoopObjectNoParametersCallback ) then begin
    fBoundedLoopObjectNoParametersCallback( start, stop );
    exit;
  end;
  if assigned( fBoundedLoopNestedCallback ) then begin
    fBoundedLoopNestedCallback( start, stop, Parameters );
    exit;
  end;
  if assigned( fBoundedLoopNestedNoParametersCallback ) then begin
    fBoundedLoopNestedNoParametersCallback( start, stop );
    exit;
  end;
  {$endif}
end;

class function TLoopDelegate.null(): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
end;

class operator TLoopDelegate.Implicit( const Value: TIndexedLoopCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TIndexedLoopCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopCallback := Value;
end;

class operator TLoopDelegate.Implicit( const Value: TIndexedLoopNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopNoParametersCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TIndexedLoopNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopNoParametersCallback := Value;
end;

class operator TLoopDelegate.Implicit( const Value: TBoundedLoopCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TBoundedLoopCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopCallback := Value;
end;

class operator TLoopDelegate.Implicit( const Value: TBoundedLoopNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopNoParametersCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TBoundedLoopNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopNoParametersCallback := Value;
end;

{$ifdef fpc}

class operator TLoopDelegate.Implicit( const Value: TIndexedLoopObjectCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopObjectCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TIndexedLoopObjectCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopObjectCallback := Value;
end;

class operator TLoopDelegate.Implicit( const Value: TIndexedLoopObjectNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopObjectNoParametersCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TIndexedLoopObjectNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopObjectNoParametersCallback := Value;
end;

class operator TLoopDelegate.Implicit( const Value: TIndexedLoopNestedCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopNestedCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TIndexedLoopNestedCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopNestedCallback := Value;
end;

class operator TLoopDelegate.Implicit( const Value: TIndexedLoopNestedNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopNestedNoParametersCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TIndexedLoopNestedNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fIndexedLoopNestedNoParametersCallback := Value;
end;

class operator TLoopDelegate.Implicit( const Value: TBoundedLoopObjectCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopObjectCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TBoundedLoopObjectCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopObjectCallback := Value;
end;

class operator TLoopDelegate.Implicit( const Value: TBoundedLoopObjectNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopObjectNoParametersCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TBoundedLoopObjectNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopObjectNoParametersCallback := Value;
end;

class operator TLoopDelegate.Implicit( const Value: TBoundedLoopNestedCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopNestedCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TBoundedLoopNestedCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopNestedCallback := Value;
end;

class operator TLoopDelegate.Implicit( const Value: TBoundedLoopNestedNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopNestedNoParametersCallback := Value;
end;

class operator TLoopDelegate.Explicit( const Value: TBoundedLoopNestedNoParametersCallback ): TLoopDelegate;
begin
  Result := Default(  TLoopDelegate );
  Result.fBoundedLoopNestedNoParametersCallback := Value;
end;

{$endif}

{$ifdef fpc}

class function MakeDelegate.From( const Callback: TNotifyCallback ): TNotifyDelegate;
begin
  Result :=  TNotifyDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TNotifyObjectCallback ): TNotifyDelegate;
begin
  Result :=  TNotifyDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TNotifyNestedCallback ): TNotifyDelegate;
begin
  Result :=  TNotifyDelegate( Callback );
end;

class function MakeDelegate.From<T>( const Callback: TNotifyWithSenderCallback<T> ): TNotifyWithSenderDelegate<T>;
begin
  Result :=  TNotifyWithSenderDelegate<T>( Callback );
end;

class function MakeDelegate.From<T>( const Callback: TNotifyWithSenderObjectCallback<T> ): TNotifyWithSenderDelegate<T>;
begin
  Result :=  TNotifyWithSenderDelegate<T>( Callback );
end;

class function MakeDelegate.From<T>( const Callback: TNotifyWithSenderNestedCallback<T> ): TNotifyWithSenderDelegate<T>;
begin
  Result :=  TNotifyWithSenderDelegate<T>( Callback );
end;

class function MakeDelegate.From( const Callback: TIndexedLoopCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TIndexedLoopNoParametersCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TIndexedLoopObjectCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TIndexedLoopObjectNoParametersCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TIndexedLoopNestedCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TIndexedLoopNestedNoParametersCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TBoundedLoopCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TBoundedLoopNoParametersCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TBoundedLoopObjectCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TBoundedLoopObjectNoParametersCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TBoundedLoopNestedCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TBoundedLoopNestedNoParametersCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

{$else}

class function MakeDelegate.From( const Callback: TNotifyCallback ): TNotifyDelegate;
begin
  Result :=  TNotifyDelegate( Callback );
end;

class function MakeDelegate.From<T>( const Callback: TNotifyWithSenderCallback<T> ): TNotifyWithSenderDelegate<T>;
begin
  Result :=  TNotifyWithSenderDelegate<T>( Callback );
end;

class function MakeDelegate.From( const Callback: TIndexedLoopCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TIndexedLoopNoParametersCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TBoundedLoopCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

class function MakeDelegate.From( const Callback: TBoundedLoopNoParametersCallback ): TLoopDelegate;
begin
  Result :=  TLoopDelegate( Callback );
end;

{$endif}

initialization
  xStatusMessages := TStatusMessages.Create;
  TStatus( stSuccess                ).Register('Success');
  TStatus( stUnknown                ).Register('?Unknown?');
  TStatus( stUTFUnknownNotSupported ).Register('TUnicodeFormat.utfUnknown cannot be used to encode or decode.');
  TStatus( stUnicodeEncodingFailed  ).Register('Unicode encoding failed for "(%format%)".');
  TStatus( stUnicodeDecodingFailed  ).Register('Unicode decoding failed for "(%format%)".');
  TStatus( stLowercaseFailed        ).Register('Attempt to encode character as lowercase failed "(%char%)".');
  TStatus( stUppercaseFailed        ).Register('Attempt to encode character as uppercase failed "(%char%)".');
  TStatus( stStrToSingleError       ).Register('String to Single conversion attempt failed: "(%str%)"');
  TStatus( stStrToDoubleError       ).Register('String to Double conversion attempt failed: "(%str%)"');
  TStatus( stStrToExtendedError     ).Register('String to Extended conversion attempt failed: "(%str%)"');
  TStatus( stStrToInt8Error         ).Register('String to Int8 conversion attempt failed: "(%str%)"');
  TStatus( stStrToInt16Error        ).Register('String to Int16 conversion attempt failed: "(%str%)"');
  TStatus( stStrToInt32Error        ).Register('String to Int32 conversion attempt failed: "(%str%)"');
  TStatus( stStrToInt64Error        ).Register('String to Int64 conversion attempt failed: "(%str%)"');
  TStatus( stStrToUInt8Error        ).Register('String to uInt8 conversion attempt failed: "(%str%)"');
  TStatus( stStrToUInt16Error       ).Register('String to uInt16 conversion attempt failed: "(%str%)"');
  TStatus( stStrToUInt32Error       ).Register('String to uInt32 conversion attempt failed: "(%str%)"');
  TStatus( stStrToUInt64Error       ).Register('String to uInt64 conversion attempt failed: "(%str%)"');
  TStatus( stStringToHexFailed      ).Register('String to hex conversion failed due to invalid char "(%char%)"');
  TStatus( stInvalidYear            ).Register('Year "(%year%)" is invalid or out of bounds while encoding date.');
  TStatus( stInvalidMonth           ).Register('Month "(%month%)" is not valid for encoding date.');
  TStatus( stInvalidDate            ).Register('Day "(%day%)" in month "(%month%)" of year "(%year%)" is invalid or out of bounds while encoding date.');
  TStatus( stGUIDConversionError    ).Register('Conversion from string to GUID failed.');
  TStatus( stCreateMutexFailed      ).Register('Failed to create mutex.');
  TStatus( stVariantTypeCoersion    ).Register('TVariant type coersion from "(%sourcetype%)" to "(%targettype%)" failed.');

finalization
  xStatusMessages.Free;

end.

